#    Copyright 2016-2020 University of Leicester & Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pexpect
from devlib import target
from devlib import exception


def setupDevice(os, platform):
    """
    This method setup the remote device and establishes the connection
    with it
    :param os: specify the operating system running on the remote target
    :param platform: specify the remote target platform
    :return: this returns a device setup for operation
    """

    SUPPORTED_PLATFORM = frozenset(["TC2","JUNO"])
    SUPPORTED_OS = frozenset(["android","debian","openembedded-lamp"])

    if os not in SUPPORTED_OS:
        raise ValueError("DEVICE ERROR: {} not a supported os".format(os))

    if platform not in SUPPORTED_PLATFORM:
        raise ValueError("DEVICE ERROR: {} not a supported platform".format(platform))

    print("Configuring a {} platform running {} os".format(platform,os))

    if ("TC2" == platform) and ("android" == os):
        dev = target.AndroidTarget(connect=False)
    elif ("TC2" == platform) and ("debian" == os):
        dev = target.LinuxTarget(connection_settings={'host': '10.1.1.199',
                                                      'username': 'testing',
                                                      'password': 'testing'},
                                 connect=False)
    else:
        print("The {} + {} combination not supported yet".format(platform,os))
        return None

    print("Successfully created the device instance")
    return dev



if __name__ == "__main__":
    try:
        device = setupDevice("android", "TC2")
        device.connect()
    except (ValueError, exception.HostError) as e:
        print("{}".format(str(e)))
    try:
        device = setupDevice("debian", "TC2")
        device.connect()
    except (ValueError, exception.HostError, exception.DevlibError,
            pexpect.pxssh.ExceptionPxssh) as e:
        print("{}".format(str(e)))




