#    Copyright 2016-2020 University of Leicester & Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from sbstauto.devices import devices
from devlib import exception


class DeviceTests(unittest.TestCase):
    """
    Tests for device methods/operations.
    """
    def test_deviceSetup_empty_args(self):
        """
        Test if deviceSetup has right number of arguments - empty args

        :return:
        """
        self.assertRaises(TypeError, devices.setupDevice)

    def test_deviceSetup_incorrect_numberof_args(self):
        """
        Test if deviceSetup Test has right number of arguments - incorrect
        number of args

        :return:
        """
        self.assertRaises(TypeError, devices.setupDevice, "linux", "Juno", "Something Extra")

    def test_deviceSetup_unsupported_os(self):
        """
        Test if deviceSetup handles invalid os and platform selections
        correctly

        :return:
        """
        self.assertRaises(ValueError, devices.setupDevice, "Not An OS", "TC2")

    def test_deviceSetup_unsupported_platform(self):
        """
        Test if deviceSetup handles invalid os and platform selections
        correctly

        :return:
        """
        self.assertRaises(ValueError, devices.setupDevice, "android", "Not A Platform")


    def test_deviceSetup_connection_failure(self):
        """
        Test if deviceSetup handles connection error after SetupDevice
        correctly

        :return:
        """
        dev = devices.setupDevice("android", "TC2")
        self.assertRaises(exception.HostError, dev.connect)


    def test_deviceSetup_invalid_os_platform_combo(self):
        """
        Test if deviceSetup handles attempt to setup invalid
        platform and os combination devices

        :return:
        """
        dev = devices.setupDevice("openembedded-lamp", "TC2")
        self.assertRaises(None, dev)


if __name__ == "__main__":
    unittest.main()
