#    Copyright 2016-2020 University of Leicester & Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pexpect
from devlib import exception
from sbstauto.devices import devices
from sbstauto import config
from sbstauto.sbst_algo.fitnessfunction import pmu_counters


def launcher():
    print("Running test suite from:"+config.SBSTAUTO_HOME)

    try:
        testplatform = devices.setupDevice("android","TC2")
        testplatform.connect()

        print("Setting up fitness function --> Perfcounter")
        pc = pmu_counters.TC2perfCounter()
        try:
            pc.configEventCounters("cci_a7_read_data_stall", "cci_a15_read_data_stall",
                                   "cci_a7_read_data_stall_trackerfull",
                                   "cci_a15_read_data_stall_trackerfull")
            pc.finaliseConfig()
        except (ValueError, OverflowError, IOError, KeyError) as e:
            print(str(e))

        targetconfig = config.SBSTAUTO_HOME+"/output/target.config"
        #print(targetconfig)
        #workinglocation = testplatform.path.join(testplatform.working_directory, 'target.config')
        testplatform.push(targetconfig, testplatform.get_workpath("target.config"))

        output = testplatform.execute("cd /data/local/tmp && mkdir -p bin")
        print(output)
        output = testplatform.execute("mkdir -p /bin && ln -s /system/bin/sh /bin/sh ")
        print(output)

        perfattarget = testplatform.install(pc.whichPerfExe())
        print(perfattarget)

    except (ValueError, exception.HostError, exception.DevlibError,
                pexpect.pxssh.ExceptionPxssh, exception.TargetError) as e:
        print("DEVICE ERROR: {}".format(str(e)))
        print("SBSTAUTOTESTER: Failed to Setup target device!")



if __name__ == '__main__':
    print('Launching SBST Auto Tester')
    launcher()
