/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/
#ifndef _COMMON_H_
#define _COMMON_H_

/*******************************************************************************
 * Add all the common headers to be used in various subcomponent in this       *
 * project to this header file.                                                *
 ******************************************************************************/
#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <mqueue.h>
#include <sys/types.h>
#include <signal.h>
#include <sched.h>
#include <time.h>
#include <math.h>

/*The following is for armv7*/

//#define dsb() __asm__ __volatile__ ("dsb" : : : "memory")

//#define dmb() __asm__ __volatile__ ("dmb" : : : "memory")

/*The following is for aarch64*/

#define dsb() __asm__ __volatile__ ("dsb sy"  : : : "memory")

#define dmb() __asm__ __volatile__ ("dmb sy"  : : : "memory")


#define REPORT_FN_DETAILS                                        \
do {                                                             \
        char* fullname;                                          \
        char delimiter[] = ".";                                  \
        char* tokens;                                            \
        fullname =  strdup(__FILE__);                            \
        tokens = strsep (&fullname, delimiter);                  \
        printf("At %s function in file %s.c\n",__func__,tokens); \
}while (0)

#define TRUE 1
#define FALSE 0

/*This define the maximum number of config sections allowed in target.config*/
#define MAX_CFG_SECTIONS 8
/*This define the maximum allowed items per config section in target.congig*/
#define MAX_ITEMS_PER_CFG_SECTION 8
/*This defines the lenght of the cfg strings from the target.config*/
#define CFG_STR_LEN 512
/*This defined the number of times each iteration is repeated to get a statistically relevant
  result*/
#define AVERAGE_STAT_COUNT 20


/*We use MessageQ as the IPC mechanism, the following lists the MQ details*/
//#define MQ_FF_NAME "/FF_Q"
#define MQ_PMU_FF_NAME "/FF_PMU_Q"
#define MQ_HWMON_FF_NAME "/FF_HWMON_Q"
#define MQ_MESSAGE_MAX_LENGTH 100
#define MQ_MAX_NUM_OF_MESSAGES 10
#define MQ_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)

#define MQ_MESSAGE_DATA "DEFAULT"
#define MQ_MESSAGE_LENGTH 10
#define MQ_MESSAGE_PRIORITY 0

#define SEC_IN_MINUTE (60)
#define SEC_IN_HOUR (60 * 60)
#define SEC_IN_DAY (24 * 60 * 60)

#define EXPERIMENT_DURATION_IN_SEC (14 * 60 * 60)


//#define BASELINE_EXPERIMENT

//#define RANDOM_TESTCASE_BASELINE_EXPERIMENT


typedef enum _error_code
{
    SUCCESS = 0, /*This starting with zero allows to use this enum as an
                   index into the error_map structure                   */
    CONFIG_FAILED,
    IO_ERROR,
    MALFORMED_CONFIG,
    NOT_IMPLEMENTED,
    INCOMPLETE_CONFIG,
    UNSUPPORTED_CONFIG,
    OS_ERROR,
    OPERATION_ERROR

} ERROR_CODE_T;

typedef struct _error_map
{
    ERROR_CODE_T err_code;
    char * err_message;
} ERROR_MAP_T;

typedef struct _cfg_element
{
    char cfg[32];
    char cfg_val[256];
} CFG_ELEMENT_T;

typedef struct _cfg_section
{
    char section[32];
    CFG_ELEMENT_T config[MAX_ITEMS_PER_CFG_SECTION];
} CFG_SECTION_T;

typedef struct _thread_data
{
	int id;
	char mq_name[16];
	pid_t threadid;
}THREADDATA_T;

/*Include all the common headers from the top level of individual components*/
#include "../../sbst_algo/fitnessfunction/include/fitness_function_common.h"
#include "../../sbst_algo/algorithm/include/sbst_common.h"
#include "../../sbst_algo/testpayload/include/payload_common.h"



#endif

