/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/sbst_adaptor.h"

/*Global variables*/
int stat_token_pmu = 0;
int stat_token_hwmon = 0;


/*******************************************************************************
 * int sbst_algo_init_config(ATG_T* algo)                                      *
 *    - This function blah blah                                                *
 ******************************************************************************/
int sbst_algo_init_config(ALG_T* algo)
{
    int ret_val = -1;
    char str[CFG_STR_LEN];
    int sbst_algo_found = FALSE;
    int hc_restart_count_found = FALSE;
    int hc_type_found = FALSE;
    int sa_temp_iterations_found = FALSE;
    int sa_type_found = FALSE;
    int sa_max_subopt_count_found = FALSE;
    char* tmpstr;
    char * token;
    int i;
    int j;
    int value;

    REPORT_FN_DETAILS;

    /*First determine if the needed configuration parameters HILLCLIMB is
      provided via the config file. The first encountered will be used*/
    for (i = 0; i < MAX_CFG_SECTIONS; i++)
    {
        strcpy(str,target_config[i].section);
        if (0 == strcmp("HILLCLIMB", str))
        {
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("type", target_config[i].config[j].cfg))
                    {
                        if (0 == strcmp("hc-rdm-restart", target_config[i].config[j].cfg_val))
                        {
                            *algo = HILLCLIMB_RDM_RESTART;
                            hc_type_found = TRUE;
                            printf("SBST algorithm type found\n");
                        }
                    }
                    else if (0 == strcmp("restart_count", target_config[i].config[j].cfg))
                    {

                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        hc_restart_count = value;
                        hc_restart_count_found = TRUE;
                        printf("hc restart count found\n");

                    }

                    if (hc_type_found && hc_restart_count_found)
                    {
                        sbst_algo_found = TRUE;
                    }


                }


        }
        else if (0 == strcmp("SIMULATEDANNEAL", str))
        {
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("type", target_config[i].config[j].cfg))
                    {
                        if (0 == strcmp("sim-anneal", target_config[i].config[j].cfg_val))
                        {
                            *algo = SIMANNEAL_BASIC;
                            sa_type_found = TRUE;
                            printf("SBST algorithm type found\n");
                        }
                        else if (0 == strcmp("sim-anneal-composite", target_config[i].config[j].cfg_val))
                        {
                            *algo = SIMANNEAL_COMPOSITE;
                            sa_type_found = TRUE;
                            printf("SBST algorithm type found\n");
                        }                   }
                    else if (0 == strcmp("temperature_iterations", target_config[i].config[j].cfg))
                    {

                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        sa_temp_iterations = value;
                        sa_composite_temp_iterations = value;
                        sa_temp_iterations_found = TRUE;
                        printf("sa temperature iterations found\n");

                    }
                    else if (0 == strcmp("max_subopt_count", target_config[i].config[j].cfg))
                    {

                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        sa_max_subopt_count = value;
                        sa_composite_max_subopt_count = value;
                        sa_max_subopt_count_found = TRUE;
                        printf("sa temperature iterations found\n");

                    }

                    if (sa_type_found && sa_temp_iterations_found && sa_max_subopt_count_found)
                    {
                        sbst_algo_found = TRUE;
                    }


                }

        }
        else if (0 == strcmp("SPEA", str))
        {
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("type", target_config[i].config[j].cfg))
                    {
                        if (0 == strcmp("spea-2", target_config[i].config[j].cfg_val))
                        {
                            *algo = SPEA_2;
                            sa_type_found = TRUE;
                            printf("SBST algorithm type found\n");
                        }
                    }
//                    else if (0 == strcmp("temperature_iterations", target_config[i].config[j].cfg))
//                    {

//                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
//                        sa_temp_iterations = value;
//                        sa_temp_iterations_found = TRUE;
//                        printf("sa temperature iterations found\n");

//                    }

//                    if (sa_type_found && sa_temp_iterations_found)
                    if (sa_type_found)
                      {
                        sbst_algo_found = TRUE;
                    }


                }

        }
        else
        {
            //TODO: Add logic when the payload type is something else

        }

        if (sbst_algo_found) /*Upon finding the first valid sbst algorithm stop the search*/
            break;


        }

    if (!sbst_algo_found)
    {
        printf("MAIN$$ ERROR: No valid sbst algorithm specified in target.config\n");

        return INCOMPLETE_CONFIG;
    }

    switch (*algo)
    {
        case HILLCLIMB_RDM_RESTART:
                        printf("MAIN$$ Using %s as SBST Algorithm with restart count set to %d times\n", "Hill Climbing - Random Restart", hc_restart_count);
                        if (hc_restart_count > 50)
                        {
                            /*This is an arbitrary upper limit of 50 restart counts*/
                            return UNSUPPORTED_CONFIG;
                        }
        break;

        case SIMANNEAL_BASIC:
                        printf("MAIN$$ Using %s as SBST Algorithm with temperature iterations set to %d times\n", "Simulated Annealing - Basic", sa_temp_iterations);
                        if (sa_temp_iterations > 100)
                        {
                            /*This is an arbitrary upper limit of 100 temperature iterations01Aug*/
                            return UNSUPPORTED_CONFIG;
                        }
        break;

        case SIMANNEAL_COMPOSITE:
                        printf("MAIN$$ Using %s as SBST Algorithm with temperature iterations set to %d times\n", "Simulated Annealing - Basic", sa_temp_iterations);
                        if (sa_temp_iterations > 100)
                        {
                            /*This is an arbitrary upper limit of 100 temperature iterations01Aug*/
                            return UNSUPPORTED_CONFIG;
                        }
        break;

        case SPEA_2:
                        printf("MAIN$$ Using %s as SBST Algorithm\n", "Spea2");
//                        if (sa_temp_iterations > 100)
//                        {
                            /*This is an arbitrary upper limit of 100 temperature iterations01Aug*/
//                            return UNSUPPORTED_CONFIG;
//                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;
    }

    return SUCCESS;
}


/*******************************************************************************
 * int sbst_algo_init_statespace_and_randomseed(ALG_T algo, PL_T payload)      *
 *    - This function blah blah                                                *
 ******************************************************************************/
int sbst_algo_init_statespace_and_randomseed(ALG_T algo, PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    switch (algo)
    {
        case HILLCLIMB_RDM_RESTART:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = initialise_hc_rdm_restart_setup(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - HillClimbing Random Restart setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }

        break;

        case SIMANNEAL_BASIC:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = initialise_sa_setup(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Simulated Annealing setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }


        break;

        case SIMANNEAL_COMPOSITE:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = initialise_sa_composite_setup(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Simulated Annealing setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }


        break;

        case SPEA_2:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = initialise_spea2_setup(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Spea2 setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }


        break;

        default:
                        return NOT_IMPLEMENTED;
        break;
    }

    return SUCCESS;
}

/*******************************************************************************
 * int sbst_algo_iterate_statespace(ALG_T algo, PL_T payload)                  *
 *    - This function blah blah                                                *
 ******************************************************************************/
int sbst_algo_iterate_statespace(ALG_T algo, PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

//    printf("In sbst_algo_iterate_statespace with ff=%d\n", fitness_fn);

    REPORT_FN_DETAILS;

    switch (algo)
    {
        case HILLCLIMB_RDM_RESTART:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = iterate_hc_rdm_restart_statespace(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - HillClimbing Random Restart iteration failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }

        break;

        case SIMANNEAL_BASIC:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = iterate_sa_statespace(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Simulated Annealing iteration failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }

        break;

        case SIMANNEAL_COMPOSITE:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = iterate_sa_composite_statespace(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Simulated Annealing iteration failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }

        break;

        case SPEA_2:
                    switch (payload)
                    {
                        case CCI_MEM_RDM_RDWR:
                                    printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                                    ret_val = iterate_spea2_statespace(payload, fitness_fn);
                                    if (SUCCESS != ret_val)
                                    {
                                        printf("MAIN$$ ERROR - Spea2 iteration failed with %s -\n",error_map[ret_val].err_message);

                                        return OPERATION_ERROR;
                                    }

                        break;

                        default:
                                    return NOT_IMPLEMENTED;
                        break;

                    }

        break;

        default:
                        return NOT_IMPLEMENTED;
        break;
    }

    return SUCCESS;
}

/*******************************************************************************
 * int sbst_algo_teardown(ALG_T algo, PL_T payload)                            *
 *    - This function blah blah                                                *
 ******************************************************************************/
int sbst_algo_teardown(ALG_T algo, PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    return SUCCESS;
}

