/********************************************************************************
 *  Copyright 2018-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/sbst_generic.h"

/*Global variables*/
/*Specifies the node id in the search space navigated by the algorithm
-1 would indicate current node id and 0 to NEXT_GEN_CANDIDATE_COUNT-1 indicate
next generation nodes*/
int node_id = -100;
int individual_index = -100;
CONSTRAINT_T sparcity_constraint;
ACTORPROFILE_T actor_profile;


/*******************************************************************************
 * static long generate_rand_col(void)                                         *
 *    - This function generates the a random value to be used as the           *
 *      cooirdinate for the col                                                *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static long generate_rand_col(void)
{

    long col_min = 0;
    long col_max;
    long rand_num;

//        REPORT_FN_DETAILS;
	col_max = (long)cols/sparcity_constraint;

//  printf ("cols = %ld sparcity index %d col_max %ld\n",cols, sparcity_constraint,col_max);

    /*We avoid a +1 in col_max - col_min as it is 0 indexed*/
    rand_num = (rand() % (col_max - col_min) ) + col_min;

//	printf("random col = %ld",rand_num);
    return (rand_num);
}


/*******************************************************************************
 * static long generate_rand_row(void)                                         *
 *    - This function generates the a random value to be used as the           *
 *      cooirdinate for the row                                                *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static long generate_rand_row(void)
{

	long row_min = 0;
	long row_max;
	long rand_num;

//        REPORT_FN_DETAILS;

	row_max = (long)rows/sparcity_constraint;

//    printf ("rows = %ld sparcity index %d row_max %ld\n",rows, sparcity_constraint,row_max);

    /*We avoid a +1 in row_max - row_min as it is 0 indexed*/
	rand_num = (rand() % (row_max - row_min)) + row_min;

//    printf("random row = %ld",rand_num);

	return (rand_num);
}


/*******************************************************************************
 * int determine_rows_cols(long long)                                          *
 *    - This function determines the number of rows and columns for the        *
 *      payload representation                                                 *
 *    - arg1 - pl_size_index - specifies the payload size index                *
 ******************************************************************************/
int determine_rows_cols(long long pl_size_index)
{

    REPORT_FN_DETAILS;

/*
64x16 = DEFAULT_CACHELINE_SZIE -> 16
1024x64 = DEFAULT_CACHELINE_SIZE -> 64
1024x512 = DEFAULT_CACHELINE_SIZE -> 512
1024x1024 = DEFAULT_CACHELINE_SIZE -> 1024
8192x4096 = DEFAULT_CACHELINE_SIZE -> 4096
16384x4096 = DEFAULT_CACHELINE_SIZE -> 4096
20480x4096 = DEFAULT_CACHELINE_SIZE -> 4096
24576x4096 = DEFAULT_CACHELINE_SIZE -> 4096
16384x8192 = DEFAULT_CACHELINE_SIZE -> 8192
32768x8192 = DEFAULT_CACHELINE_SIZE -> 8192
65536x8192 = DEFAULT_CACHELINE_SIZE -> 8192
*/

    if (pl_size_index == 1)
	{
		pl_cols_max = 16;
		pl_rows_max = 16;
	}
	else if (pl_size_index == 64)
	{
		pl_cols_max = 64;
		pl_rows_max = 256;
	}
	else if (pl_size_index <= 512)
	{
		pl_cols_max = 512;
		pl_rows_max = (pl_size_index*256)/pl_cols_max;
	}
	else if (pl_size_index <= 1024)
	{
        pl_cols_max = 1024;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
	}
    else if (pl_size_index <= 32768)
    {
        pl_cols_max = 4096;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 65536)
    {
        pl_cols_max = 4096;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 81920)
    {
        pl_cols_max = 4096;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 98304)
    {
        pl_cols_max = 4096;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 131072)
    {
        pl_cols_max = 8192;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 262144)
    {
        pl_cols_max = 8192;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
    else if (pl_size_index <= 524288)
    {
        pl_cols_max = 8192;
        pl_rows_max = (pl_size_index*256)/pl_cols_max;
    }
	else
	{
		printf("Invalid computation of max rows and cols for payload !!!!Terminating!!!!\n");
		return OPERATION_ERROR;
	}

    return SUCCESS;
}

/*******************************************************************************
 * int populate_node(int)                                                      *
 *    - This function populates the payload of the specified index with the    *
 *      random set of memory locations honoring the sparcity constraint        *
 *    - arg1 - node_index - -1 means the current node. other indices are items *
 *                           from the OPEN list in next generation candidates  *
 ******************************************************************************/
int populate_node (int node_index)
{

    long long i=0;
    int j=0;
    long x=0;
    long y=0;

    REPORT_FN_DETAILS;

    /*If the node_index == -1 then this means you are operating on the current node*/
    if (-1 == node_index)
	{
		determine_rows_cols(current_node.pl_size);
		printf("MAIN$$ Populating current node with data payload[%dx%d]\n",pl_rows_max,pl_cols_max);
		current_node.pl_size_row = pl_rows_max;
		current_node.pl_size_col = pl_cols_max;
		sparcity_constraint = current_node.sparcity_index;
		pl_size = current_node.pl_size;

		for (i = 0; i < pl_size; i++)
		{
			x = generate_rand_row();
			y = generate_rand_col();
			cci_mem_rdwr_payload[i].x_pos = x;
			cci_mem_rdwr_payload[i].y_pos = y;
//	        if (i%10 == 0)
//	            printf("\n");
//	      	else
//	            printf("%lld -> (%d %d, %d %d) ", i, x, cci_mem_rdwr_payload[i].x_pos, y, cci_mem_rdwr_payload[i].y_pos);

		}
	}
    /*Else you are operating on the 0 indexed next generation candidate nodes*/
	else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
	{
		determine_rows_cols(next_gen[node_index].pl_size);
		printf("MAIN$$ Populating next node %d with data payload[%dx%d]\n",node_index, pl_rows_max, pl_cols_max);
		next_gen[node_index].pl_size_row = pl_rows_max;
		next_gen[node_index].pl_size_col = pl_cols_max;
		sparcity_constraint = next_gen[node_index].sparcity_index;
		pl_size = next_gen[node_index].pl_size;

		for (i = 0; i < pl_size; i++)
		{
			x = generate_rand_row();
			y = generate_rand_col();
			cci_mem_rdwr_payload[i].x_pos = x;
			cci_mem_rdwr_payload[i].y_pos = y;

//          if (i%10 == 0)
//              printf("\n");
//          else
//              printf("%lld -> (%d %d, %d %d) ", i, x, cci_mem_rdwr_payload[i].x_pos, y, cci_mem_rdwr_payload[i].y_pos);
		}
	}

	return SUCCESS;
}

/*******************************************************************************
 * int populate_node(int)                                                      *
 *    - This function populates the payload of the specified index with the    *
 *      random set of memory locations honoring the sparcity constraint        *
 *    - arg1 - node_index - -1 means the current node. other indices are items *
 *                           from the OPEN list in next generation candidates  *
 ******************************************************************************/
int populate_population_node (int node_index)
{

    long long i=0;
    int j=0;
    long x=0;
    long y=0;

    REPORT_FN_DETAILS;

    /*You are operating on the 0 indexed population nodes*/
    if ((node_index >= 0) && (node_index < SPEA2_POPULATION_SIZE))
	{
		determine_rows_cols(pop[node_index].pl_size);
		printf("MAIN$$ Populating next node %d with data payload[%dx%d]\n",node_index, pl_rows_max, pl_cols_max);
		pop[node_index].pl_size_row = pl_rows_max;
		pop[node_index].pl_size_col = pl_cols_max;
		sparcity_constraint = pop[node_index].sparcity_index;
		pl_size = pop[node_index].pl_size;

		for (i = 0; i < pl_size; i++)
		{
			x = generate_rand_row();
			y = generate_rand_col();
			cci_mem_rdwr_payload[i].x_pos = x;
			cci_mem_rdwr_payload[i].y_pos = y;

//          if (i%10 == 0)
//              printf("\n");
//          else
//              printf("%lld -> (%d %d, %d %d) ", i, x, cci_mem_rdwr_payload[i].x_pos, y, cci_mem_rdwr_payload[i].y_pos);
		}
	}

	return SUCCESS;
}


/*******************************************************************************
 * int clear_computed_scores(EVAL_SCOPE_T clear_scope)                         *
 *    - This function resets the computed_fitness_function element of the      *
 *      payload in current_node and next_gen data structures                   *
 *    - arg1 - clear_scope - indicates what is the scope to clear              *
 ******************************************************************************/
int clear_computed_scores(EVAL_SCOPE_T clear_scope)
{

	int i;
	int j;

    REPORT_FN_DETAILS;

	if (CURRENT_ONLY == clear_scope)
	{
	    for (i = 0; i < perf_counters; i++)
	    {
    		sprintf(current_node.pmu_computed_fitness_score.counter_name+64*i,"%s"," ");
	    	current_node.pmu_computed_fitness_score.counter_score[i] = 0;

    		sprintf(current_node.hwmon_computed_fitness_score.counter_name+64*i,"%s"," ");
	    	current_node.hwmon_computed_fitness_score.counter_score[i] = 0;
	    }
	    current_node.composite_fitness_score.pmu_normalised =  0;
	    current_node.composite_fitness_score.hwmon_normalised =  0;
	}
	else if (CURRENT_AND_BEST == clear_scope)
	{
	    for (i = 0; i < perf_counters; i++)
	    {
            sprintf(current_node.pmu_computed_fitness_score.counter_name+64*i,"%s"," ");
            current_node.pmu_computed_fitness_score.counter_score[i] = 0;

            sprintf(current_node.hwmon_computed_fitness_score.counter_name+64*i,"%s"," ");
            current_node.hwmon_computed_fitness_score.counter_score[i] = 0;

            sprintf(best_node.pmu_computed_fitness_score.counter_name+64*i,"%s"," ");
            best_node.pmu_computed_fitness_score.counter_score[i] = 0;

            sprintf(best_node.hwmon_computed_fitness_score.counter_name+64*i,"%s"," ");
            best_node.hwmon_computed_fitness_score.counter_score[i] = 0;
        }
        current_node.composite_fitness_score.pmu_normalised = 0;
        current_node.composite_fitness_score.hwmon_normalised = 0;
        best_node.composite_fitness_score.pmu_normalised = 0;
        best_node.composite_fitness_score.hwmon_normalised = 0;
	}

	for (i = 0; i < NEXT_GEN_CANDIDATE_COUNT; i++)
	{
        for (j = 0; j < perf_counters; j++ )
        {
    	    sprintf(next_gen[i].pmu_computed_fitness_score.counter_name+64*j,"%s"," ");
            next_gen[i].pmu_computed_fitness_score.counter_score[j] = 0;

    	    sprintf(next_gen[i].hwmon_computed_fitness_score.counter_name+64*j,"%s"," ");
            next_gen[i].hwmon_computed_fitness_score.counter_score[j] = 0;
        }

        next_gen[i].composite_fitness_score.pmu_normalised = 0;
        next_gen[i].composite_fitness_score.hwmon_normalised = 0;

	}


	for (i = 0; i < SPEA2_POPULATION_SIZE; i++)
	{
        for (j = 0; j < perf_counters; j++ )
        {
    	    sprintf(pop[i].pmu_computed_fitness_score.counter_name+64*j,"%s"," ");
            pop[i].pmu_computed_fitness_score.counter_score[j] = 0;

    	    sprintf(pop[i].hwmon_computed_fitness_score.counter_name+64*j,"%s"," ");
            pop[i].hwmon_computed_fitness_score.counter_score[j] = 0;
        }
        pop[i].fitness = 0;

	}


    return SUCCESS;
}

/*******************************************************************************
 * int report_ccitest_fitness_function(int node_index)                      *
 *    - This function reports the fitness function outcome                     *
 *    - arg1 - node_index - represents the node which the data pertains to     *
 ******************************************************************************/
int report_ccitest_fitness_function(int node_index, FIT_FN_T fitness_fn)
{

    static int counter = 0;
    int ret_val = -1;

    REPORT_FN_DETAILS;

	ret_val = fitness_function_evaluate(fitness_fn, node_index);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Fitness Function reporting failed with %s -\n",error_map[ret_val].err_message);
        return OPERATION_ERROR;
    }

    counter++;
	printf("MAIN$$ Counter value = %d\n",counter);

    return SUCCESS;

}


/*******************************************************************************
 * int sbst_ccitest_iterator(FIT_FN_T fitness_fn                               *
 *    - This function send the messages to the payload actors and fitness      *
 *      function inorder to iterate across the state space                     *
 *    - arg1 - fitness_fn - specifies the fitness function                     *
 ******************************************************************************/
int sbst_ccitest_iterator(FIT_FN_T fitness_fn)
{
	int ret;
	int i;

	REPORT_FN_DETAILS;

    /*This global variable indicates the processing of the generated perf report*/
	dat_process_done_pmu = 0;
	dat_process_done_hwmon = 0;

	printf("==============>\n");
//	printf("==============> generation[ %d ] %d %d %d\n",generations,x_pos,y_pos,explorer_index);
	printf("==============>\n");

//    sprintf(output_file,"Iter%d_individual[%d]-token%d.dat",iterator,individual_index,stat_token);
    sprintf(output_file_pmu,"Iter%d_individual[%d]-PMU-token%d.dat",iterator,individual_index,stat_token_pmu);
    sprintf(output_file_hwmon,"Iter%d_individual[%d]-HWMON-token%d.dat",iterator,individual_index,stat_token_hwmon);
//    sprintf(output_file,"Iter%d_individual-token%d.dat",iterator,stat_token);

    /*This should hopefully set the cache state to a similar initial state*/
    reset_to_init_state();

    /*Send the barriers instructions to each data generator and data observer threads*/
    for (i = 0; i < dg_thread_count; i++)
    {
        printf("MAIN$$ Sending message BARRIERS DG[%d].... \n", i);
        ret = mq_send(mqd_wr[i], "BARRIERS", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");
    }

    for (i = 0; i < do_thread_count; i++)
    {
        printf("MAIN$$ Sending message BARRIERS DO[%d].... \n", i);
        ret = mq_send(mqd_rd[i], "BARRIERS", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");
    }

    /*Now invoke the barriers instruction from the test harness main thread itself*/
    set_barriers();

//sleep hack, giving time before FF thread is initiated for counting events
     sleep(1);

//     if (MULTI_OBJ == fitness_fn)
//     {
        /*Initiate the payload execution start with fitness function monitoring being initiated*/
        printf("MAIN$$ Sending message PMU FF.... ");
        ret = mq_send(mqd_pmu_ff, "INITIATE", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");

       /*Initiate the payload execution start with fitness function monitoring being initiated*/
        printf("MAIN$$ Sending message HWMON FF.... ");
        ret = mq_send(mqd_hwmon_ff, "INITIATE", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");
//     }
//     else
//     {
        /*Assuming single objective fitness function in this case*/
        /*Initiate the payload execution start with fitness function monitoring being initiated*/
//        printf("MAIN$$ Sending message FF.... ");
//        ret = mq_send(mqd_singleobj_ff, "INITIATE", 9, MQ_MESSAGE_PRIORITY);
//        if(ret)
//            printf("MAIN$$ Failed\n");
//        else
//            printf("MAIN$$ Done\n");
//     }

//sleep hack, need sync semaphore to ensure perf started running when DG and DO messages are sent
     sleep(1);

//	printf("Hit any key after FF being triggered\n");
//    getchar();
    for (i = 0; i < dg_thread_count; i++)
    {
        /*Using MsgQ instruct data generator to perform some operations*/
        printf("MAIN$$ Sending message DG[%d].... \n", i);
        ret = mq_send(mqd_wr[i], "INITIATE", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");

//        printf("Hit any key after DG being triggered\n");
//        getchar();

        if ((SINGLE_RDWR_AGENT == actor_profile) && (0 == i))
        {
            /*if single agent then break the loop*/
            break;
        }

        if ((DUAL_RDWR_AGENT == actor_profile) && (1 == i))
        {
            /*if double agent then break the loop*/
            break;
        }
        /*In case of MULTIPLE_RDWR_AGENT proceed to send the initiation signal across all
          data generator threads, so don't break the loop*/

    }
    /*Do we really need to sleep???*/
	sleep(1);

    for (i = 0; i < do_thread_count; i++)
    {
        /*Using MsgQ instruct data observor to perform some operations*/
        printf("MAIN$$ Sending message DO[%d].... \n", i);
        ret = mq_send(mqd_rd[i], "INITIATE", 9, MQ_MESSAGE_PRIORITY);
        if(ret)
            printf("MAIN$$ Failed\n");
        else
            printf("MAIN$$ Done\n");

//        printf("Hit any key after DO being triggered\n");
//        getchar();

        if ((SINGLE_RDWR_AGENT == actor_profile) && (0 == i))
        {
            /*if single agent then break the loop*/
            break;
        }

        if ((DUAL_RDWR_AGENT == actor_profile) && (1 == i))
        {
            /*if double agent then break the loop*/
            break;
        }
        /*In case of MULTIPLE_RDWR_AGENT proceed to send the initiation signal across all
          data generator threads, so don't break the loop*/

    }

    printf("MAIN$$ Waiting for dat processing being completed.... \n");

//		while(!dat_process_done)
//		{
//			/*Waiting till data processing is done*/
//			sleep(1);
//		}

    iterator++;

    return SUCCESS;
}


/*******************************************************************************
 * static int evaluate_current_node (EVAL_SCOPE_T clear_scope)                 *
 *    - This function evaluates the scores for current node                    *
 *    - arg1 - clear_scope - indicates what is the scope to clear              *
 ******************************************************************************/
static int evaluate_current_node(EVAL_SCOPE_T clear_score, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

	if (CURRENT_AND_BEST == clear_score)
        clear_computed_scores(CURRENT_AND_BEST);
	else if (CURRENT_ONLY == clear_score)
		clear_computed_scores(CURRENT_ONLY);

    pl_size = current_node.pl_size;
	printf("MAIN$$ -----------------> PL size is set as %lld %lld\n",pl_size,current_node.pl_size);
    actor_profile = current_node.actor_index;

    individual_index = -1;
    /*Pass -1 to indicate that current node is the one needing to be populated*/
    populate_node(-1);

//    for (stat_token = 0; stat_token < AVERAGE_STAT_COUNT; stat_token++)
    for (stat_token_pmu = 0, stat_token_hwmon = 0; (stat_token_pmu < AVERAGE_STAT_COUNT) && (stat_token_hwmon < AVERAGE_STAT_COUNT); stat_token_pmu++, stat_token_hwmon++)
    {
//	    ret_val = reset_to_init_state();
//        if (SUCCESS != ret_val)
//        {
//            printf("MAIN$$ ERROR - HillClimbing Random Restart setup initialisation failed with %s -\n",error_map[ret_val].err_message);
//            return OPERATION_ERROR;
//        }

        /*This global variable indicates the processing of the generated perf report*/
//        dat_process_done = 0;

        ret_val = sbst_ccitest_iterator(fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - CCI search space iteration failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }
        /*We know for sure that perf userspace utility is configured to run for
          perf_monotor_duration, So wait for a little more than that before
          attempting to process the perf output*/
        sleep(perf_monitor_duration+2);

        /*-1 indicate the current node*/
        ret_val = report_ccitest_fitness_function(-1, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - HillClimbing Random Restart search space iteration failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }
    }

    return SUCCESS;
}

/*******************************************************************************
 * int evaluate_population_nodes (EVAL_SCOPE_T clear_scope)                    *
 *    - This function evaluates the scores for current node                    *
 *    - arg1 - clear_scope - indicates what is the scope to clear              *
 ******************************************************************************/
int evaluate_population_nodes(EVAL_SCOPE_T clear_score, FIT_FN_T fitness_fn, int node_index)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

	/*Check if the candidate in the population is active before evaluating*/
	if (1 == pop[node_index].active)
	{
        pl_size = pop[node_index].pl_size;
	    printf("MAIN$$ -----------------> PL size is set as %lld %lld\n",pl_size,pop[node_index].pl_size);
        actor_profile = pop[node_index].actor_index;

        individual_index = node_index;
        /*Pass -1 to indicate that current node is the one needing to be populated*/
        populate_population_node(node_index);

//      for (stat_token = 0; stat_token < AVERAGE_STAT_COUNT; stat_token++)
        for (stat_token_pmu = 0, stat_token_hwmon = 0; (stat_token_pmu < AVERAGE_STAT_COUNT) && (stat_token_hwmon < AVERAGE_STAT_COUNT); stat_token_pmu++, stat_token_hwmon++)
        {
//	          ret_val = reset_to_init_state();
//            if (SUCCESS != ret_val)
//            {
//                  printf("MAIN$$ ERROR - HillClimbing Random Restart setup initialisation failed with %s -\n",error_map[ret_val].err_message);
//                  return OPERATION_ERROR;
//            }

              /*This global variable indicates the processing of the generated perf report*/
//            dat_process_done = 0;

            ret_val = sbst_ccitest_iterator(fitness_fn);
            if (SUCCESS != ret_val)
            {
                printf("MAIN$$ ERROR - CCI search space iteration failed with %s -\n",error_map[ret_val].err_message);
                return OPERATION_ERROR;
            }
            /*We know for sure that perf userspace utility is configured to run for
                perf_monotor_duration, So wait for a little more than that before
                attempting to process the perf output*/
            sleep(perf_monitor_duration+2);

            /*-1 indicate the current node*/
            ret_val = report_ccitest_fitness_function(node_index, fitness_fn);
            if (SUCCESS != ret_val)
            {
                printf("MAIN$$ ERROR - SPEA2 search space iteration failed with %s -\n",error_map[ret_val].err_message);
                return OPERATION_ERROR;
            }
        }
    }


    return SUCCESS;
}




#ifdef BASELINE_EXPERIMENT

/*******************************************************************************
 * int generate_initial_seed(PL_T payload, int first_instance)                 *
 *    - This function blah blah                                                *
 ******************************************************************************/
int generate_initial_seed(PL_T payload, int first_instance, FIT_FN_T fitness_fn)
{
	int rand_size;
	int rand_constraint;
	int remainder;
	int rand_actor_profile;
    int ret_val = -1;

	REPORT_FN_DETAILS;

    /*Initialise the random generator method once in the exeriment*/
	if (TRUE == first_instance)
		srand(time(NULL));

    /*Initialise rand_size predefined size from
    PL_SIZE_1K = 1
    PL_SIZE_64K = 64
    PL_SIZE_512K = 512
    PL_SIZE_1M = 1024
    PL_SIZE_32M = 32768
    PL_SIZE_64M = 65536
    PL_SIZE_80M = 81920
    PL_SIZE_96M = 98304
    PL_SIZE_128M = 131072
    PL_SIZE_256M = 262144
    PL_SIZE_512M = 524288*/
	rand_size = 32768;

    /*Now populate the current node size attribute with this*/
    current_node.pl_size = rand_size;

    /*Pick sparcity index from predefined entries of
    CONSTRAINT_HIGHEST
    CONSTRAINT_HIGH_MEDIUM
    CONSTRAINT_LOW_MEDIUM
    UNCONSTRAINT*/
	rand_constraint = (rand() % 4) + 1;

    /*Now populate the current node sparcity index with this*/
	current_node.sparcity_index = UNCONSTRAINT;

    /*Pick a random actor profile*/
	rand_actor_profile =  (rand() % 3) + 1;

    /*Now populate the current node sparcity index from predefined entries of
    SINGLE_RDWR_AGENT
    DUAL_RDWR_AGENT
    MULTIPLE_RDWR_AGENT*/
	current_node.actor_index = MULTIPLE_RDWR_AGENT;

    printf("MAIN$$ PL Size initialised with random seed value %lld and Sparcity Index %d Actor Index %d :: \n", current_node.pl_size, current_node.sparcity_index, current_node.actor_index);



	if (TRUE == first_instance)
	{
		ret_val = evaluate_current_node(CURRENT_AND_BEST, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

            return OPERATION_ERROR;
        }
	}
	else
    {
        ret_val = evaluate_current_node(CURRENT_ONLY, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

            return OPERATION_ERROR;
        }

    }

	return SUCCESS;

}

#else

/*******************************************************************************
 * int generate_initial_seed(PL_T payload, int first_instance)                 *
 *    - This function blah blah                                                *
 ******************************************************************************/
int generate_initial_seed(PL_T payload, int first_instance, FIT_FN_T fitness_fn)
{
	int rand_size;
	int rand_constraint;
	int remainder;
	int rand_actor_profile;
    int ret_val = -1;

	REPORT_FN_DETAILS;

    /*Initialise the random generator method once in the exeriment*/
	if (TRUE == first_instance)
		srand(time(NULL));

    /*Pick a random size attribute*/
	rand_size = (rand() % (PL_SIZE_MAX + 1 - PL_SIZE_MIN)) + PL_SIZE_MIN;

    /*Ensure the size is a multiple of defined size step value*/
	remainder = rand_size % PL_SIZE_STEP;

	if (remainder != 0)
	{
		if (rand_size < PL_SIZE_STEP)
			rand_size = PL_SIZE_MIN;
		else
			rand_size = rand_size + PL_SIZE_STEP - remainder;
	}

    /*Now populate the current node size attribute with this*/
    current_node.pl_size = rand_size;

    /*Pick a random sparcity index attribute*/
	rand_constraint = (rand() % 4) + 1;

    /*Now populate the current node sparcity index with this*/
	if (rand_constraint == 4)
		current_node.sparcity_index = CONSTRAINT_HIGHEST;
	else if (rand_constraint == 3)
		current_node.sparcity_index = CONSTRAINT_HIGH_MEDIUM;
	else if (rand_constraint == 2)
		current_node.sparcity_index = CONSTRAINT_LOW_MEDIUM;
	else
		current_node.sparcity_index = UNCONSTRAINT;

    /*Pick a random actor profile*/
	rand_actor_profile =  (rand() % 3) + 1;

    /*Now populate the current node sparcity index with this*/
	if (rand_actor_profile == 1)
		current_node.actor_index = SINGLE_RDWR_AGENT;
	else if (rand_actor_profile == 2)
		current_node.actor_index = DUAL_RDWR_AGENT;
	else
		current_node.actor_index = MULTIPLE_RDWR_AGENT;

    printf("MAIN$$ PL Size initialised with random seed value %lld and Sparcity Index %d Actor Index %d :: \n", current_node.pl_size, current_node.sparcity_index, current_node.actor_index);



	if (TRUE == first_instance)
	{
		ret_val = evaluate_current_node(CURRENT_AND_BEST, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

            return OPERATION_ERROR;
        }
	}
	else
    {
        ret_val = evaluate_current_node(CURRENT_ONLY, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

            return OPERATION_ERROR;
        }

    }

	return SUCCESS;

}

#endif

/*******************************************************************************
 * int generate_initial_population(PL_T payload, int first_instance)                 *
 *    - This function blah blah                                                *
 ******************************************************************************/
int generate_initial_population(PL_T payload, int first_instance, FIT_FN_T fitness_fn)
{
	int rand_size;
	int rand_constraint;
	int remainder;
	int rand_actor_profile;
    int ret_val = -1;
    int i = 0;

	REPORT_FN_DETAILS;

    /*Initialise the random generator method once in the exeriment*/
	if (TRUE == first_instance)
	{
		srand(time(NULL));
        clear_computed_scores(CURRENT_AND_BEST);
	}

	for (i = 0; i < SPEA2_POPULATION_SIZE; i++)
	{

        /*Pick a random size attribute*/
	    rand_size = (rand() % (PL_SIZE_MAX + 1 - PL_SIZE_MIN)) + PL_SIZE_MIN;

        /*Ensure the size is a multiple of defined size step value*/
	    remainder = rand_size % PL_SIZE_STEP;

	    if (remainder != 0)
	    {
		    if (rand_size < PL_SIZE_STEP)
			    rand_size = PL_SIZE_MIN;
		    else
			    rand_size = rand_size + PL_SIZE_STEP - remainder;
	    }

        /*Now populate the current node size attribute with this*/
        pop[i].pl_size = rand_size;

        /*Pick a random sparcity index attribute*/
	    rand_constraint = (rand() % 4) + 1;

        /*Now populate the current node sparcity index with this*/
	    if (rand_constraint == 4)
		    pop[i].sparcity_index = CONSTRAINT_HIGHEST;
	    else if (rand_constraint == 3)
		    pop[i].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
	    else if (rand_constraint == 2)
		    pop[i].sparcity_index = CONSTRAINT_LOW_MEDIUM;
	    else
		    pop[i].sparcity_index = UNCONSTRAINT;

        /*Pick a random actor profile*/
	    rand_actor_profile =  (rand() % 3) + 1;

        /*Now popolate the current node sparcity index with this*/
	    if (rand_actor_profile == 1)
		    pop[i].actor_index = SINGLE_RDWR_AGENT;
	    else if (rand_actor_profile == 2)
		    pop[i].actor_index = DUAL_RDWR_AGENT;
	    else
		    pop[i].actor_index = MULTIPLE_RDWR_AGENT;

        printf("MAIN$$ PL Size initialised with random seed value %lld and Sparcity Index %d Actor Index %d :: \n", pop[i].pl_size, pop[i].sparcity_index, pop[i].actor_index);

        pop[i].active = 1;

        ret_val = evaluate_population_nodes(CURRENT_AND_BEST, fitness_fn, i);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

            return OPERATION_ERROR;
        }

        /*Here we set the nextgen population as inactive*/
        pop_nextgen[i].active = 0;
	}


	for (i = SPEA2_POPULATION_SIZE; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
	    pop[i].active = 0;
	    pop_nextgen[i].active = 0;
    }

	return SUCCESS;

}
