/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/sbst_hc.h"

/*Global variables*/
int hc_restart_count = 0;
int hc_seeding_firstflag = TRUE;


/*******************************************************************************
 * void function01 (void)                                                      *
 *    - This function blah blah                                                *
 ******************************************************************************/

void function01 (void)
{
	REPORT_FN_DETAILS;


}


/*******************************************************************************
 * static long compute_size_step(void)                                         *
 *    - This function generates the a random step for the payload size         *
 *      attribute                                                *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static long compute_size_step(void)
{

// Capturing the frequency index as
// 1/25 chance for size step [1MB, 4MB]
// 3/25 chance for size step [4MB, 8MB]
// 17/25 chance for size step [8MB, 12MB]
// 3/25 chance for size step [12MB, 16MB]
// 1/25 chance for size step [16MB, 20MB]
    int freq_array[] = {3,3,2,3,3,4,3,3,3,1,3,3,2,3,3,3,4,3,5,3,2,3,4,3,3};
    long size_step;
    int index;

//        REPORT_FN_DETAILS;

    index = rand() % 25; //25 because we want to generate an index
                         //into the freq_array which is a 25 element array

//	printf("freq_array[%d] = %d", index, freq_array[index]);

    if (1 == freq_array[index])
    {
        size_step = (rand() % 4) + 1;
    }
    else if (5 == freq_array[index])
    {
        size_step = (rand() % 5) + 16;
    }
    else if (2 == freq_array[index])
    {
        size_step = (rand() % 5) + 4;
    }
    else if (4 == freq_array[index])
    {
        size_step = (rand() % 5) + 12;
    }
    else
    {
        /*We default to the max probability case of freq_array[index] == 3*/
        size_step = (rand() % 5) + 8;

    }

//	printf("pl_size_step = %ld", size_step * 1024);

    return (size_step * PL_SIZE_STEP);
}


/*******************************************************************************
 * int initialise_hc_rdm_restart_setup(PL_T payload)                           *
 *    - This function blah blah                                                *
 ******************************************************************************/
int initialise_hc_rdm_restart_setup(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                ret_val = generate_initial_seed(payload, hc_seeding_firstflag, fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                    return OPERATION_ERROR;
                }
                hc_seeding_firstflag = FALSE;

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}

#ifdef BASELINE_EXPERIMENT
/*******************************************************************************
 * int generate_next_nodes(void)                                               *
 *    - This function generates the next set of neighbours in the state space  *
 *      Do note that in the current implementation the number of next gen nodes*
 *      is hardcoded and fixed to 11 neighbouring candidates                   *
 *    - arg1 - none                                                            *
 ******************************************************************************/
int generate_next_nodes(void)
{

	int i = 0;

    REPORT_FN_DETAILS;


//	Size same as current node
	next_gen[0].pl_size = current_node.pl_size;
//	Sparcity index same as current node
	next_gen[0].sparcity_index = current_node.sparcity_index;
//	Actor index same as current node
	next_gen[0].actor_index = current_node.actor_index;

//	Size same as current node
	next_gen[1].pl_size = current_node.pl_size;
//	Sparcity index same as current node
	next_gen[1].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
    next_gen[1].actor_index = current_node.actor_index;

//	Size same as current node
	next_gen[2].pl_size = current_node.pl_size;
    next_gen[3].pl_size = current_node.pl_size;
    next_gen[4].pl_size = current_node.pl_size;
//	Actor index same as current node
    next_gen[2].actor_index = current_node.actor_index;
    next_gen[3].actor_index = current_node.actor_index;
    next_gen[4].actor_index = current_node.actor_index;


//	Sparcity index same as current node
	next_gen[2].sparcity_index = current_node.sparcity_index;
	next_gen[3].sparcity_index = current_node.sparcity_index;
	next_gen[4].sparcity_index = current_node.sparcity_index;

//	actor profile different from current node
	next_gen[5].pl_size = current_node.pl_size;
	next_gen[6].pl_size = current_node.pl_size;

	next_gen[5].sparcity_index = current_node.sparcity_index;
	next_gen[6].sparcity_index = current_node.sparcity_index;

	next_gen[5].actor_index = current_node.actor_index;
	next_gen[6].actor_index = current_node.actor_index;

//  Size same as current node
////    next_gen[7].pl_size = current_node.pl_size;
//  Sparcity index same as current node
////    next_gen[7].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
////    next_gen[7].actor_index = current_node.actor_index;

//  Size same as current node
////    next_gen[8].pl_size = current_node.pl_size;
//  Sparcity index same as current node
////    next_gen[8].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
////    next_gen[8].actor_index = current_node.actor_index;


//  Size same as current node
////    next_gen[9].pl_size = current_node.pl_size;
//  Sparcity index same as current node
////    next_gen[9].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
////    next_gen[9].actor_index = current_node.actor_index;

//  Size same as current node
////    next_gen[10].pl_size = current_node.pl_size;
//  Sparcity index same as current node
////    next_gen[10].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
////    next_gen[10].actor_index = current_node.actor_index;

       return SUCCESS;
}

#else

#ifdef RANDOM_TESTCASE_BASELINE_EXPERIMENT


/*******************************************************************************
 * int generate_next_nodes(void)                                               *
 *    - This function generates the next set of neighbours in the state space  *
 *      Do note that in the current implementation the number of next gen nodes*
 *      is hardcoded and fixed to 11 neighbouring candidates                   *
 *    - arg1 - none                                                            *
 ******************************************************************************/
int generate_next_nodes(void)
{

	int i = 0;
	int rand_size;
	int rand_constraint;
	int remainder;
	int rand_actor_profile;
    int ret_val = -1;

    REPORT_FN_DETAILS;

    for (i = 0; i < NEXT_GEN_CANDIDATE_COUNT; i++)
    {
        /*Pick a random size attribute*/
        rand_size = (rand() % (PL_SIZE_MAX + 1 - PL_SIZE_MIN)) + PL_SIZE_MIN;

        /*Ensure the size is a multiple of defined size step value*/
        remainder = rand_size % PL_SIZE_STEP;

        if (remainder != 0)
        {
            if (rand_size < PL_SIZE_STEP)
                rand_size = PL_SIZE_MIN;
            else
                rand_size = rand_size + PL_SIZE_STEP - remainder;
        }

        /*Now populate the current node size attribute with this*/
        next_gen[i].pl_size = rand_size;

        /*Pick a random sparcity index attribute*/
        rand_constraint = (rand() % 4) + 1;

        /*Now populate the current node sparcity index with this*/
        if (rand_constraint == 4)
            next_gen[i].sparcity_index = CONSTRAINT_HIGHEST;
        else if (rand_constraint == 3)
            next_gen[i].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
        else if (rand_constraint == 2)
            next_gen[i].sparcity_index = CONSTRAINT_LOW_MEDIUM;
        else
            next_gen[i].sparcity_index = UNCONSTRAINT;

        /*Pick a random actor profile*/
        rand_actor_profile =  (rand() % 3) + 1;

        /*Now popolate the current node sparcity index with this*/
        if (rand_actor_profile == 1)
            next_gen[i].actor_index = SINGLE_RDWR_AGENT;
        else if (rand_actor_profile == 2)
            next_gen[i].actor_index = DUAL_RDWR_AGENT;
        else
            next_gen[i].actor_index = MULTIPLE_RDWR_AGENT;

        printf("MAIN$$ PL Size initialised with random seed value %lld and Sparcity Index %d Actor Index %d :: \n", next_gen[i].pl_size, next_gen[i].sparcity_index, next_gen[i].actor_index);

    }

      return SUCCESS;
}


# else


/*******************************************************************************
 * int generate_next_nodes(void)                                               *
 *    - This function generates the next set of neighbours in the state space  *
 *      Do note that in the current implementation the number of next gen nodes*
 *      is hardcoded and fixed to 11 neighbouring candidates                   *
 *    - arg1 - none                                                            *
 ******************************************************************************/
int generate_next_nodes(void)
{

	int i = 0;
	long computed_pl_size_step = 0;

    REPORT_FN_DETAILS;


    computed_pl_size_step = compute_size_step();
//	Size + step candidate
	next_gen[0].pl_size = ((current_node.pl_size + computed_pl_size_step) >= PL_SIZE_MAX) ? PL_SIZE_MAX : current_node.pl_size + computed_pl_size_step ;
//	Sparcity index same as current node
	next_gen[0].sparcity_index = current_node.sparcity_index;
//	Actor index same as current node
	next_gen[0].actor_index = current_node.actor_index;

//	Size - step candidate
	next_gen[1].pl_size = ((current_node.pl_size - computed_pl_size_step) <= PL_SIZE_MIN) ? PL_SIZE_MIN : current_node.pl_size - computed_pl_size_step;
//	Sparcity index same as current node
	next_gen[1].sparcity_index = current_node.sparcity_index;
//  Actor index same as current node
    next_gen[1].actor_index = current_node.actor_index;

//	Size same as current node
	next_gen[2].pl_size = current_node.pl_size;
    next_gen[3].pl_size = current_node.pl_size;
    next_gen[4].pl_size = current_node.pl_size;
//	Actor index same as current node
    next_gen[2].actor_index = current_node.actor_index;
    next_gen[3].actor_index = current_node.actor_index;
    next_gen[4].actor_index = current_node.actor_index;


//	Sparcity index different from current node
	if (CONSTRAINT_HIGHEST == current_node.sparcity_index)
	{
		next_gen[2].sparcity_index = UNCONSTRAINT;
		next_gen[3].sparcity_index = CONSTRAINT_LOW_MEDIUM;
		next_gen[4].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
	}
	else if (CONSTRAINT_HIGH_MEDIUM == current_node.sparcity_index)
	{
        next_gen[2].sparcity_index = UNCONSTRAINT;
        next_gen[3].sparcity_index = CONSTRAINT_LOW_MEDIUM;
        next_gen[4].sparcity_index = CONSTRAINT_HIGHEST;
	}
    else if (CONSTRAINT_LOW_MEDIUM == current_node.sparcity_index)
    {
        next_gen[2].sparcity_index = UNCONSTRAINT;
        next_gen[3].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
        next_gen[4].sparcity_index = CONSTRAINT_HIGHEST;
    }
    else if (UNCONSTRAINT == current_node.sparcity_index)
    {
        next_gen[2].sparcity_index = CONSTRAINT_LOW_MEDIUM;
        next_gen[3].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
        next_gen[4].sparcity_index = CONSTRAINT_HIGHEST;
    }
	else
	{
		printf("MAIN$$ Should never reach here...!!!\n");
		return OPERATION_ERROR;
	}

//	actor profile different from current node
	next_gen[5].pl_size = current_node.pl_size;
	next_gen[6].pl_size = current_node.pl_size;

	next_gen[5].sparcity_index = current_node.sparcity_index;
	next_gen[6].sparcity_index = current_node.sparcity_index;

	if (SINGLE_RDWR_AGENT == current_node.actor_index)
	{
		next_gen[5].actor_index = DUAL_RDWR_AGENT;
		next_gen[6].actor_index = MULTIPLE_RDWR_AGENT;
	}
	else if (DUAL_RDWR_AGENT == current_node.actor_index)
	{
		next_gen[5].actor_index = SINGLE_RDWR_AGENT;
		next_gen[6].actor_index = MULTIPLE_RDWR_AGENT;
	}
	else if (MULTIPLE_RDWR_AGENT == current_node.actor_index)
	{
		next_gen[5].actor_index = SINGLE_RDWR_AGENT;
		next_gen[6].actor_index = DUAL_RDWR_AGENT;
	}
    else
    {
        printf("Should never reach here...!!!\n");
    }

//      Size + 2*step candidate
////        next_gen[7].pl_size = ((current_node.pl_size + 2*PL_SIZE_STEP) >= PL_SIZE_MAX) ? PL_SIZE_MAX : current_node.pl_size + 2*PL_SIZE_STEP ;
//      Sparcity index same as current node
////        next_gen[7].sparcity_index = current_node.sparcity_index;
//      Actor index same as current node
////        next_gen[7].actor_index = current_node.actor_index;

//      Size - 2*step candidate
////        next_gen[8].pl_size = ((current_node.pl_size - 2*PL_SIZE_STEP) <= PL_SIZE_MIN) ? PL_SIZE_MIN : current_node.pl_size - 2*PL_SIZE_STEP;
//      Sparcity index same as current node
////        next_gen[8].sparcity_index = current_node.sparcity_index;
//      Actor index same as current node
////        next_gen[8].actor_index = current_node.actor_index;


//      Size + 3*step candidate
////        next_gen[9].pl_size = ((current_node.pl_size + 3*PL_SIZE_STEP) >= PL_SIZE_MAX) ? PL_SIZE_MAX : current_node.pl_size + 3*PL_SIZE_STEP ;
//      Sparcity index same as current node
////        next_gen[9].sparcity_index = current_node.sparcity_index;
//      Actor index same as current node
////        next_gen[9].actor_index = current_node.actor_index;

//      Size - 3*step candidate
////        next_gen[10].pl_size = ((current_node.pl_size - 3*PL_SIZE_STEP) <= PL_SIZE_MIN) ? PL_SIZE_MIN : current_node.pl_size - 3*PL_SIZE_STEP;
//      Sparcity index same as current node
////        next_gen[10].sparcity_index = current_node.sparcity_index;
//      Actor index same as current node
////        next_gen[10].actor_index = current_node.actor_index;

       return SUCCESS;
}

#endif //For RANDOM_TESTCASE_BASELINE_EXPERIMENT

#endif // For BASELINE_EXPERIMENT


#ifdef RANDOM_TESTCASE_BASELINE_EXPERIMENT

/*******************************************************************************
 * static int evaluate_and_select_next_gen_nodes(void)                         *
 *    - This function evaluates the scores for next generation nodes and       *
 *      selects the next best candidate to progress the state space search     *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int evaluate_and_select_next_gen_nodes(FIT_FN_T fitness_fn)
{

    int i = 0;
	int index = 0;
	int tracker[NEXT_GEN_CANDIDATE_COUNT];
	int best_found = 0;
	long best_score = 0;
	char str[128];
	int node_index = -1;
	FILE* report_fd;
	FILE* shadowreport_fd;
	int ret_val = -1;
	int j = 0;

    REPORT_FN_DETAILS;

    clear_computed_scores(NEXTGEN_ONLY);

	for (i = 0; i < NEXT_GEN_CANDIDATE_COUNT; i++)
		tracker[i] = 0;

	i = 0;
	/*Here with single objective evaluation we take into account only
	  the first of the selected counters for evaluation*/
	if (PMU_MON == fitness_fn)
	{
    	best_score = current_node.pmu_computed_fitness_score.counter_score[0];
	}
	else if (HW_MON == fitness_fn)
	{
    	best_score = current_node.hwmon_computed_fitness_score.counter_score[0];
	}
	else
	{
        printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
	}

//	while (!best_found && i < NEXT_GEN_CANDIDATE_COUNT)
//Do it exhaustively
	  while (i < NEXT_GEN_CANDIDATE_COUNT)
      {
    	index = rand() % NEXT_GEN_CANDIDATE_COUNT;
		if (0 == tracker[index])
		{
			tracker[index] = 1;
			i++;

            pl_size = next_gen[index].pl_size;
			printf("MAIN$$ INDEX is ----> %d\n",index);
            printf("MAIN$$ -----------------> PL size is set as %lld %lld\n",pl_size,next_gen[index].pl_size);

            actor_profile = next_gen[index].actor_index;

            individual_index = index;
            populate_node(index);

//            for (stat_token = 0; stat_token < AVERAGE_STAT_COUNT; stat_token++)
            for (stat_token_pmu = 0, stat_token_hwmon = 0; (stat_token_pmu < AVERAGE_STAT_COUNT) && (stat_token_hwmon < AVERAGE_STAT_COUNT); stat_token_pmu++, stat_token_hwmon++)
            {

        	    ret_val = reset_to_init_state();
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }

                /*This global variable indicates the processing of the generated perf report*/
//                dat_process_done = 0;

                ret_val = sbst_ccitest_iterator(fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart search space iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }

                /*We know for sure that perf userspace utility is configured to run for
                  perf_monotor_duration, So wait for a little more than that before
                  attempting to process the perf output*/
                sleep(perf_monitor_duration+2);

                /*index indicate the next gen candidate node*/
                ret_val = report_ccitest_fitness_function(index, fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart search space iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }


            }/*End of for loop*/

			//Check if there is a best score than current?
            /*Here with single objective evaluation we take into account only
	          the first of the selected counters for evaluation*/
// We don't need this check. Just record the current and next
//			if (next_gen[index].computed_fitness_score.counter_score[0] >= current_node.computed_fitness_score.counter_score[0])
			{
// We don't set best_found as we want to iterate across them all
//				best_found = 1;

			    printf("MAIN$$ SUCCESS Next best candidate found @ next_gen[%d] with score %ld -- size[%d,%d]\n ",index,next_gen[index].computed_fitness_score.counter_score[0],next_gen[index].pl_size_row,\
                next_gen[index].pl_size_col);

            	printf("MAIN$$ Before generational change\n");
        	    printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        		printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        		printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        		printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        		printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        		if (PMU_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                    }
        		}
        		else if (HW_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                    }
        		}
        		else
        		{
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        		}

        		report_fd = fopen(report_singleobj_name,"a+");
        		if (NULL == report_fd)
        		{
                    printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
                	return IO_ERROR;
        		}

        		shadowreport_fd = fopen(report_shadowobj_name,"a+");
        		if (NULL == shadowreport_fd)
        		{
                    printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
                	return IO_ERROR;
        		}

        		sprintf(str,"%s","\n\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^");
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

                if (PMU_MON == fitness_fn)
                {
                    sprintf(str,"\n$$$ Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                        current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
                }
                else if (HW_MON == fitness_fn)
                {
                    sprintf(str,"\n$$$ Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                        current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
                }
                else
                {
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
                }

        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

               	// Copy the new next gen candidate as current node
        		memcpy(&current_node,&(next_gen[index]),sizeof(current_node));

//      		  current_node.pl_size = next_gen[index].pl_size;
//      		  current_node.pl_size_row = next_gen[index].pl_size_row;
//      		  current_node.pl_size_col = next_gen[index].pl_size_col;
//      		  current_node.sparcity_index = next_gen[index].sparcity_index;
//      		  current_node.actor_index = next_gen[index].actor_index;
//      		  strcpy(&(current_node.computed_fitness_score.counter1_name),&(next_gen[index].computed_fitness_score.counter1_name));
//      		  current_node.computed_fitness_score.counter1_score = next_gen[index].computed_fitness_score.counter1_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter2_name),&(next_gen[index].computed_fitness_score.counter2_name));
//      		  current_node.computed_fitness_score.counter2_score = next_gen[index].computed_fitness_score.counter2_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter3_name),&(next_gen[index].computed_fitness_score.counter3_name));
//      		  current_node.computed_fitness_score.counter3_score = next_gen[index].computed_fitness_score.counter3_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter4_name),&(next_gen[index].computed_fitness_score.counter4_name));
//        		  current_node.computed_fitness_score.counter4_score = next_gen[index].computed_fitness_score.counter4_score;

			    printf("MAIN$$ After generational change\n");
        		printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        		printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        		printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        		printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        		printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        		if (PMU_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                    }
                    sprintf(str,"\n$$$ NEXTGEN[%d] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index,current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                    current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
        		}
        		else if (HW_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                    }
                    sprintf(str,"\n$$$ NEXTGEN[%d] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index,current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                    current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
        		}
        		else
        		{
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        		}
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

        		sprintf(str,"%s","\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^\n");
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

        		fclose(report_fd);
        		fclose(shadowreport_fd);

//                printf("Hit any key to continue...\n");
//                getchar();

			}


		} /*End of if 0 == tracker[index] condition*/
    } /*End of while loop*/

//Just return success at this point wo we start with another set of random test candidates

// return search exhausted so the restart count can increment.
		return SEARCH_EXHAUSTED;
}

#else

/*******************************************************************************
 * static int evaluate_and_select_next_gen_nodes(void)                         *
 *    - This function evaluates the scores for next generation nodes and       *
 *      selects the next best candidate to progress the state space search     *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int evaluate_and_select_next_gen_nodes(FIT_FN_T fitness_fn)
{

    int i = 0;
	int index = 0;
	int tracker[NEXT_GEN_CANDIDATE_COUNT];
	int best_found = 0;
	long best_score = 0;
	char str[128];
	int node_index = -1;
	FILE* report_fd;
	FILE* shadowreport_fd;
	int ret_val = -1;
	int j = 0;
	long nextgen_score = 0;
	long current_score = 0;
	long bestnode_score = 0;

    REPORT_FN_DETAILS;

    clear_computed_scores(NEXTGEN_ONLY);

	for (i = 0; i < NEXT_GEN_CANDIDATE_COUNT; i++)
		tracker[i] = 0;

	i = 0;
	/*Here with single objective evaluation we take into account only
	  the first of the selected counters for evaluation*/
	if (PMU_MON == fitness_fn)
	{
    	best_score = current_node.pmu_computed_fitness_score.counter_score[0];
	}
	else if (HW_MON == fitness_fn)
	{
    	best_score = current_node.hwmon_computed_fitness_score.counter_score[0];
	}
	else
	{
        printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");

	}

	while (!best_found && i < NEXT_GEN_CANDIDATE_COUNT)
    {
    	index = rand() % NEXT_GEN_CANDIDATE_COUNT;
		if (0 == tracker[index])
		{
			tracker[index] = 1;
			i++;

            pl_size = next_gen[index].pl_size;
			printf("MAIN$$ INDEX is ----> %d\n",index);
            printf("MAIN$$ -----------------> PL size is set as %lld %lld\n",pl_size,next_gen[index].pl_size);

            actor_profile = next_gen[index].actor_index;

            individual_index = index;
            populate_node(index);

            for (stat_token_pmu = 0, stat_token_hwmon = 0; (stat_token_pmu < AVERAGE_STAT_COUNT) && (stat_token_hwmon < AVERAGE_STAT_COUNT); stat_token_pmu++, stat_token_hwmon++)
            {

        	    ret_val = reset_to_init_state();
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }

                /*This global variable indicates the processing of the generated perf report*/
//                dat_process_done = 0;

                ret_val = sbst_ccitest_iterator(fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart search space iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }

                /*We know for sure that perf userspace utility is configured to run for
                  perf_monotor_duration, So wait for a little more than that before
                  attempting to process the perf output*/
                sleep(perf_monitor_duration+2);

                /*index indicate the next gen candidate node*/
                ret_val = report_ccitest_fitness_function(index, fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - HillClimbing Random Restart search space iteration failed with %s -\n",error_map[ret_val].err_message);
                    return OPERATION_ERROR;
                }


            }/*End of for loop*/

			//Check if there is a best score than current?

			if (PMU_MON == fitness_fn)
			{
                nextgen_score = next_gen[index].pmu_computed_fitness_score.counter_score[0];
                current_score = current_node.pmu_computed_fitness_score.counter_score[0];
			}
			else if (HW_MON == fitness_fn)
			{
                nextgen_score = next_gen[index].hwmon_computed_fitness_score.counter_score[0];
                current_score = current_node.hwmon_computed_fitness_score.counter_score[0];
			}
			else
			{
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
			}

            /*Here with single objective evaluation we take into account only
	          the first of the selected counters for evaluation*/
//			if (next_gen[index].computed_fitness_score.counter_score[0] >= current_node.computed_fitness_score.counter_score[0])
			if (nextgen_score >= current_score)
			{
				best_found = 1;

			    printf("MAIN$$ SUCCESS Next best candidate found @ next_gen[%d] with score %ld -- size[%d,%d]\n ",index,nextgen_score,next_gen[index].pl_size_row,\
                next_gen[index].pl_size_col);

            	printf("MAIN$$ Before generational change\n");
        	    printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        		printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        		printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        		printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        		printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        		if (PMU_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                    }
        		}
        		else if (HW_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                    }
        		}
        		else
        		{
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        		}

        		report_fd = fopen(report_singleobj_name,"a+");
        		if (NULL == report_fd)
        		{
                    printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
                	return IO_ERROR;
        		}

        		shadowreport_fd = fopen(report_shadowobj_name,"a+");
        		if (NULL == shadowreport_fd)
        		{
                    printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
                	return IO_ERROR;
        		}

        		sprintf(str,"%s","\n\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^");
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

                if (PMU_MON == fitness_fn)
                {
                    sprintf(str,"\n$$$ Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                        current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
                }
                else if (HW_MON == fitness_fn)
                {
                    sprintf(str,"\n$$$ Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                        current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
                }
                else
                {
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
                }

        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

               	// Copy the new next gen candidate as current node
        		memcpy(&current_node,&(next_gen[index]),sizeof(current_node));

//      		  current_node.pl_size = next_gen[index].pl_size;
//      		  current_node.pl_size_row = next_gen[index].pl_size_row;
//      		  current_node.pl_size_col = next_gen[index].pl_size_col;
//      		  current_node.sparcity_index = next_gen[index].sparcity_index;
//      		  current_node.actor_index = next_gen[index].actor_index;
//      		  strcpy(&(current_node.computed_fitness_score.counter1_name),&(next_gen[index].computed_fitness_score.counter1_name));
//      		  current_node.computed_fitness_score.counter1_score = next_gen[index].computed_fitness_score.counter1_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter2_name),&(next_gen[index].computed_fitness_score.counter2_name));
//      		  current_node.computed_fitness_score.counter2_score = next_gen[index].computed_fitness_score.counter2_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter3_name),&(next_gen[index].computed_fitness_score.counter3_name));
//      		  current_node.computed_fitness_score.counter3_score = next_gen[index].computed_fitness_score.counter3_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter4_name),&(next_gen[index].computed_fitness_score.counter4_name));
//        		  current_node.computed_fitness_score.counter4_score = next_gen[index].computed_fitness_score.counter4_score;

			    printf("MAIN$$ After generational change\n");
        		printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        		printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        		printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        		printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        		printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        		if (PMU_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                    }
                    sprintf(str,"\n$$$ NEXTGEN[%d] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index,current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                    current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
        		}
        		else if (HW_MON == fitness_fn)
        		{
                    for (j = 0; j < perf_max_counters; j++)
                    {
                        printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                    }
                    sprintf(str,"\n$$$ NEXTGEN[%d] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index,current_node.pl_size_row,current_node.pl_size_col,current_node.sparcity_index,current_node.actor_index,\
                                    current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);

        		}
        		else
        		{
        		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        		    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        		}
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

        		sprintf(str,"%s","\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^\n");
        		fputs(str,report_fd);
        		fputs(str,shadowreport_fd);

        		fclose(report_fd);;
        		fclose(shadowreport_fd);;

//                printf("Hit any key to continue...\n");
//                getchar();

			}


		} /*End of if 0 == tracker[index] condition*/
    } /*End of while loop*/

	if (0 == best_found)
	{
	    if (PMU_MON == fitness_fn)
	    {
                bestnode_score = best_node.pmu_computed_fitness_score.counter_score[0];
                current_score = current_node.pmu_computed_fitness_score.counter_score[0];
	    }
	    else if (HW_MON == fitness_fn)
	    {
                bestnode_score = best_node.hwmon_computed_fitness_score.counter_score[0];
                current_score = current_node.pmu_computed_fitness_score.counter_score[0];
	    }
	    else
	    {
		    printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
    	    sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
	    }

		if (current_score >= bestnode_score)
		{
			printf("MAIN$$ New global best found; Hence save it\n");
			// Copy the new next gen candidate as the new best node
			memcpy(&best_node,&current_node,sizeof(current_node));

            report_fd = fopen(report_singleobj_name,"a+");
            if (NULL == report_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
                return IO_ERROR;
            }

            shadowreport_fd = fopen(report_shadowobj_name,"a+");
            if (NULL == shadowreport_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
                return IO_ERROR;
            }

            sprintf(str,"%s","\n\n^^^^^^^^^%%%%%%%%%%%%%%%%%%^^^^^^^^^^");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            if (PMU_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ [NEW] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                    best_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.pmu_computed_fitness_score.counter_score[0]);
            }
            else if (HW_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ [NEW] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                    best_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.hwmon_computed_fitness_score.counter_score[0]);
            }
            else
            {
                printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
                sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
            }

            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            sprintf(str,"%s","\n^^^^^^^^^%%%%%%%%%%%%%%%%%%%^^^^^^^^^^\n");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            fclose(report_fd);
            fclose(shadowreport_fd);
		}
		else
		{
			report_fd = fopen(report_singleobj_name,"a+");
            if (NULL == report_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
                return IO_ERROR;
            }

			shadowreport_fd = fopen(report_shadowobj_name,"a+");
            if (NULL == shadowreport_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
                return IO_ERROR;
            }

            sprintf(str,"%s","\n\n^^^^^^^^^%%%%%%%%%%%%%%%%%%^^^^^^^^^^");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            if (PMU_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ [UNCHANGED] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                    best_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.pmu_computed_fitness_score.counter_score[0]);
            }
            else if (HW_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ [UNCHANGED] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                    best_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.hwmon_computed_fitness_score.counter_score[0]);
            }
            else
            {
                printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
                sprintf(str,"MAIN$$ Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
            }

            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            sprintf(str,"%s","\n^^^^^^^^^%%%%%%%%%%%%%%%%%%%^^^^^^^^^^\n");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            fclose(report_fd);
            fclose(shadowreport_fd);

		}

		return SEARCH_EXHAUSTED; /*Returning -1 indicate experiment cannot proceed, Exhausted exploration*/
	}
	else
		return SUCCESS;
}



#endif

/*******************************************************************************
 * int iterate_hc_rdm_restart_statespace(PL_T payload)                           *
 *    - This function blah blah                                                *
 ******************************************************************************/
int iterate_hc_rdm_restart_statespace(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;
	int finished = 0;
    int restart_count = 0;

	REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Iterate CCI Testing statespace with Random read write payload\n");

                hc_seeding_firstflag = FALSE;
		        while (!finished)
		        {
			        ret_val = generate_next_nodes();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - HillClimbing Random Restart iteration failed to generate next gen nodes with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }
                    /*-1 indicates exploration exhausted from the current set of neighbourhood
                      attempt restarting with a new random seed*/
			        if (SEARCH_EXHAUSTED == evaluate_and_select_next_gen_nodes(fitness_fn))
			        {
				        restart_count++;

//				        Do random restart here
//				        if (-1 == select_next_gen())
                        ret_val = generate_initial_seed(payload, hc_seeding_firstflag, fitness_fn);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - HillClimbing Random Restart iteration failed with %s -\n",error_map[ret_val].err_message);
                            return OPERATION_ERROR;
                        }

                        printf("Hit any key to continue...[Search exhaused in immediate neighbourhood with restart count %d current time {%lld} finish time {%lld} ]\n", restart_count, (long long)time(NULL), (long long)finish_time);
                        getchar();


			         }

			        if ((restart_count > hc_restart_count) || (time(NULL) > finish_time))
			        {
				        finished = 1;
			        }
                }

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}