/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/sbst_sa.h"

/*Global variables*/
int sa_temp_iterations = 0;
int sa_seeding_firstflag = TRUE;
int sa_max_subopt_count = 0;


/*******************************************************************************
 * int initialise_sa_setup(PL_T payload)                                       *
 *    - This function blah blah                                                *
 ******************************************************************************/
int initialise_sa_setup(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                ret_val = generate_initial_seed(payload, sa_seeding_firstflag, fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - Simulated Annealing setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                    return OPERATION_ERROR;
                }
                sa_seeding_firstflag = FALSE;

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}

/*******************************************************************************
 * static double compute_acceptance_probability(double oldcost, double newcost,*
 *                                       int higher_the_better, float temp)    *
 *    - This function blah blah                                                *
 ******************************************************************************/
static float compute_acceptance_probability(float oldcost, float newcost, int higher_the_better, float temp)
{

    float cost_diff = 0.0;
    float ap = 0.0;

    if (1 == higher_the_better)
    {
        cost_diff = (newcost/newcost) - (oldcost/newcost);
    }
    else
    {
        cost_diff = (oldcost/oldcost) - (newcost/oldcost);
    }

    ap = pow(2.71828, cost_diff/temp);

    printf("MAIN$$ Computing ap[%f] as e^(%f = diff(newcost[%f],oldcost[%f]))/%f with higher_the_better set to %d\n",ap, cost_diff/temp, newcost, oldcost, temp, higher_the_better);

    return ap;
}


/*******************************************************************************
 * static int evaluate_and_select_next_gen_nodes(int higher_the_better,        *
 *                                                float temperature)           *
 *    - This function evaluates the scores for next generation nodes and       *
 *      selects the next best candidate to progress the state space search     *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int evaluate_and_select_next_gen_nodes(int higher_the_better, float temperature, FIT_FN_T fitness_fn)
{

    int i = 0;
	int index = 0;
	int best_found = 0;
	char str[164];
	int node_index = -1;
	FILE* report_fd;
	FILE* shadowreport_fd;
	int ret_val = -1;
	int j = 0;
	float ap = 0.0;
	float random_index = 0.0;
	long flooring_score = 0;
	long nextgen_score = 0;
	long current_score = 0;
	long bestnode_score = 0;
	int return_search_exhausted = 0;
	static int num_subopt_in_temp = 0;


    REPORT_FN_DETAILS;

    clear_computed_scores(NEXTGEN_ONLY);

	/*Here with single objective evaluation we take into account only
	  the first of the selected counters for evaluation*/

    index = rand() % NEXT_GEN_CANDIDATE_COUNT;

    pl_size = next_gen[index].pl_size;
    printf("MAIN$$ INDEX is ----> %d\n",index);
    printf("MAIN$$ -----------------> PL size is set as %lld %lld\n",pl_size,next_gen[index].pl_size);

    actor_profile = next_gen[index].actor_index;

    individual_index = index;
    populate_node(index);

//    for (stat_token = 0; stat_token < AVERAGE_STAT_COUNT; stat_token++)
    for (stat_token_pmu = 0, stat_token_hwmon = 0; (stat_token_pmu < AVERAGE_STAT_COUNT) && (stat_token_hwmon < AVERAGE_STAT_COUNT); stat_token_pmu++, stat_token_hwmon++)
    {

        ret_val = reset_to_init_state();
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - Simulated Annealing iteration failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }

        /*This global variable indicates the processing of the generated perf report*/
//                dat_process_done = 0;

        ret_val = sbst_ccitest_iterator(fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - Simulated Annealing search space iteration failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }

        /*We know for sure that perf userspace utility is configured to run for
          perf_monotor_duration, So wait for a little more than that before
          attempting to process the perf output*/
        sleep(perf_monitor_duration+2);

        /*index indicate the next gen candidate node*/
        ret_val = report_ccitest_fitness_function(index, fitness_fn);
        if (SUCCESS != ret_val)
        {
            printf("MAIN$$ ERROR - Simulated Annealing search space iteration failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }


    }/*End of for loop*/

    //Check if there is a best score than current?

    if (PMU_MON == fitness_fn)
    {
        nextgen_score = next_gen[index].pmu_computed_fitness_score.counter_score[0];
        current_score = current_node.pmu_computed_fitness_score.counter_score[0];
    }
    else if (HW_MON == fitness_fn)
    {
        nextgen_score = next_gen[index].hwmon_computed_fitness_score.counter_score[0];
        current_score = current_node.hwmon_computed_fitness_score.counter_score[0];
    }
    else
    {
        printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
    }
    /*Here with single objective evaluation we take into account only
      the first of the selected counters for evaluation*/
    if (nextgen_score >= current_score)
    {

        /*Here we definitely go along with the new candidate*/
        best_found = 1;

        printf("MAIN$$ SUCCESS [@temperature %f] Next best candidate found @ next_gen[%d] with score %ld -- size[%d,%d]\n ", temperature, index,nextgen_score,next_gen[index].pl_size_row,\
        next_gen[index].pl_size_col);


//        printf("Hit any key to continue...\n");
//        getchar();

        printf("MAIN$$ Before generational change\n");
        printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        if (PMU_MON == fitness_fn)
        {
            for (j = 0; j < perf_max_counters; j++)
            {
                printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
            }
        }
        else if (HW_MON == fitness_fn)
        {
            for (j = 0; j < perf_max_counters; j++)
            {
                printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
            }
        }
        else
        {
        }

        report_fd = fopen(report_singleobj_name,"a+");
        if (NULL == report_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
            return IO_ERROR;
        }

        shadowreport_fd = fopen(report_shadowobj_name,"a+");
        if (NULL == shadowreport_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
            return IO_ERROR;
        }

        sprintf(str,"%s","\n\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        if (PMU_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ Current_node[%d][%d] [@temperature %f] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row, current_node.pl_size_col, temperature, current_node.sparcity_index, current_node.actor_index,\
                                current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
        }
        else if (HW_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ Current_node[%d][%d] [@temperature %f] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row, current_node.pl_size_col, temperature, current_node.sparcity_index, current_node.actor_index,\
                                current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
            sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        }

        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        // Copy the new next gen candidate as current node
        memcpy(&current_node,&(next_gen[index]),sizeof(current_node));

//      		  current_node.pl_size = next_gen[index].pl_size;
//      		  current_node.pl_size_row = next_gen[index].pl_size_row;
//      		  current_node.pl_size_col = next_gen[index].pl_size_col;
//      		  current_node.sparcity_index = next_gen[index].sparcity_index;
//      		  current_node.actor_index = next_gen[index].actor_index;
//      		  strcpy(&(current_node.computed_fitness_score.counter1_name),&(next_gen[index].computed_fitness_score.counter1_name));
//      		  current_node.computed_fitness_score.counter1_score = next_gen[index].computed_fitness_score.counter1_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter2_name),&(next_gen[index].computed_fitness_score.counter2_name));
//      		  current_node.computed_fitness_score.counter2_score = next_gen[index].computed_fitness_score.counter2_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter3_name),&(next_gen[index].computed_fitness_score.counter3_name));
//      		  current_node.computed_fitness_score.counter3_score = next_gen[index].computed_fitness_score.counter3_score;
//      		  strcpy(&(current_node.computed_fitness_score.counter4_name),&(next_gen[index].computed_fitness_score.counter4_name));
//        		  current_node.computed_fitness_score.counter4_score = next_gen[index].computed_fitness_score.counter4_score;

        printf("MAIN$$ After generational change\n");
        printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
        printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
        printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
        printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
        printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

        if (PMU_MON == fitness_fn)
        {
            for (j = 0; j < perf_max_counters; j++)
            {
                printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
            }
            sprintf(str,"\n$$$ NEXTGEN[%d] [@temperature %f] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index, temperature, current_node.pl_size_row, current_node.pl_size_col, current_node.sparcity_index, current_node.actor_index,\
                            current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
        }
        else if (HW_MON == fitness_fn)
        {
            for (j = 0; j < perf_max_counters; j++)
            {
                printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
            }
            sprintf(str,"\n$$$ NEXTGEN[%d] [@temperature %f] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",index, temperature, current_node.pl_size_row, current_node.pl_size_col, current_node.sparcity_index, current_node.actor_index,\
                            current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
            sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        }

        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        sprintf(str,"%s","\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^\n");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        fclose(report_fd);
        fclose(shadowreport_fd);

//                printf("Hit any key to continue...\n");
//                getchar();

    }
    else
    {
        /*Here we compute the acceptance criteria for the sub-optimal candidate and decide whether to proceed
          with this or not*/

        if (PMU_MON == fitness_fn)
        {
            current_score = current_node.pmu_computed_fitness_score.counter_score[0];
            nextgen_score = next_gen[index].pmu_computed_fitness_score.counter_score[0];
        }
        else if (HW_MON == fitness_fn)
        {
            current_score = current_node.hwmon_computed_fitness_score.counter_score[0];
            nextgen_score = next_gen[index].hwmon_computed_fitness_score.counter_score[0];
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        }

        ap = compute_acceptance_probability(current_score, nextgen_score, higher_the_better, temperature);

        random_index =  (float)rand() / (float)(RAND_MAX);
        printf("MAIN$$ random index generated %lf is compared against computed ap %lf\n",random_index, ap);

//        printf("Hit any key to continue...\n");
//        getchar();

        /*Defined the flooring_score based on the current best_score recorded*/
        if (PMU_MON == fitness_fn)
        {
            flooring_score = best_node.pmu_computed_fitness_score.counter_score[0] * 0.9;
        }
        else if (HW_MON == fitness_fn)
        {
            flooring_score = best_node.hwmon_computed_fitness_score.counter_score[0] * 0.9;
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
        }

        if ((ap > random_index) && (nextgen_score >= flooring_score) && (num_subopt_in_temp < sa_max_subopt_count))
        {
            /*Here we decide to move with the sub optimal solution if the acceptance probability is
              better than the random index generated and the next_gen score is atleast 90% of the global best*/

            printf("MAIN$$ SUCCESS [@temperature %f] Next [SUB OPTIMAL {%d}] candidate found @ next_gen[%d] with score %ld -- size[%d,%d]\n ", temperature, num_subopt_in_temp, index, nextgen_score, next_gen[index].pl_size_row,\
            next_gen[index].pl_size_col);

            /*Increment the count tracking the suboptimal candidate selection in a given temperature*/
            num_subopt_in_temp++;

//            printf("Hit any key to continue...\n");
//            getchar();

            printf("MAIN$$ Before generational change\n");
            printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
            printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
            printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
            printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
            printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

            if (PMU_MON == fitness_fn)
            {
                for (j = 0; j < perf_max_counters; j++)
                {
                    printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                }
            }
            else if (HW_MON == fitness_fn)
            {
                for (j = 0; j < perf_max_counters; j++)
                {
                    printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                }
            }
            else
            {
                printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
            }

            report_fd = fopen(report_singleobj_name,"a+");
            if (NULL == report_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
                return IO_ERROR;
            }

            shadowreport_fd = fopen(report_shadowobj_name,"a+");
            if (NULL == shadowreport_fd)
            {
                printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
                return IO_ERROR;
            }

            sprintf(str,"%s","\n\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            if (PMU_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ Current_node[%d][%d] [@temperature %f] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row, current_node.pl_size_col, temperature, current_node.sparcity_index, current_node.actor_index,\
                                    current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
            }
            else if (HW_MON == fitness_fn)
            {
                sprintf(str,"\n$$$ Current_node[%d][%d] [@temperature %f] SI = %d AI = %d Score[%s] = %ld",current_node.pl_size_row, current_node.pl_size_col, temperature, current_node.sparcity_index, current_node.actor_index,\
                                    current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
            }
            else
            {
                printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
                sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
            }
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            // Copy the new next gen candidate as current node
            memcpy(&current_node,&(next_gen[index]),sizeof(current_node));

    //      		  current_node.pl_size = next_gen[index].pl_size;
    //      		  current_node.pl_size_row = next_gen[index].pl_size_row;
    //      		  current_node.pl_size_col = next_gen[index].pl_size_col;
    //      		  current_node.sparcity_index = next_gen[index].sparcity_index;
    //      		  current_node.actor_index = next_gen[index].actor_index;
    //      		  strcpy(&(current_node.computed_fitness_score.counter1_name),&(next_gen[index].computed_fitness_score.counter1_name));
    //      		  current_node.computed_fitness_score.counter1_score = next_gen[index].computed_fitness_score.counter1_score;
    //      		  strcpy(&(current_node.computed_fitness_score.counter2_name),&(next_gen[index].computed_fitness_score.counter2_name));
    //      		  current_node.computed_fitness_score.counter2_score = next_gen[index].computed_fitness_score.counter2_score;
    //      		  strcpy(&(current_node.computed_fitness_score.counter3_name),&(next_gen[index].computed_fitness_score.counter3_name));
    //      		  current_node.computed_fitness_score.counter3_score = next_gen[index].computed_fitness_score.counter3_score;
    //      		  strcpy(&(current_node.computed_fitness_score.counter4_name),&(next_gen[index].computed_fitness_score.counter4_name));
    //        		  current_node.computed_fitness_score.counter4_score = next_gen[index].computed_fitness_score.counter4_score;

            printf("MAIN$$ After generational change\n");
            printf("MAIN$$ current_node.pl_size = %lld\n",current_node.pl_size);
            printf("MAIN$$ current_node.pl_size_row = %d\n",current_node.pl_size_row);
            printf("MAIN$$ current_node.pl_size_col = %d\n",current_node.pl_size_col);
            printf("MAIN$$ current_node.sparcity_index = %d\n",current_node.sparcity_index);
            printf("MAIN$$ current_node.pl_actor_index = %d\n",current_node.actor_index);

            if (PMU_MON == fitness_fn)
            {
                for (j = 0; j < perf_max_counters; j++)
                {
                    printf("MAIN$$ current_node.pmu_computed_fitness_score[%s] = %ld\n",current_node.pmu_computed_fitness_score.counter_name[j],current_node.pmu_computed_fitness_score.counter_score[j]);
                }

                sprintf(str,"\n$$$ NEXTGEN[%d] [SUB OPTIMAL {%d}] [@temperature %f] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld", index, num_subopt_in_temp, temperature, current_node.pl_size_row, current_node.pl_size_col, current_node.sparcity_index, current_node.actor_index ,\
                                current_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.pmu_computed_fitness_score.counter_score[0]);
            }
            else if (HW_MON == fitness_fn)
            {
                for (j = 0; j < perf_max_counters; j++)
                {
                    printf("MAIN$$ current_node.hwmon_computed_fitness_score[%s] = %ld\n",current_node.hwmon_computed_fitness_score.counter_name[j],current_node.hwmon_computed_fitness_score.counter_score[j]);
                }

                sprintf(str,"\n$$$ NEXTGEN[%d] [SUB OPTIMAL {%d}] [@temperature %f] Current_node[%d][%d] SI = %d AI = %d Score[%s] = %ld", index, num_subopt_in_temp, temperature, current_node.pl_size_row, current_node.pl_size_col, current_node.sparcity_index, current_node.actor_index ,\
                                current_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,current_node.hwmon_computed_fitness_score.counter_score[0]);
            }
            else
            {
                printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
                sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
            }
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            sprintf(str,"%s","\n^^^^^^^^^$$$$$$$$$$$$^^^^^^^^^^\n");
            fputs(str,report_fd);
            fputs(str,shadowreport_fd);

            fclose(report_fd);;
            fclose(shadowreport_fd);;

//                printf("Hit any key to continue...\n");
//                getchar();
        }
        else
        {
//            printf("Hit any key to continue as we are exhausting the search...\n");
//            getchar();

            return_search_exhausted = 1;
            /*As we will reduce the temperature in the next re-entry of this function
              reset the tracker for number of sub optimal candidates selected in a
              given temperature*/
            num_subopt_in_temp = 0;

        }

    }

/*  Check if I have the global best in which case save it and continue
*/
    if (PMU_MON == fitness_fn)
    {
        current_score = current_node.pmu_computed_fitness_score.counter_score[0];
        bestnode_score = best_node.pmu_computed_fitness_score.counter_score[0];
    }
    else if (HW_MON == fitness_fn)
    {
        current_score = current_node.hwmon_computed_fitness_score.counter_score[0];
        bestnode_score = best_node.hwmon_computed_fitness_score.counter_score[0];
    }
    else
    {
        printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
    }

    if (current_score >= bestnode_score)
    {
        printf("MAIN$$ New global best found; Hence save it\n");
        // Copy the new next gen candidate as the new best node
        memcpy(&best_node,&current_node,sizeof(current_node));

        report_fd = fopen(report_singleobj_name,"a+");
        if (NULL == report_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
            return IO_ERROR;
        }

        shadowreport_fd = fopen(report_shadowobj_name,"a+");
        if (NULL == shadowreport_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
            return IO_ERROR;
        }

        sprintf(str,"%s","\n\n^^^^^^^^^%%%%%%%%%%%%%%%%%%^^^^^^^^^^");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        if (PMU_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ [NEW] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                best_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.pmu_computed_fitness_score.counter_score[0]);
        }
        else if (HW_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ [NEW] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                best_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.hwmon_computed_fitness_score.counter_score[0]);
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
            sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        }

        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        sprintf(str,"%s","\n^^^^^^^^^%%%%%%%%%%%%%%%%%%%^^^^^^^^^^\n");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        fclose(report_fd);
        fclose(shadowreport_fd);
    }
    else
    {
        report_fd = fopen(report_singleobj_name,"a+");
        if (NULL == report_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_singleobj_name);
            return IO_ERROR;
        }

        shadowreport_fd = fopen(report_shadowobj_name,"a+");
        if (NULL == shadowreport_fd)
        {
            printf("MAIN$$ File open error for %s, Terminating...\n", report_shadowobj_name);
            return IO_ERROR;
        }

        sprintf(str,"%s","\n\n^^^^^^^^^%%%%%%%%%%%%%%%%%%^^^^^^^^^^");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        if (PMU_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ [UNCHANGED] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                best_node.pmu_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.pmu_computed_fitness_score.counter_score[0]);
        }
        else if (HW_MON == fitness_fn)
        {
            sprintf(str,"\n$$$ [UNCHANGED] GLOBAL BEST best_node[%d][%d] SI = %d AI = %d Score[%s] = %ld",best_node.pl_size_row,best_node.pl_size_col,best_node.sparcity_index,best_node.actor_index,\
                best_node.hwmon_computed_fitness_score.counter_name+0*COUNTER_NAME_STR_LEN,best_node.hwmon_computed_fitness_score.counter_score[0]);
        }
        else
        {
            printf("MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!\n");
            sprintf(str,"MAIL$$ ERROR:: Incorrect fitness function selection used to evaluate in sbst algorithm!!!");
        }

        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        sprintf(str,"%s","\n^^^^^^^^^%%%%%%%%%%%%%%%%%%%^^^^^^^^^^\n");
        fputs(str,report_fd);
        fputs(str,shadowreport_fd);

        fclose(report_fd);
        fclose(shadowreport_fd);

    }

    if (1 == return_search_exhausted)
    {
//        printf("Hit any key to continue as we are exhausting the search...\n");
//        getchar();


        return SEARCH_EXHAUSTED; /*Returning -1 indicate experiment cannot proceed, Exhausted exploration*/
    }


    return SUCCESS;
}

/*******************************************************************************
 * int iterate_sa_statespace(PL_T payload)                                     *
 *    - This function blah blah                                                *
 ******************************************************************************/
int iterate_sa_statespace(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;
	float temp_start = 1.0;
	float temp_min = 0.1;
	float alpha = 0.8;
	int temp_iter = 0;

    printf("MAIN$$ Iterate CCI Testing statespace with Random read write payload\n");

	REPORT_FN_DETAILS;
    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Iterate CCI Testing statespace with Random read write payload\n");

                sa_seeding_firstflag = FALSE;
            	while ((temp_start > temp_min) && (time(NULL) < finish_time))
		        {
		            temp_iter = 0;

		            while ((temp_iter < sa_temp_iterations) && (time(NULL) < finish_time))
		            {
                        ret_val = generate_next_nodes();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Simulated Annealing iteration failed to generate next gen nodes with %s -\n",error_map[ret_val].err_message);
                            return OPERATION_ERROR;
                        }

                        /*-1 indicates exploration exhausted from the current set of neighbourhood
                          attempt restarting with a new random seed*/
                        if (SEARCH_EXHAUSTED == evaluate_and_select_next_gen_nodes(TRUE, temp_start, fitness_fn))
                        {
                            /*Continue exploration with the same temperature by attempting with a new neighbour from current*/
                            temp_iter++;
                            printf("MAIN$$ Search exhausted repeating temp iterations at %d\n",temp_iter);

//                            printf("Hit any key to continue...[Search exhauseted iterating at temp %f for %d th time] with current time {%lld} finish time {%lld}\n",temp_start, temp_iter, (long long)time(NULL), (long long)finish_time);
//                            getchar();

                        }

		            }

		            /*Cool the temperature by alpha and continue search*/
		            temp_start*=alpha;

                    printf("Hit any key to continue...[Cooling the temperature to %f]\n", temp_start);
//                    getchar();

                }

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }
    return SUCCESS;
}