/********************************************************************************
 *  Copyright 2018-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

 #include "include/sbst_spea2.h"

/*Global variables*/
int spea2_seeding_firstflag = TRUE;
int generation_index = 0;
/*+1 so we can use a special delimiter of -100 to the ordered list when full
    elements are in play*/
static int fitness_ordered_index[SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE + 1];
static int non_dominates_index[SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE + 1];
static int distance_ordered_index[SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE + 1];

/*******************************************************************************
 * static int check_for_crossover(void)                                        *
 *    - This function generates crossover action to be done                    *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int check_for_crossover()
{

    int min = 1;
    int max = 100;
    int random;

//        REPORT_FN_DETAILS;

    random = (rand() % (max - min + 1) ) + min;

    if (random > (int)SPEA2_CROSSOVER_RATE*100)
    {
        /*Here we return crossover to be denied*/
        random = 0;
    }
    else
    {
        /*This returns 1 for crossover to happen*/
        random = 1;
    }

//	printf("random index = %d",rand_index);
    return (random);
}

/*******************************************************************************
 * static int check_for_mutation(void)                                         *
 *    - This function generates mutation action to be done                     *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int check_for_mutation()
{

    int min = 1;
    int max = 100;
    int random;

//        REPORT_FN_DETAILS;

    random = (rand() % (max - min + 1) ) + min;

    if (random > (int)SPEA2_MUTATION_RATE*100)
    {
        /*Here we return mutation to be denied*/
        random = 0;
    }
    else
    {
        /*This returns 1 for mutation in Si, 2 for mutation in Sp and 3 for
            mutation in Ac*/
        random = (rand() % (3) ) + 1;
    }

//	printf("random index = %d",rand_index);
    return (random);
}

/*******************************************************************************
 * static int generate_rand_crossover_point(void)                              *
 *    - This function generates the a random cross over point                  *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int generate_rand_crossover_point()
{

    int index_min = 0;
    int index_max = 1;
    int rand_index;

//        REPORT_FN_DETAILS;

    rand_index = (rand() % (index_max - index_min + 1) ) + index_min;

//	printf("random index = %d",rand_index);
    return (rand_index);
}

/*******************************************************************************
 * static int generate_rand_matingpool_index(void)                             *
 *    - This function generates the a random index within the mating pool      *
 *      population                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int generate_rand_matingpool_index()
{

    int index_min = 0;
    int index_max = SPEA2_MATINGPOOL_SIZE-1;
    int rand_index;

//        REPORT_FN_DETAILS;

    rand_index = (rand() % (index_max - index_min + 1) ) + index_min;

//	printf("random index = %d",rand_index);
    return (rand_index);
}


/*******************************************************************************
 * static int generate_rand_archive_index(void)                                *
 *    - This function generates the a random index within the archive          *
 *      population                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int generate_rand_archive_index()
{

    int index_min = SPEA2_POPULATION_SIZE;
    int index_max = SPEA2_POPULATION_SIZE+SPEA2_ARCHIVE_SIZE-1;
    int rand_index;

//        REPORT_FN_DETAILS;

    rand_index = (rand() % (index_max - index_min + 1) ) + index_min;

//	printf("random index = %d",rand_index);
    return (rand_index);
}

/*******************************************************************************
 * static int write_population_details(void)                                   *
 *    - This function write population details to report file                  *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int write_population_details()
{

    int i, j;
    FILE* report_spea2_fd;
    char str[PERF_OUT_STR_LEN];
    char timestamp[128];
    time_t current_time;


    REPORT_FN_DETAILS;

	report_spea2_fd = fopen("report-spea2.txt","a+");
	if (NULL == report_spea2_fd)
	{
		printf("MAIN$$ File open error for %s, Terminating...\n","report-spea2.txt");
		return IO_ERROR;
	}

    sprintf(str,">>>>>>Generation[%d]<<<<<<<\n",generation_index);
	fputs(str,report_spea2_fd);

    current_time = time(NULL);
    sprintf(timestamp,"# started %s\n\n",asctime(localtime(&current_time)));
	fputs(timestamp,report_spea2_fd);

    sprintf(str,"\n[%s]\n","Population Members");
	fputs(str,report_spea2_fd);

//        printf("Before writing anything from pop\n");
//        getchar();

	for (i = 0; i < SPEA2_POPULATION_SIZE; i++)
	{
        if (1 == pop[i].active)
        {
//        printf("Before writing anything from pop[%d]\n",i);
//        getchar();
            sprintf(str,"\npop[%d] --> {pmu[%ld],hwmon[%ld]} --> Fitness[%ld] --> Density[%lf] \n", i, \
                    pop[i].pmu_computed_fitness_score.counter_score[0], \
                    pop[i].hwmon_computed_fitness_score.counter_score[0], \
                    pop[i].fitness, pop[i].density);
            fputs(str,report_spea2_fd);

//        printf("Before writing distance numbers\n");
//        getchar();
            /*Write the computed distance information*/
            for (j = 0; j < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); j++)
            {
                if (1 == pop[j].active)
                {
                    sprintf(str,"{distance[%d,%d]=%lf},  ", i, j, pop[i].distance[j]);
                    fputs(str,report_spea2_fd);
                }
            }
//        printf("After writing distance numbers\n");
//        getchar();

        }

	}

    sprintf(str,"\n\n[%s]\n","Archive Population Members");
	fputs(str,report_spea2_fd);

	for (i = SPEA2_POPULATION_SIZE; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
        if (1 == pop[i].active)
        {
            sprintf(str,"\narchive_pop[%d] --> {pmu[%ld],hwmon[%ld]} --> Fitness[%ld] --> Density[%lf] \n", i, \
                    pop[i].pmu_computed_fitness_score.counter_score[0], \
                    pop[i].hwmon_computed_fitness_score.counter_score[0], \
                    pop[i].fitness, pop[i].density);
            fputs(str,report_spea2_fd);

//        printf("Before writing distance numbers\n");
//        getchar();

            /*Write the computed distance information*/
            for (j = 0; j < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); j++)
            {
                if (1 == pop[j].active)
                {
                    sprintf(str,"{distance[%d,%d]=%lf},  ", i, j, pop[i].distance[j]);
                    fputs(str,report_spea2_fd);
                }
            }

//        printf("After writing distance numbers\n");
//        getchar();


        }

	}

    sprintf(str,"\n%s\n","~~~~~~~~~~~~~~~~~~~~~~~~");
	fputs(str,report_spea2_fd);

	fclose(report_spea2_fd);

    return SUCCESS;
}

/*******************************************************************************
 * static int a_dominates_b(int a_index, int b_index)                          *
 *    - This function checks whether a dominates b or not based multiple       *
 *      objectives                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int a_dominates_b(int a_index, int b_index)
{

    int i;
    int a_is_worse = 0;
    int both_equal = 1;
    long a_score, b_score;

//    REPORT_FN_DETAILS;

    /*Check if a dominates b and this happens only if one of the below conditions
        is satisfied:
        1. fitness_fn1 and fitness_fn2 of a is greater than b
        2. one of the fitness_fn is equal in both a and b and for the other a is
            greater than b
    */
	for (i = 0; (i < SPEA2_FITNESS_DIMENSION) && !a_is_worse; i++)
	{
	    if (0 == i)
	    {
	        a_score = pop[a_index].pmu_computed_fitness_score.counter_score[0];
	        b_score = pop[b_index].pmu_computed_fitness_score.counter_score[0];
	    }
	    else
	    {
	        a_score = pop[a_index].hwmon_computed_fitness_score.counter_score[0];
	        b_score = pop[b_index].hwmon_computed_fitness_score.counter_score[0];
	    }
	    /*Higher the fitness value is better*/
        a_is_worse = a_score < b_score;
        both_equal = (a_score == b_score) && both_equal;

	}

    return (!both_equal && !a_is_worse);
}


/*******************************************************************************
 * static int do_population_fitness_assignment(void)*
 *    - This function performs the fitness assignment for the candidates  based*
 *      on spea2                                                              *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int do_population_fitness_assignment()
{

    int candidate_strength[SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE];
    long candidate_fitness = 0;
    int i, j;

    REPORT_FN_DETAILS;

    /*First compute the strength of each active candidate in the population
        based on how many other solutions it dominates*/
	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
	    if (1 == pop[i].active)
	    {
	        candidate_strength[i] = 0;

            for (j = 0; j < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); j++)
            {
                if (1 == pop[j].active)
                {
                    if (TRUE == a_dominates_b(i,j))
                    {
                        candidate_strength[i]++;
                    }
                }

            }

	    }
	}


	/*Now assign fitness as the sum of strength values of its dominators*/

	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
	    if (1 == pop[i].active)
	    {
	        candidate_fitness = 0;

            for (j = 0; j < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); j++)
            {
                if (1 == pop[j].active)
                {
                    if (TRUE == a_dominates_b(j,i))
                    {
                        candidate_fitness += candidate_strength[j];
                    }
                }

            }

            pop[i].fitness = candidate_fitness;

	    }



	}


    return SUCCESS;
}

/*******************************************************************************
 * static int a_isequalto_b(int a_index, int b_index)                          *
 *    - This function computes the whether a and b candidates are equal or not *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int a_isequalto_b(int a_index, int b_index)
{

    int i;
    int both_equal = 1;

//    REPORT_FN_DETAILS;

	for (i = 0; i < SPEA2_FITNESS_DIMENSION; i++)
	{
	    if (0 == i)
	    {
	        both_equal = (pop[a_index].pmu_computed_fitness_score.counter_score[0] == \
	                    pop[b_index].pmu_computed_fitness_score.counter_score[0]) && \
	                    both_equal;
	    }
	    else
	    {
	        both_equal = (pop[a_index].hwmon_computed_fitness_score.counter_score[0] == \
	                    pop[b_index].hwmon_computed_fitness_score.counter_score[0]) && \
	                    both_equal;
	    }

	}

    return (both_equal);
}

/*******************************************************************************
 * static int sort_candidate_index_on_distance(int candidate_index)            *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int sort_candidate_index_on_distance(int candidate_index)
{

    int i, j, temp;
    int index = 0;

    REPORT_FN_DETAILS;

    /*First extract valid indices to the distance index array*/
	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
        if (1 == pop[i].active)
        {
            distance_ordered_index[index] = i;
            index++;
        }
	}
	/*Add the special delimiter*/
    distance_ordered_index[index] = -100;



    /*Using bubble sort; At this moment we have all valid candidate indices in the
        fitness ordered index array; Just sort it; index holds the ordered array size*/
	for (i = 0; i < (index - 1); i++)
	{
	    for (j = (i + 1); j < index; j++)
	    {
            if (pop[candidate_index].distance[distance_ordered_index[i]] > pop[candidate_index].distance[distance_ordered_index[j]])
            {
                temp = distance_ordered_index[i];
                distance_ordered_index[i] = distance_ordered_index[j];
                distance_ordered_index[j] = temp;
            }
	    }
	}


    return index;
}


/*******************************************************************************
 * static int do_compute_density(void)                                         *
 *    - This function computes the the density of the elements based on spea2  *
 *                                                                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int do_compute_density()
{

    int i, k;
    int both_equal = 1;
    int active_candidates = 0;
    double distance_kth = 0;


    REPORT_FN_DETAILS;


	/*Once the distance of ith element to all jth elements in the population and archive
        is available then compute the density as the inverse of the distance to the kth
        element in the sorted list of distances where k = sqrt{pop_size+archive_size}*/

	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
    {
	    if (1 == pop[i].active)
	    {
            /*First sort based on distance*/
            active_candidates = sort_candidate_index_on_distance(i);

            k = (int)sqrt(active_candidates);
            distance_kth = pop[i].distance[distance_ordered_index[k]];
            pop[i].density = 1 / (distance_kth + 2);
        }

    }


    return SUCCESS;
}



/*******************************************************************************
 * static double a_distance_from_b(int a_index, int b_index)                   *
 *    - This function computes the distance between a and b candidates based   *
 *      on spea2                                                               *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static double a_distance_from_b(int a_index, int b_index)
{

    int i;
    double distance = 0;
    long a_score, b_score;

//    REPORT_FN_DETAILS;


    if (1 == a_isequalto_b(a_index, b_index))
        return 0;

	for (i = 0; i < SPEA2_FITNESS_DIMENSION; i++)
	{
	    if (0 == i)
	    {
	        a_score = pop[a_index].pmu_computed_fitness_score.counter_score[0];
	        b_score = pop[b_index].pmu_computed_fitness_score.counter_score[0];
	    }
	    else
	    {
	        a_score = pop[a_index].hwmon_computed_fitness_score.counter_score[0];
	        b_score = pop[b_index].hwmon_computed_fitness_score.counter_score[0];
	    }

	    distance += pow(a_score - b_score, 2);
	}

	if (0.0 == distance)
	    distance = SPEA2_MINDOUBLE;

    return (sqrt(distance));
}



/*******************************************************************************
 * static int do_population_distance_compute(void)                             *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int do_population_distance_compute()
{

    double distance;
    int i, j, ret_val;

    REPORT_FN_DETAILS;

    /*First compute the distance of each element from the other element in the
        population and archive population*/
	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
	    if (1 == pop[i].active)
	    {

            for (j = 0; j < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); j++)
            {
                if (1 == pop[j].active)
                {
        	        distance = a_distance_from_b(i,j);
                    pop[i].distance[j] = distance;
                }

            }

	    }
	}


    ret_val = do_compute_density();
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - SPEA2 compute density failed with %s -\n",error_map[ret_val].err_message);
        return OPERATION_ERROR;
    }


    return SUCCESS;
}

/*******************************************************************************
 * static int sort_population_index_on_fitness(void)                           *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int sort_population_index_on_fitness()
{

    int i, j, temp;
    int index = 0;

    REPORT_FN_DETAILS;

    /*First extract valid indices to the fitness index array*/
	for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
	{
        if (1 == pop[i].active)
        {
            fitness_ordered_index[index] = i;
            index++;
        }
	}
	/*Add the special delimiter*/
    fitness_ordered_index[index] = -100;



    /*Using bubble sort; At this moment we have all valid candidate indices in the
        fitness ordered index array; Just sort it; index holds the ordered array size*/
	for (i = 0; i < (index - 1); i++)
	{
	    for (j = (i + 1); j < index; j++)
	    {
            if (pop[fitness_ordered_index[i]].fitness > pop[fitness_ordered_index[j]].fitness)
            {
                temp = fitness_ordered_index[i];
                fitness_ordered_index[i] = fitness_ordered_index[j];
                fitness_ordered_index[j] = temp;
            }
	    }
	}


    return SUCCESS;
}

/*******************************************************************************
 * static int copy_non_dominates_index(void)                                   *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int copy_non_dominates_index()
{

    int i, j, temp;
    int index = 0;

    REPORT_FN_DETAILS;

    while (0 == pop[fitness_ordered_index[index]].fitness)
    {
        non_dominates_index[index] = fitness_ordered_index[index];
        index++;
    }

    return SUCCESS;
}

/*******************************************************************************
 * static int truncate_non_dominates(int non_dominate_count)                   *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int truncate_non_dominates(int non_dominate_count)
{

    int i, j, temp;
    int non_dominates_to_truncate = 0;
    int index = 0;
    int pair1, pair2;
    double smallest_distance;
    /*+1 so we can use -100 as special delimiter*/


    REPORT_FN_DETAILS;

    non_dominates_to_truncate  = non_dominate_count - SPEA2_ARCHIVE_SIZE;
    while (0 < non_dominates_to_truncate)
    {
        printf("non_dominates_to_truncate = %d non_dominate_count = %d\n",non_dominates_to_truncate, non_dominate_count);
        /*Find which is the solution that is most densely packed with another
            non-dominate node.*/
        /*Now compare the mutual distances between each of these non dominate
            nodes and pick the one with smallest and mark that as inactive*/
        /*Set to an arbitrary large value*/
        smallest_distance = 32000;
        for (i = 0; i < non_dominate_count; i++)
        {
        printf("non_dominate[%i] \n",i);

            if (1 == pop[non_dominates_index[i]].active)
            {
                for (j = 0; j < non_dominate_count; j++)
                {
                    /*Avoid comparison of same indexed items*/
                    if (non_dominates_index[i] != non_dominates_index[j])
                    {
                        if (1 == pop[non_dominates_index[j]].active)
                        {
                            if (pop[non_dominates_index[i]].distance[non_dominates_index[j] < smallest_distance])
                            {
                                pair1 = non_dominates_index[i];
                                pair2 = non_dominates_index[j];
                                smallest_distance = pop[pair1].distance[pair2];
                                printf("Smallest distance is pop[%d] <--> pop[%d] = %lf\n",pair1,pair2,smallest_distance);
                            }
                        }
                    }

                }
            }

        }

        /*Now pair1 and pair2 represent the 2 non-dominate solutions which are having the smallest
            distance among themselves, or most dense. Mark the pair1 as inactive as part of truncation*/

        pop[non_dominates_index[pair1]].active = 0;
        /*Now proceed to progress with the truncation from the remaining non-dominate candidates*/
        non_dominates_to_truncate--;
    }

    /*At this point we have got the non dominate solutions truncated. Do the copy operation
        there will be exactly SPEA2_ACHIVE_SIZE active elements in non_dominates_index array*/
    for (i = 0; i < non_dominate_count; i++)
    {
        if (1 == pop[non_dominates_index[i]].active)
        {
            memcpy(&(pop_nextgen[SPEA2_POPULATION_SIZE + i]),&(pop[non_dominates_index[i]]),sizeof(current_node));
        }

    }

    return SUCCESS;
}

/*******************************************************************************
 * static int do_environmental_selection(void)                                 *
 *    - This function performs the computation of distance for the candidates  *
 *      based on spea2                                                         *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int do_environmental_selection()
{
    int count = 0;
    int index = 0;
    int iterator = 0;
    int i, j, ret_val;

    REPORT_FN_DETAILS;

    /*First populate the next generation archive with the non-dominate solutions
        There could be 3 options
        A: Where the number of non-dominate solutions directly fit into the
            archive bucket
        B: Where the number of non-dominate solutions are less than the size of
            archive bucket, here we move dominate solutions from the sorted list where
            fitness > 1
        C: Where the number of non-dominate solutions are more than the size of
            archive bucket, here we truncate the non-dominate solution by iteratively
            removing the solutions that are clustered together*/

    /*First counting how many non-dominate solutions are there*/

    for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
    {
	    if ((1 == pop[i].active) && (0 == pop[i].fitness))
	    {
	        count++;
        }
    }

    if (SPEA2_ARCHIVE_SIZE == count)
    {
        /*Condition A satisfied, simply copy the current non-dominant set of
            solutions to the next gen archive*/
        for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
        {
            if ((1 == pop[i].active) && (0 == pop[i].fitness))
            {
                memcpy(&(pop_nextgen[SPEA2_POPULATION_SIZE + index]),&(pop[i]),sizeof(current_node));
                index++;
            }
        }

    }
    else if (SPEA2_ARCHIVE_SIZE > count)
    {
        /*Condition B satisfied, simply copy the current non-dominant set of
            solutions to the next gen archive, and move dominate solutions from
            the sorted list where Fitness > 1*/

        for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
        {
            if ((1 == pop[i].active) && (0 == pop[i].fitness))
            {
                memcpy(&(pop_nextgen[SPEA2_POPULATION_SIZE + index]),&(pop[i]),sizeof(current_node));
                index++;
            }
        }

        /*sort based on increasing fitness score for those candidates;
            so we can select from those having fitness >= 1*/
        if ( SUCCESS != sort_population_index_on_fitness())
        {
            printf("MAIN$$ ERROR - Spea2 sort fitness ordered list failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }

        /*Copy the dominant solutions from the sorted list to nextgen population archive*/
        while (((SPEA2_POPULATION_SIZE + index) < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE)) \
                && (-100 != fitness_ordered_index[iterator]))
        {
            if (1 <= pop[fitness_ordered_index[iterator]].fitness)
            {
                memcpy(&(pop_nextgen[SPEA2_POPULATION_SIZE + index]),&(pop[fitness_ordered_index[iterator]]),sizeof(current_node));
                index++;
            }

            iterator++;
        }

    }
    else
    {
        /*SPEA2_ARCHIVE_SIZE < count*/
        /*Condition C satisfied, truncate non-dominate solution based on density
         information and copy the result to archive*/

        /*sort based on increasing fitness score for those candidates;
            so we can select from those having fitness == 0*/
        if ( SUCCESS != sort_population_index_on_fitness())
        {
            printf("MAIN$$ ERROR - Spea2 sort fitness ordered list failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }

        if ( SUCCESS != copy_non_dominates_index())
        {
            printf("MAIN$$ ERROR - Spea2 copy non dominates list failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }

        if ( SUCCESS != truncate_non_dominates(count))
        {
            printf("MAIN$$ ERROR - Spea2 truncate non dominates list failed with %s -\n",error_map[ret_val].err_message);
            return OPERATION_ERROR;
        }
    }

    return SUCCESS;
}

/*******************************************************************************
 * static int do_generate_next_population(void)                                *
 *    - This function performs process of generating the next generation of    *
 *      population based on spea2                                              *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int do_generate_next_population()
{

    int i, j, temp;
    int candidate1, candidate2;
    int crossover_point;
    int mutate_action;
    int crossover_action;
    long long size;
    CONSTRAINT_T sparcity;
    ACTORPROFILE_T actors;

    REPORT_FN_DETAILS;

    /*We have the archive list in nextgen population already populated with the
        non-dominate candidates. We perform binary tournament and single point
        crossover and mutation to fill the next generation population*/

    /*First do the mating pool selection from archive population using
        binary tournament*/
	for (i = 0; i < SPEA2_MATINGPOOL_SIZE; i++)
    {
        /*Select 2 random candidates from archive population and add to
            mating pool based on fitness+dessity comparison*/
        candidate1 = generate_rand_archive_index();
        candidate2 = generate_rand_archive_index();
        if ((pop_nextgen[candidate1].fitness + pop_nextgen[candidate1].density) > \
            (pop_nextgen[candidate2].fitness + pop_nextgen[candidate2].density))
        {
                /*winner is candidate 1*/
                memcpy(&(pop_matingpool[i]),&(pop_nextgen[candidate1]),sizeof(current_node));
        }
        else
        {
                /*winner is candidate 2*/
                memcpy(&(pop_matingpool[i]),&(pop_nextgen[candidate2]),sizeof(current_node));
        }

    }

    /*Perform cross-over with mutation from mating pool till population is filled up*/
    for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
    {
        do

        {
            /*Select 2 random candidates from mating pool to perform single-point cross-over with mutation*/
            candidate1 = generate_rand_matingpool_index();
            candidate2 = generate_rand_matingpool_index();
        } while (candidate1 == candidate2); /*We need 2 separate parents to start mating*/


        crossover_action = check_for_crossover();
        /*Check if crossover action is favourable*/
        if (0 != crossover_action)
        {

            crossover_point = generate_rand_crossover_point();

            /*Here we select the cross over point as the one after Size as in
                Si ^ Sp Ac*/
            if (0 == crossover_point)
            {
                /*Child 1 generated*/
                /*Size from parent1*/
                pop_nextgen[i].pl_size = pop_matingpool[candidate1].pl_size;
                pop_nextgen[i].pl_size_row = pop_matingpool[candidate1].pl_size_row;
                pop_nextgen[i].pl_size_col = pop_matingpool[candidate1].pl_size_col;

                /*Sparcity and Actor count from parent*/
                pop_nextgen[i].sparcity_index = pop_matingpool[candidate2].sparcity_index;
                pop_nextgen[i].actor_index = pop_matingpool[candidate2].actor_index;
                pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate2].primary_fitness_fn;

                pop_nextgen[i].active = 1;

                i++;

                if (i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE))
                {
                    /*Size from parent1*/
                    pop_nextgen[i].pl_size = pop_matingpool[candidate2].pl_size;
                    pop_nextgen[i].pl_size_row = pop_matingpool[candidate2].pl_size_row;
                    pop_nextgen[i].pl_size_col = pop_matingpool[candidate2].pl_size_col;

                    /*Sparcity and Actor count from parent*/
                    pop_nextgen[i].sparcity_index = pop_matingpool[candidate1].sparcity_index;
                    pop_nextgen[i].actor_index = pop_matingpool[candidate1].actor_index;
                    pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate1].primary_fitness_fn;

                    pop_nextgen[i].active = 1;
                }
                else
                {
                    /*Check to ensure we break the copying if the population was odd numbered*/
                    break;
                }

            }
            else
            {
                /*Here we select the cross over point as the one after Sparcity as in
                    Si Sp ^ Ac*/
                /*Child 1 generated*/
                /*Size from parent1*/
                pop_nextgen[i].pl_size = pop_matingpool[candidate1].pl_size;
                pop_nextgen[i].pl_size_row = pop_matingpool[candidate1].pl_size_row;
                pop_nextgen[i].pl_size_col = pop_matingpool[candidate1].pl_size_col;
                pop_nextgen[i].sparcity_index = pop_matingpool[candidate1].sparcity_index;

                /*Sparcity and Actor count from parent*/
                pop_nextgen[i].actor_index = pop_matingpool[candidate2].actor_index;
                pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate2].primary_fitness_fn;

                i++;

                if (i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE))
                {
                    /*Size from parent1*/
                    pop_nextgen[i].pl_size = pop_matingpool[candidate2].pl_size;
                    pop_nextgen[i].pl_size_row = pop_matingpool[candidate2].pl_size_row;
                    pop_nextgen[i].pl_size_col = pop_matingpool[candidate2].pl_size_col;
                    pop_nextgen[i].sparcity_index = pop_matingpool[candidate2].sparcity_index;

                    /*Sparcity and Actor count from parent*/
                    pop_nextgen[i].actor_index = pop_matingpool[candidate1].actor_index;
                    pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate1].primary_fitness_fn;
                }
                else
                {
                    /*Check to ensure we break the copying if the population was odd numbered*/
                    break;
                }

            }
        }
        else
        {
                /*Child 1 copied without crossover*/
                pop_nextgen[i].pl_size = pop_matingpool[candidate1].pl_size;
                pop_nextgen[i].pl_size_row = pop_matingpool[candidate1].pl_size_row;
                pop_nextgen[i].pl_size_col = pop_matingpool[candidate1].pl_size_col;
                pop_nextgen[i].sparcity_index = pop_matingpool[candidate1].sparcity_index;
                pop_nextgen[i].actor_index = pop_matingpool[candidate1].actor_index;
                pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate1].primary_fitness_fn;

                pop_nextgen[i].active = 1;

                i++;

                if (i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE))
                {
                    /*Size from parent1*/
                    pop_nextgen[i].pl_size = pop_matingpool[candidate2].pl_size;
                    pop_nextgen[i].pl_size_row = pop_matingpool[candidate2].pl_size_row;
                    pop_nextgen[i].pl_size_col = pop_matingpool[candidate2].pl_size_col;
                    pop_nextgen[i].sparcity_index = pop_matingpool[candidate2].sparcity_index;
                    pop_nextgen[i].actor_index = pop_matingpool[candidate2].actor_index;
                    pop_nextgen[i].primary_fitness_fn = pop_matingpool[candidate2].primary_fitness_fn;

                    pop_nextgen[i].active = 1;
                }
                else
                {
                    /*Check to ensure we break the copying if the population was odd numbered*/
                    break;
                }

        }

        mutate_action = check_for_mutation();

        if (0 != mutate_action)
        {
            /*Perform mutation*/
            if (1 == mutate_action)
            {
                /*Mutate size attribute*/
                size = pop_nextgen[i].pl_size;
    //          printf("Before MUTATION pop_nextgen[%d] size = %lld\n",i, size);
                size+=PL_SIZE_STEP;
                if (size > PL_SIZE_MAX)
                {
                    size = PL_SIZE_MAX;
                }
                pop_nextgen[i].pl_size = size;

    //          printf("After MUTATION pop_nextgen[%d] size = %lld\n",i, size);
                determine_rows_cols(pop_nextgen[i].pl_size);
                pop_nextgen[i].pl_size_row = pl_rows_max;
                pop_nextgen[i].pl_size_col = pl_cols_max;

                size = pop_nextgen[i-1].pl_size;
    //          printf("Before MUTATION pop_nextgen[%d] size = %lld\n",i-1, size);
                size+=PL_SIZE_STEP;
                if (size > PL_SIZE_MAX)
                {
                    size = PL_SIZE_MAX;
                }
                pop_nextgen[i-1].pl_size = size;
    //          printf("After MUTATION pop_nextgen[%d] size = %lld\n",i, size);

                determine_rows_cols(pop_nextgen[i-1].pl_size);
                pop_nextgen[i-1].pl_size_row = pl_rows_max;
                pop_nextgen[i-1].pl_size_col = pl_cols_max;

            }
            else if (2 == mutate_action)
            {
                /*Mutate sparcity index*/
                sparcity = pop_nextgen[i].sparcity_index;
                if (UNCONSTRAINT == sparcity)
                {
                    pop_nextgen[i].sparcity_index = CONSTRAINT_LOW_MEDIUM;
                }
                else if (CONSTRAINT_LOW_MEDIUM == sparcity)
                {
                    pop_nextgen[i].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
                }
                else if (CONSTRAINT_HIGH_MEDIUM == sparcity)
                {
                    pop_nextgen[i].sparcity_index = CONSTRAINT_HIGHEST;
                }
                else
                {
                    pop_nextgen[i].sparcity_index = UNCONSTRAINT;
                }

                sparcity = pop_nextgen[i-1].sparcity_index;
                if (UNCONSTRAINT == sparcity)
                {
                    pop_nextgen[i-1].sparcity_index = CONSTRAINT_LOW_MEDIUM;
                }
                else if (CONSTRAINT_LOW_MEDIUM == sparcity)
                {
                    pop_nextgen[i-1].sparcity_index = CONSTRAINT_HIGH_MEDIUM;
                }
                else if (CONSTRAINT_HIGH_MEDIUM == sparcity)
                {
                    pop_nextgen[i-1].sparcity_index = CONSTRAINT_HIGHEST;
                }
                else
                {
                    pop_nextgen[i-1].sparcity_index = UNCONSTRAINT;
                }

            }
            else
            {
                /*Mutate actor index*/
                actors = pop_nextgen[i].actor_index;
                if (SINGLE_RDWR_AGENT == actors)
                {
                    pop_nextgen[i].actor_index = DUAL_RDWR_AGENT;
                }
                else if (DUAL_RDWR_AGENT == actors)
                {
                    pop_nextgen[i].actor_index = MULTIPLE_RDWR_AGENT;
                }
                else
                {
                    pop_nextgen[i].actor_index = SINGLE_RDWR_AGENT;
                }

                actors = pop_nextgen[i-1].actor_index;
                if (SINGLE_RDWR_AGENT == actors)
                {
                    pop_nextgen[i-1].actor_index = DUAL_RDWR_AGENT;
                }
                else if (DUAL_RDWR_AGENT == actors)
                {
                    pop_nextgen[i-1].actor_index = MULTIPLE_RDWR_AGENT;
                }
                else
                {
                    pop_nextgen[i-1].actor_index = SINGLE_RDWR_AGENT;
                }
            }

        }

    }

    /*Copy from next gen to pop*/
    for (i = 0; i < (SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE); i++)
    {
        memcpy(&(pop[i]),&(pop_nextgen[i]),sizeof(current_node));
    }

    return SUCCESS;
}
/*******************************************************************************
 * int initialise_spea2_setup(PL_T payload, FIT_FN_T fitness_fn)                *
 *    - This function blah blah                                                 *
 ******************************************************************************/
int initialise_spea2_setup(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Initialise CCI Testing statespace with Random read write payload\n");
                ret_val = generate_initial_population(payload, spea2_seeding_firstflag, fitness_fn);
                if (SUCCESS != ret_val)
                {
                    printf("MAIN$$ ERROR - Spea2 setup initialisation failed with %s -\n",error_map[ret_val].err_message);

                    return OPERATION_ERROR;
                }
                spea2_seeding_firstflag = FALSE;

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}




/*******************************************************************************
 * int iterate_spea2_statespace(PL_T payload, FIT_FN_T fitness_fn)             *
 *    - This function blah blah                                                *
 ******************************************************************************/
int iterate_spea2_statespace(PL_T payload, FIT_FN_T fitness_fn)
{
    int ret_val = -1;
    int termination_counter = SPEA2_MAX_GENERATIONS;
    int i;

	REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                printf("MAIN$$ Iterate CCI Testing statespace with Random read write payload\n");

		        while ((time(NULL) < finish_time) && termination_counter >= 0)
		        {

		            ret_val = do_population_fitness_assignment();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - SPEA2 population assignment operation failed with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }
		            ret_val = do_population_distance_compute();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - SPEA2 population distance compute operation failed with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }

                    ret_val = do_environmental_selection();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - SPEA2 environmental selection failed with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }

                    /*This is where the archive will have the selection done from the current generation*/
                    ret_val = write_population_details();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - SPEA2 write population failed with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }

                    ret_val = do_generate_next_population();
                    if (SUCCESS != ret_val)
                    {
                        printf("MAIN$$ ERROR - SPEA2 generate next population failed with %s -\n",error_map[ret_val].err_message);
                        return OPERATION_ERROR;
                    }

                    for (i = 0; i < SPEA2_POPULATION_SIZE; i++)
                	{

                        ret_val = evaluate_population_nodes(CURRENT_AND_BEST, fitness_fn, i);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - SBST algorithm evaluation failed with %s -\n",error_map[ret_val].err_message);

                            return OPERATION_ERROR;
                        }
                    }


                    generation_index++;

		            termination_counter--;

                }

        break;

        default:
            return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}