/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/arm_cci_pmu_fitness_function.h"

/*Global variables*/
char output_file_pmu[PERF_OUT_STR_LEN];
//char output_file_hwmon[PERF_OUT_STR_LEN];
char perf_event_list[MAX_COUNTERS][COUNTER_NAME_STR_LEN];
int perf_counters = 0;
//int dat_process_done = 0;
int dat_process_done_pmu = 0;
int perf_monitor_duration = 5;



/*******************************************************************************
 * void function02 (void)                                                      *
 *    - This function blah blah                                                *
 ******************************************************************************/

void function02 (void)
{
	REPORT_FN_DETAILS;


}

/*******************************************************************************
 * static int discover_cci_pmu_events (char* filter)                           *
 *    - This function lists the cci pmu events available to perf userspace tool*
 *      for monitoring and lists them based on the filter specified via the    *
 *      input argument                                                         *
 *    - arg1 - filter - specifies the predefined filter                        *
 ******************************************************************************/

static int discover_cci_pmu_events (char* filter)
{

	char command_string[64];

	REPORT_FN_DETAILS;

	printf("Listing the supported perf\n");
	sprintf(command_string,"perf list %s > perf_list.txt",filter);
	printf("The command is %s\n",command_string);
	system(command_string);

	return SUCCESS;
}

/*******************************************************************************
 * static int record_cci_pmu_events (void * config)                            *
 *    - This function records the cci pmu events specified by the user as      *
 *      defined by the config input argument                                   *
 *    - arg1 - config - specifies the predefined config definde by the user    *
 ******************************************************************************/

static void * record_cci_pmu_events (void * ptr)
{
	char* perf_args[16];
	char* perf_cmd[512];
	int ret_val = -1;
	pid_t pid;
//	pid_t forked_pid;
	int finished = 0;
	int iter = 0;
    mqd_t mqd_pmu;
    int ret;
 	char msg_buff[MQ_MESSAGE_MAX_LENGTH];
 	ssize_t msg_len;
	char pid_str[8];


	REPORT_FN_DETAILS;

	memset(msg_buff,'\0',sizeof(msg_buff));

	/*Get the process id for the test suite process, so we can pass it to
	  perf utility to monitor/record events*/
    pid = getpid();

	sprintf(pid_str,"%d",pid);

	printf("PERF_THREAD$$ Process being targeted with perf is :: %s\n",pid_str);

    /*Prepare the perf command to be issues in a separate shell process*/
	perf_args[0] = "perf";
	perf_args[1] = "stat";
	perf_args[2] = "-x,";
	perf_args[3] = "-o";
//	perf_args[4] = output_file;
	perf_args[4] = output_file_pmu;
	perf_args[5] = "-a";
 	perf_args[6] = "-e";
    /*perf_args[7] is where perf_cmdline contents will fit*/
	perf_args[8] = "sleep";
	/*perf_args[9] is the perf_monitor_duration
	  ensure the sleep duration in the main thread also matches this*/
	perf_args[10] = NULL;



    printf("PERF_THREAD## Launching perf in a separate thread\n");

    /* Open the message queue. Message queue is already created from a different process */
    mqd_pmu = mq_open(MQ_PMU_FF_NAME, O_RDONLY);
//    mqd_pmu = mq_open(MQ_PMU_FF_NAME, O_RDONLY);

    if( (mqd_t)-1 != mqd_pmu )
    {
	    printf("PERF_THREAD## Message Queue Opened\n");
    }
    else
    {
        printf("PERF_THREAD## ERROR: Message queue %s open failed\n",MQ_PMU_FF_NAME);
//        printf("PERF_THREAD## ERROR: Message queue %s open failed\n",MQ_PMU_FF_NAME);
        ret_val = OS_ERROR;

        pthread_exit(&ret_val);
    }

    while (!finished)
    {

  	    printf ("PERF_THREAD## Iteration %d\n",iter++);
        printf ("PERF_THREAD## Restart_Count %d Generations %d Iterator %d\n",restart_count,generations,iterator);

        printf("PERF_THREAD## Receiving message .... \n");
        msg_len = mq_receive(mqd_pmu, msg_buff, MQ_MESSAGE_MAX_LENGTH, NULL);
        if(msg_len < 0)
        {
  		    printf("PERF_THREAD##  ... Failed\n");
        }
        else
        {
            msg_buff[MQ_MESSAGE_MAX_LENGTH-1] = '\0';
 //           printf("PERF_THREAD## Successfully received %d bytes\n", (int)msg_len);
            printf("PERF_THREAD## FF:: Message : %s \n", msg_buff);
        }

        if (0 == strcmp(msg_buff,"INITIATE"))
        {
//        	printf("PERF_THREAD## Will generate %s ....\n",output_file);
        	printf("PERF_THREAD## Will generate %s ....\n",output_file_pmu);
            printf("PERF_THREAD## Process being targeted with perf is :: %s\n",pid_str);

            /*IMPORTANT - A trail space is important ate end of perf_cmd string*/
			sprintf(perf_cmd,"./%s %s %s %s %s %s %s %s %s %d",perf_args[0],perf_args[1],perf_args[2],perf_args[3],perf_args[4],perf_args[5],perf_args[6],perf_cmdline,perf_args[8],perf_monitor_duration);

			printf("PERF_THREAD## Invoking %s\n",perf_cmd);
//			printf("PERF_THREAD## perf_cmdline is %s\n", perf_cmdline);
            /*First determine whether system() can execute successfully*/
            ret_val = system(NULL);
            if (ret_val == 0)
            {
				printf("PERF_THREAD$$ ERROR Could not invoke any cmd using system(). Fatal Error! Terminating...\n");
				ret_val = OS_ERROR;
				pthread_exit(&ret_val);

            }
            /*Since system could be invoked proceed to launch perf userspace utility*/
			if (-1 == system(perf_cmd))
    		{
				printf("PERF_THREAD$$ ERROR Perf invokation failed\n");
				ret_val = OS_ERROR;
				pthread_exit(&ret_val);
			}
			else
			{
                printf("MAIN$$ system('./perf <blah>') invocation returned %d\n", ret_val);

			}
			/*Force the fitness function thread to wait while perf execution concludes
			  based on the monitoring duration set*/
			sleep(perf_monitor_duration);

			printf("PERF_THREAD## After perf execution\n");

        }
        else if (0 == strcmp(msg_buff,"TERMINATE"))
        {
			printf("PERF_THREAD## FF:: TERMINATE\n");
            finished = 1;
        }
        else
        {
            printf("PERF_THREAD## perf thread waiting for message\n");
        }

	}
    printf("PERF_THREAD## Exiting from FF thread\n");

    ret_val = SUCCESS;
	pthread_exit(&ret_val);
}


/*******************************************************************************
 * static int process_pmu_dat_files (int node_index)                           *
 *    - This function appends the .dat counter values to a report              *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
static int process_pmu_dat_files(int node_index)
{

	FILE* fd = NULL;
	FILE* report_pmu_fd = NULL;
	char str[PERF_OUT_STR_LEN];
	char* tempstr;
	char* freemem;
	char file_info[64];
	char delim[2] = ",";
	char delim1[2] = "_";
	char delim2[2] = ".";
	char * token;
	int tmp_score = -1;
	int x_pos = -1;
	int y_pos = -1;
	int found_best = 0;
	int loads = -1;
	int stores = -1;
	float score_ratio = -1;
	long score;
	float average = 0;
	int counters = 0;
	int termination_countdown = 10; //Avoid an infinite loop by providing a countdown while waiting for the perf output to be generated
    int i =0;

    REPORT_FN_DETAILS;

    /*Progressively append the perf output to the final report.txt while computing the fitness scores
      in the internal data structures as we go along*/
	report_pmu_fd = fopen(report_pmu_name,"a+");
	if (NULL == report_pmu_fd)
	{
		printf("MAIN$$ File open error for %s, Terminating...\n", report_pmu_name);
		return IO_ERROR;
	}
	while ((fd == NULL) && (termination_countdown > 0))
	{
//	    printf("Trying to open %s attempt %d\n", output_file, 11-termination_countdown);
		fd = fopen(output_file_pmu,"r");
		if (NULL == fd)
		{
			printf("MAIN$$ WARNING Is perf system exec prerequisite satisfied? mkdir -p /bin; ln -s /system/bin/sh /bin/sh\n");
			printf("MAIN$$ File open error for dat file, Retrying attempt %d...\n",11-termination_countdown);
			sleep(3);
			termination_countdown--;
		}
	}
//    printf("Termination count is %d\n",termination_countdown);
    if (0 == termination_countdown)
    {
        return OS_ERROR;
    }
	printf("MAIN$$ Processing %s ...\n",output_file_pmu);


//	printf("and cat %s...\n",output_file);
//	sprintf(debugstr,"cat %s",output_file);
//	system(debugstr);

    sprintf(file_info,"\n[%s]\n",output_file_pmu);
	fputs(file_info,report_pmu_fd);

    /*Process the perf output file line by line*/
	while (NULL != fgets(str,PERF_OUT_STR_LEN,fd))
	{
		fputs(str,report_pmu_fd);

		for (i=0; i<perf_counters; i++)
		{
    		/*In addition to copying the string save the best
    		  If the line contains the scores of the searched event save it*/
	    	if (NULL != strstr(str,perf_event_list[i]))
		    {
                tempstr = strdup(str);
                freemem = tempstr;
                token = strsep(&tempstr,delim);
//                printf("token %s tempstr %s",token, tempstr);
//                printf("%s is %d long\n",tempstr,strlen(tempstr));
                tempstr[strcspn(tempstr,"\n")] = '\0';
//    		    printf("Before MAIN$$ %s %s Perf event -> %s\n", token, tempstr, perf_event_list[i]);
                if (0 == strcmp(tempstr,perf_event_list[i]))
                {
//    		        printf("MAIN$$ %s Perf event -> %s\n", tempstr, perf_event_list[i]);
			        score = atol(token);
    		        //printf("MAIN$$ %s Perf event -> %s\n", str, perf_event_list[i]);
			        printf("MAIN$$ %s:: %ld \n", perf_event_list[i], score);

			        if (-1 == node_index)
			        {
				        sprintf(current_node.pmu_computed_fitness_score.counter_name[i],"%s",perf_event_list[i]);
				        if (stat_token_pmu == AVERAGE_STAT_COUNT-1)
				        {
					        current_node.pmu_computed_fitness_score.counter_score[i] += score;
					        average = current_node.pmu_computed_fitness_score.counter_score[i] / AVERAGE_STAT_COUNT;
					        printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_pmu, current_node.pmu_computed_fitness_score.counter_name[i], average, (long)average);
					        current_node.pmu_computed_fitness_score.counter_score[i] = (long)average;
				        }
				        else
				        {
					        current_node.pmu_computed_fitness_score.counter_score[i] += score;
					        printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_pmu, current_node.pmu_computed_fitness_score.counter_name[i], current_node.pmu_computed_fitness_score.counter_score[i]);
				        }

			        }
			        else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
                    {
                        sprintf(next_gen[node_index].pmu_computed_fitness_score.counter_name[i],"%s",perf_event_list[i]);
                        if (stat_token_pmu == AVERAGE_STAT_COUNT-1)
                        {
					       	next_gen[node_index].pmu_computed_fitness_score.counter_score[i] += score;
                            average = next_gen[node_index].pmu_computed_fitness_score.counter_score[i] / AVERAGE_STAT_COUNT;
                            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_pmu, next_gen[node_index].pmu_computed_fitness_score.counter_name[i], average, (long)average);
                            next_gen[node_index].pmu_computed_fitness_score.counter_score[i] = (long)average;
                        }
                        else
                        {
                            next_gen[node_index].pmu_computed_fitness_score.counter_score[i] += score;
                            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_pmu, next_gen[node_index].pmu_computed_fitness_score.counter_name[i], next_gen[node_index].pmu_computed_fitness_score.counter_score[i]);
                        }
                    }
			        else
	                    printf("MAIN$$ Should not reach here.....!!!!\n");


	            }

      		    free(freemem);

	        }

	    }
	} //end of While loop processing the .dat file

	printf("MAIN$$ Global Best score is ::\n");
//	printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",global_best.gen,global_best.x_pos, global_best.y_pos, global_best.explorer_index, global_best.LLC_load_misses, global_best.LLC_store_misses, global_best.load_miss_ratio, global_best.store_miss_ratio);
        printf("MAIN$$ Current generation Best score is ::\n");
//        printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",current_best.gen,current_best.x_pos, current_best.y_pos, current_best.explorer_index, current_best.LLC_load_misses, current_best.LLC_store_misses, current_best.load_miss_ratio, current_best.store_miss_ratio);

    //We want to write a summary after the end of all statistical samples are taken into account
	if (stat_token_pmu == AVERAGE_STAT_COUNT-1)
	{
		if (-1 == node_index)
		{
			sprintf(str,"\n^^^^^Current_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
			fputs(str,report_pmu_fd);
			for (i=0; i<perf_counters; i++)
			{
                sprintf(str,"\n^^^^Current_node -> %s = %ld", current_node.pmu_computed_fitness_score.counter_name[i], current_node.pmu_computed_fitness_score.counter_score[i]);
                fputs(str,report_pmu_fd);
			}
		}
		else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
		{
            sprintf(str,"\n^^^^^Nextgen_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
            fputs(str,report_pmu_fd);
			for (i=0; i<perf_counters; i++)
			{
                sprintf(str,"\n^^^^Nextgen_node[%d] -> %s = %ld", node_index,next_gen[node_index].pmu_computed_fitness_score.counter_name[i], next_gen[node_index].pmu_computed_fitness_score.counter_score[i]);
	    		fputs(str,report_pmu_fd);
    		}
    	}
		else
			printf("MAIN$$ Should not reach here...!!!!\n");
	}


	fclose(fd);
	printf("MAIN$$ Trying to delete %s\n",output_file_pmu);

	remove(output_file_pmu);
	fclose(report_pmu_fd);

    /*Indicate to both single objective and multi-objective indicator variables*/
	dat_process_done_pmu = 1;

//	printf("Hit any key: \n");
//	getchar();

    return SUCCESS;

}



/*******************************************************************************
 * int report_pmu_fitness_function (int node_index)                            *
 *    - This function evaluates the pmu based fitness function using perf      *
 *      based pmu counters                                                     *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_pmu_fitness_function(int node_index)
{

    static int counter_pmu = 0;
    int ret_val = -1;

    REPORT_FN_DETAILS;

	ret_val = process_pmu_dat_files(node_index);
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to report pmu fitness function\n");

	    return OPERATION_ERROR;
	}

    counter_pmu++;
	printf("MAIN$$ PMU FF Counter value = %d\n",counter_pmu);

    return SUCCESS;

}


/*******************************************************************************
 * static int process_pop_based_pmu_dat_files (int node_index)                           *
 *    - This function appends the .dat counter values to a report              *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
static int process_pop_based_pmu_dat_files(int node_index)
{

	FILE* fd = NULL;
	FILE* report_pmu_fd = NULL;
	char str[PERF_OUT_STR_LEN];
	char* tempstr;
	char* freemem;
	char file_info[64];
	char delim[2] = ",";
	char delim1[2] = "_";
	char delim2[2] = ".";
	char * token;
	int tmp_score = -1;
	int x_pos = -1;
	int y_pos = -1;
	int found_best = 0;
	int loads = -1;
	int stores = -1;
	float score_ratio = -1;
	long score;
	float average = 0;
	int counters = 0;
	int termination_countdown = 10; //Avoid an infinite loop by providing a countdown while waiting for the perf output to be generated
    int i =0;

    REPORT_FN_DETAILS;

    /*Progressively append the perf output to the final report.txt while computing the fitness scores
      in the internal data structures as we go along*/
	report_pmu_fd = fopen(report_pmu_name,"a+");
	if (NULL == report_pmu_fd)
	{
		printf("MAIN$$ File open error for %s, Terminating...\n", report_pmu_name);
		return IO_ERROR;
	}
	while ((fd == NULL) && (termination_countdown > 0))
	{
//	    printf("Trying to open %s attempt %d\n", output_file, 11-termination_countdown);
		fd = fopen(output_file_pmu,"r");
		if (NULL == fd)
		{
			printf("MAIN$$ WARNING Is perf system exec prerequisite satisfied? mkdir -p /bin; ln -s /system/bin/sh /bin/sh\n");
			printf("MAIN$$ File open error for dat file, Retrying attempt %d...\n",11-termination_countdown);
			sleep(3);
			termination_countdown--;
		}
	}
//    printf("Termination count is %d\n",termination_countdown);
    if (0 == termination_countdown)
    {
        return OS_ERROR;
    }
	printf("MAIN$$ Processing %s ...\n",output_file_pmu);


//	printf("and cat %s...\n",output_file);
//	sprintf(debugstr,"cat %s",output_file);
//	system(debugstr);

    sprintf(file_info,"\n[%s]\n",output_file_pmu);
	fputs(file_info,report_pmu_fd);

    /*Process the perf output file line by line*/
	while (NULL != fgets(str,PERF_OUT_STR_LEN,fd))
	{
		fputs(str,report_pmu_fd);

		for (i=0; i<perf_counters; i++)
		{
    		/*In addition to copying the string save the best
    		  If the line contains the scores of the searched event save it*/
	    	if (NULL != strstr(str,perf_event_list[i]))
		    {
                tempstr = strdup(str);
                freemem = tempstr;
                token = strsep(&tempstr,delim);
//                printf("token %s tempstr %s",token, tempstr);
//                printf("%s is %d long\n",tempstr,strlen(tempstr));
                tempstr[strcspn(tempstr,"\n")] = '\0';
//    		    printf("Before MAIN$$ %s %s Perf event -> %s\n", token, tempstr, perf_event_list[i]);
                if (0 == strcmp(tempstr,perf_event_list[i]))
                {
//    		        printf("MAIN$$ %s Perf event -> %s\n", tempstr, perf_event_list[i]);
			        score = atol(token);
    		        //printf("MAIN$$ %s Perf event -> %s\n", str, perf_event_list[i]);
			        printf("MAIN$$ %s:: %ld \n", perf_event_list[i], score);

			        if ((node_index >= 0) && (node_index < SPEA2_POPULATION_SIZE))
                    {
                        sprintf(pop[node_index].pmu_computed_fitness_score.counter_name[i],"%s",perf_event_list[i]);
                        if (stat_token_pmu == AVERAGE_STAT_COUNT-1)
                        {
					       	pop[node_index].pmu_computed_fitness_score.counter_score[i] += score;
                            average = pop[node_index].pmu_computed_fitness_score.counter_score[i] / AVERAGE_STAT_COUNT;
                            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_pmu, pop[node_index].pmu_computed_fitness_score.counter_name[i], average, (long)average);
                            pop[node_index].pmu_computed_fitness_score.counter_score[i] = (long)average;
                        }
                        else
                        {
                            pop[node_index].pmu_computed_fitness_score.counter_score[i] += score;
                            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_pmu, pop[node_index].pmu_computed_fitness_score.counter_name[i], pop[node_index].pmu_computed_fitness_score.counter_score[i]);
                        }
                    }
			        else
	                    printf("MAIN$$ Should not reach here.....!!!!\n");


	            }

      		    free(freemem);

	        }

	    }
	} //end of While loop processing the .dat file

	printf("MAIN$$ Global Best score is ::\n");
//	printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",global_best.gen,global_best.x_pos, global_best.y_pos, global_best.explorer_index, global_best.LLC_load_misses, global_best.LLC_store_misses, global_best.load_miss_ratio, global_best.store_miss_ratio);
        printf("MAIN$$ Current generation Best score is ::\n");
//        printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",current_best.gen,current_best.x_pos, current_best.y_pos, current_best.explorer_index, current_best.LLC_load_misses, current_best.LLC_store_misses, current_best.load_miss_ratio, current_best.store_miss_ratio);

    //We want to write a summary after the end of all statistical samples are taken into account
	if (stat_token_pmu == AVERAGE_STAT_COUNT-1)
	{
		if ((node_index >= 0) && (node_index < SPEA2_POPULATION_SIZE))
		{
            sprintf(str,"\n^^^^^Nextgen_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
            fputs(str,report_pmu_fd);
			for (i=0; i<perf_counters; i++)
			{
                sprintf(str,"\n^^^^Nextgen_node[%d] -> %s = %ld", node_index,pop[node_index].pmu_computed_fitness_score.counter_name[i], pop[node_index].pmu_computed_fitness_score.counter_score[i]);
	    		fputs(str,report_pmu_fd);
    		}
    	}
		else
			printf("MAIN$$ Should not reach here...!!!!\n");
	}


	fclose(fd);
	printf("MAIN$$ Trying to delete %s\n",output_file_pmu);

	remove(output_file_pmu);
	fclose(report_pmu_fd);

    /*Indicate to both single objective and multi-objective indicator variables*/
	dat_process_done_pmu = 1;

//	printf("Hit any key: \n");
//	getchar();

    return SUCCESS;

}



/*******************************************************************************
 * int report_pop_based_pmu_fitness_function (int node_index)                            *
 *    - This function evaluates the pmu based fitness function using perf      *
 *      based pmu counters                                                     *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_pop_based_pmu_fitness_function(int node_index)
{

    static int counter_pmu = 0;
    int ret_val = -1;

    REPORT_FN_DETAILS;

	ret_val = process_pop_based_pmu_dat_files(node_index);
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to report pmu fitness function\n");

	    return OPERATION_ERROR;
	}

    counter_pmu++;
	printf("MAIN$$ PMU FF Counter value = %d\n",counter_pmu);

    return SUCCESS;

}


/*******************************************************************************
 * int launch_pmu_fitness_function(void)                                       *
 *    - This function creates a thread to evaluate the fitness function        *
 *      using perf based pmu counters                                          *
 ******************************************************************************/

int launch_pmu_fitness_function(void)
{

	int ret_val = -1;

	REPORT_FN_DETAILS;

    ret_val = pthread_create( &fitness_fn_pmu_thread, NULL, record_cci_pmu_events, (void*) NULL);
//    ret_val = pthread_create( &fitness_fn_pmu_thread, NULL, record_cci_pmu_events, (void*) NULL);
    if (ret_val)
	{
		printf("MAIN$$ ERROR: Unable to create pmu fitness function thread\n");

	    return OS_ERROR;
	}

	return SUCCESS;

}