/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

 #include "include/arm_hwmon_fitness_function.h"
/*Global variables*/
char output_file_hwmon[PERF_OUT_STR_LEN];
int dat_process_done_hwmon = 0;

/*******************************************************************************
 * static int record_hwmon_events (void * config)                              *
 *    - This function records the cci pmu events specified by the user as      *
 *      defined by the config input argument                                   *
 *    - arg1 - config - specifies the predefined config definde by the user    *
 ******************************************************************************/

static void * record_hwmon_events (void * ptr)
{
	char* hwmon_args[16];
	char hwmon_cmd[512];
	int ret_val = -1;
	pid_t pid;
//	pid_t forked_pid;
	int finished = 0;
	int iter = 0;
    mqd_t mqd_hwmon;
    int ret;
 	char msg_buff[MQ_MESSAGE_MAX_LENGTH];
 	ssize_t msg_len;
	char pid_str[8];


	REPORT_FN_DETAILS;

	memset(msg_buff,'\0',sizeof(msg_buff));

	/*Get the process id for the test suite process, so we can pass it to
	  perf utility to monitor/record events*/
    pid = getpid();

	sprintf(pid_str,"%d",pid);

	printf("PERF_THREAD$$ Process being targeted with perf is :: %s\n",pid_str);

    /*Prepare the perf command to be issues in a separate shell process*/
	hwmon_args[0] = "readenergy";
	hwmon_args[1] = "-o";
//	hwmon_args[2] = output_file;
	hwmon_args[2] = output_file_hwmon;
	hwmon_args[3] = "-d";
	/*hwmon_args[4] is the pmu_monitor_duration
	  ensure the sleep duration in the main thread also matches this*/
    /*hwmon_args[5] is where hwmon_cmdline contents will fit*/
	hwmon_args[6] = NULL;

    printf("HWMON_THREAD## Launching readenergy in a separate thread\n");

    /* Open the message queue. Message queue is already created from a different process */
    mqd_hwmon = mq_open(MQ_HWMON_FF_NAME, O_RDONLY);
//    mqd_hwmon = mq_open(MQ_HWMON_FF_NAME, O_RDONLY);

    if( (mqd_t)-1 != mqd_hwmon )
    {
	    printf("HWMON_THREAD## Message Queue Opened\n");
    }
    else
    {
        printf("HWMON_THREAD## ERROR: Message queue %s open failed\n",MQ_HWMON_FF_NAME);
//        printf("HWMON_THREAD## ERROR: Message queue %s open failed\n",MQ_HWMON_FF_NAME);
        ret_val = OS_ERROR;

        pthread_exit(&ret_val);
    }

    while (!finished)
    {

  	    printf ("HWMON_THREAD## Iteration %d\n",iter++);
        printf ("HWMON_THREAD## Restart_Count %d Generations %d Iterator %d\n",restart_count,generations,iterator);

        printf("HWMON_THREAD## Receiving message .... \n");
        msg_len = mq_receive(mqd_hwmon, msg_buff, MQ_MESSAGE_MAX_LENGTH, NULL);
        if(msg_len < 0)
        {
  		    printf("HWMON_THREAD##  ... Failed\n");
        }
        else
        {
            msg_buff[MQ_MESSAGE_MAX_LENGTH-1] = '\0';
 //           printf("HWMON_THREAD## Successfully received %d bytes\n", (int)msg_len);
            printf("HWMON_THREAD## FF:: Message : %s \n", msg_buff);
        }

        if (0 == strcmp(msg_buff,"INITIATE"))
        {
//        	printf("HWMON_THREAD## Will generate %s ....\n",output_file);
        	printf("HWMON_THREAD## Will generate %s ....\n",output_file_hwmon);
            printf("HWMON_THREAD## Process being targeted with perf is :: %s\n",pid_str);

            /*IMPORTANT - A trail space is important at the end of perf_cmd string*/
			sprintf(hwmon_cmd,"./%s %s %s %s %d %s ",hwmon_args[0],hwmon_args[1],hwmon_args[2],hwmon_args[3],perf_monitor_duration,hwmon_cmdline);

			printf("HWMON_THREAD## Invoking %s\n",hwmon_cmd);
//			printf("HWMON_THREAD## perf_cmdline is %s\n", hwmon_cmd);
            /*First determine whether system() can execute successfully*/
            ret_val = system(NULL);
            if (ret_val == 0)
            {
				printf("HWMON_THREAD$$ ERROR Could not invoke any cmd using system(). Fatal Error! Terminating...\n");
				ret_val = OS_ERROR;
				pthread_exit(&ret_val);

            }
            /*Since system could be invoked proceed to launch perf userspace utility*/
			if (-1 == system(hwmon_cmd))
    		{
				printf("HWMON_THREAD$$ ERROR readenergy invokation failed\n");
				ret_val = OS_ERROR;
				pthread_exit(&ret_val);
			}
			else
			{
                printf("MAIN$$ system('./perf <blah>') invocation returned %d\n", ret_val);

			}
			/*Force the fitness function thread to wait while perf execution concludes
			  based on the monitoring duration set*/
			sleep(perf_monitor_duration);

			printf("HWMON_THREAD## After perf execution\n");

        }
        else if (0 == strcmp(msg_buff,"TERMINATE"))
        {
			printf("HWMON_THREAD## FF:: TERMINATE\n");
            finished = 1;
        }
        else
        {
            printf("HWMON_THREAD## perf thread waiting for message\n");
        }

	}
    printf("HWMON_THREAD## Exiting from FF thread\n");

    ret_val = SUCCESS;
	pthread_exit(&ret_val);
}

/*******************************************************************************
 * static int process_hwmon_dat_files (int node_index)                         *
 *    - This function appends the .dat counter values to a report              *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
static int process_hwmon_dat_files(int node_index)
{

	FILE* fd = NULL;
	FILE* report_hwmon_fd = NULL;
	char str[HWMON_OUT_STR_LEN];
	char* tempstr;
	char* freemem;
	char file_info[64];
	char write_str[64];
	char delim[2] = ",";
	char delim1[2] = "_";
	char delim2[2] = ".";
	char* token;
	int tmp_score = -1;
	int x_pos = -1;
	int y_pos = -1;
	int found_best = 0;
	int loads = -1;
	int stores = -1;
    float average;
	float a53_score, a57_score, sys_score, gpu_score;
	float a53_average, a57_average, gpu_average, sys_average;
	int a53_score_count, a57_score_count, sys_score_count, gpu_score_count;
	int termination_countdown = 10; //Avoid an infinite loop by providing a countdown while waiting for the perf output to be generated
    int i = 0;
    int header_found, sys_found, a53_found, a57_found, gpu_found;
    int sys_index, a53_index, a57_index, gpu_index;
    time_t current_time;
	char timestamp[128];

    REPORT_FN_DETAILS;

    /*Progressively append the perf output to the final report.txt while computing the fitness scores
      in the internal data structures as we go along*/
	report_hwmon_fd = fopen(report_hwmon_name,"a+");
	if (NULL == report_hwmon_fd)
	{
		printf("MAIN$$ File open error for %s, Terminating...\n",report_hwmon_name);
		return IO_ERROR;
	}
	while ((fd == NULL) && (termination_countdown > 0))
	{
//	    printf("Trying to open %s attempt %d\n", output_file, 11-termination_countdown);
		fd = fopen(output_file_hwmon,"r");
		if (NULL == fd)
		{
			printf("MAIN$$ WARNING Is perf system exec prerequisite satisfied? mkdir -p /bin; ln -s /system/bin/sh /bin/sh\n");
			printf("MAIN$$ File open error for dat file, Retrying attempt %d...\n",11-termination_countdown);
			sleep(3);
			termination_countdown--;
		}
	}
//    printf("Termination count is %d\n",termination_countdown);
    if (0 == termination_countdown)
    {
        return OS_ERROR;
    }
	printf("MAIN$$ Processing %s ...\n",output_file_hwmon);


//	printf("and cat %s...\n",output_file);
//	sprintf(debugstr,"cat %s",output_file);
//	system(debugstr);

//    sprintf(file_info,"\n[%s]\n",output_file);
    sprintf(file_info,"\n[%s]\n",output_file_hwmon);
	fputs(file_info,report_hwmon_fd);

    current_time = time(NULL);
    sprintf(timestamp,"# started %s\n\n",asctime(localtime(&current_time)));
	fputs(timestamp,report_hwmon_fd);

	header_found = a53_found = a57_found = sys_found = gpu_found = 0;
	sys_index = a57_index = a53_index = gpu_index = -1;
    sys_score = a53_score = a57_score = gpu_score = 0;
    sys_average = a53_average = a57_average = gpu_average = 0;
    sys_score_count = a53_score_count = a57_score_count = gpu_score_count = 0;


    /*Process the perf output file line by line*/
	while (NULL != fgets(str,HWMON_OUT_STR_LEN,fd))
	{


        if (0 == header_found)
        {
            //Parse the header to determine the sys_pow measurement position in the csv line
            tempstr = strdup(str);
            freemem = tempstr;
            i = 0;


            while( (token = strsep(&tempstr,delim)) != NULL)
            {

                if (0 == strcmp(token,"sys_pow"))
                {
                    sys_index = i;
                    sys_found = 1;
                }
                if (0 == strcmp(token,"a53_pow"))
                {
                    a53_index = i;
                    a53_found = 1;
                }
                if (0 == strcmp(token,"a57_pow"))
                {
                    a57_index = i;
                    a57_found = 1;
                }
                if (0 == strcmp(token,"gpu_pow"))
                {
                    gpu_index = i;
                    gpu_found = 1;
               }

                if (sys_found && a53_found && a57_found && gpu_found)
                {
                    header_found = 1;
                    //Got all indices hence break
                    break;
                }

                i++;
            }
            free(freemem);


        }
        else
        {
            //Here compute the average score from the instantaneous measurements

            tempstr = strdup(str);
            freemem = tempstr;
            i = 0;

            while( (token = strsep(&tempstr,delim)) != NULL)
            {
                if (sys_index == i)
                {
                    //Extract the values and compute the average
                    sys_score = (float)atof(token);
                    sys_average += sys_score;
                    sys_score_count++;
                }
                if (a53_index == i)
                {
                    //Extract the values and compute the average
                    a53_score = (float)atof(token);
                    a53_average += a53_score;
                    a53_score_count++;
                }
                if (a57_index == i)
                {
                    //Extract the values and compute the average
                    a57_score = (float)atof(token);
                    a57_average += a57_score;
                    a57_score_count++;
                }
                if (gpu_index == i)
                {
                    //Extract the values and compute the average
                    gpu_score = (float)atof(token);
                    gpu_average += gpu_score;
                    gpu_score_count++;
                }
                i++;
            }
            free(freemem);


        }


	} //end of While loop processing the .dat file


//    printf("\nsys_average = %f a53_average = %f a57_average = %f gpu_average = %f", sys_average, a53_average, a57_average, gpu_average);
//    printf("\nsys_score_count = %d a53_score_count = %d a57_score_count = %d gpu_score_count = %d", sys_score_count, a53_score_count, a57_score_count, gpu_score_count);
    sys_average /= sys_score_count;
    a53_average /= a53_score_count;
    a57_average /= a57_score_count;
    gpu_average /= gpu_score_count;
//    printf("\nsys_average = %f a53_average = %f a57_average = %f gpu_average = %f", sys_average, a53_average, a57_average, gpu_average);

    sprintf(write_str,"%ld,system_power\n",(long)(sys_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,a57_power\n",(long)(a57_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,a53_power\n",(long)(a53_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,gpu_power\n",(long)(gpu_average*1000));
	fputs(write_str,report_hwmon_fd);

    printf("MAIN$$ system_power:: %ld \n", (long)(sys_average*1000));
    printf("MAIN$$ a53_power:: %ld \n", (long)(a53_average*1000));
    printf("MAIN$$ a57_power:: %ld \n", (long)(a57_average*1000));
    printf("MAIN$$ gpu_power:: %ld \n", (long)(gpu_average*1000));
//	printf("Hit any key: \n");
//	getchar();

    if (-1 == node_index)
	{
	    sprintf(current_node.hwmon_computed_fitness_score.counter_name[0],"%s","system_power");
		sprintf(current_node.hwmon_computed_fitness_score.counter_name[1],"%s","a53_power");
		sprintf(current_node.hwmon_computed_fitness_score.counter_name[2],"%s","a57_power");
		sprintf(current_node.hwmon_computed_fitness_score.counter_name[3],"%s","gpu_power");
		if (stat_token_hwmon == AVERAGE_STAT_COUNT-1)
		{
		    current_node.hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
			average = current_node.hwmon_computed_fitness_score.counter_score[0] / AVERAGE_STAT_COUNT;
			printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[0], average, (long)average);
			current_node.hwmon_computed_fitness_score.counter_score[0] = (long)average;

			current_node.hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
			average = current_node.hwmon_computed_fitness_score.counter_score[1] / AVERAGE_STAT_COUNT;
			printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[1], average, (long)average);
			current_node.hwmon_computed_fitness_score.counter_score[1] = (long)average;

			current_node.hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
			average = current_node.hwmon_computed_fitness_score.counter_score[2] / AVERAGE_STAT_COUNT;
			printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[2], average, (long)average);
			current_node.hwmon_computed_fitness_score.counter_score[2] = (long)average;

			current_node.hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
			average = current_node.hwmon_computed_fitness_score.counter_score[3] / AVERAGE_STAT_COUNT;
			printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[3], average, (long)average);
			current_node.hwmon_computed_fitness_score.counter_score[3] = (long)average;
		}
		else
		{
		    current_node.hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
			printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[0], current_node.hwmon_computed_fitness_score.counter_score[0]);

			current_node.hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
			printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[1], current_node.hwmon_computed_fitness_score.counter_score[1]);

			current_node.hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
			printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[2], current_node.hwmon_computed_fitness_score.counter_score[2]);

			current_node.hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
			printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, current_node.hwmon_computed_fitness_score.counter_name[3], current_node.hwmon_computed_fitness_score.counter_score[3]);
		}

    }
	else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
    {
        sprintf(next_gen[node_index].hwmon_computed_fitness_score.counter_name[0],"%s","system_power");
        sprintf(next_gen[node_index].hwmon_computed_fitness_score.counter_name[1],"%s","a53_power");
        sprintf(next_gen[node_index].hwmon_computed_fitness_score.counter_name[2],"%s","a57_power");
        sprintf(next_gen[node_index].hwmon_computed_fitness_score.counter_name[3],"%s","gpu_power");

        if (stat_token_hwmon == AVERAGE_STAT_COUNT-1)
        {
		    next_gen[node_index].hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
            average = next_gen[node_index].hwmon_computed_fitness_score.counter_score[0] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[0], average, (long)average);
            next_gen[node_index].hwmon_computed_fitness_score.counter_score[0] = (long)average;

			next_gen[node_index].hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
            average = next_gen[node_index].hwmon_computed_fitness_score.counter_score[1] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[1], average, (long)average);
            next_gen[node_index].hwmon_computed_fitness_score.counter_score[1] = (long)average;

			next_gen[node_index].hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
            average = next_gen[node_index].hwmon_computed_fitness_score.counter_score[2] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[2], average, (long)average);
            next_gen[node_index].hwmon_computed_fitness_score.counter_score[2] = (long)average;

			next_gen[node_index].hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
            average = next_gen[node_index].hwmon_computed_fitness_score.counter_score[3] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[3], average, (long)average);
            next_gen[node_index].hwmon_computed_fitness_score.counter_score[3] = (long)average;
        }
        else
        {
            next_gen[node_index].hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[0], next_gen[node_index].hwmon_computed_fitness_score.counter_score[0]);

            next_gen[node_index].hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[1], next_gen[node_index].hwmon_computed_fitness_score.counter_score[1]);

            next_gen[node_index].hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[2], next_gen[node_index].hwmon_computed_fitness_score.counter_score[2]);

            next_gen[node_index].hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, next_gen[node_index].hwmon_computed_fitness_score.counter_name[3], next_gen[node_index].hwmon_computed_fitness_score.counter_score[3]);
        }
    }
	else
	    printf("MAIN$$ Should not reach here.....!!!!\n");



	printf("MAIN$$ Global Best score is ::\n");
//	printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",global_best.gen,global_best.x_pos, global_best.y_pos, global_best.explorer_index, global_best.LLC_load_misses, global_best.LLC_store_misses, global_best.load_miss_ratio, global_best.store_miss_ratio);
        printf("MAIN$$ Current generation Best score is ::\n");
//        printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",current_best.gen,current_best.x_pos, current_best.y_pos, current_best.explorer_index, current_best.LLC_load_misses, current_best.LLC_store_misses, current_best.load_miss_ratio, current_best.store_miss_ratio);

    //We want to write a summary after the end of all statistical samples are taken into account
	if (stat_token_hwmon == AVERAGE_STAT_COUNT-1)
	{
		if (-1 == node_index)
		{
			sprintf(str,"\n^^^^^Current_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
			fputs(str,report_hwmon_fd);
			for (i=0; i<4; i++)
			{
                sprintf(str,"\n^^^^Current_node -> %s = %ld", current_node.hwmon_computed_fitness_score.counter_name[i], current_node.hwmon_computed_fitness_score.counter_score[i]);
                fputs(str,report_hwmon_fd);
			}
		}
		else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
		{
            sprintf(str,"\n^^^^^Nextgen_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
            fputs(str,report_hwmon_fd);
			for (i=0; i<4; i++)
			{
                sprintf(str,"\n^^^^Nextgen_node[%d] -> %s = %ld", node_index,next_gen[node_index].hwmon_computed_fitness_score.counter_name[i], next_gen[node_index].hwmon_computed_fitness_score.counter_score[i]);
	    		fputs(str,report_hwmon_fd);
    		}
    	}
		else
			printf("MAIN$$ Should not reach here...!!!!\n");
	}


	fclose(fd);
	printf("MAIN$$ Trying to delete %s\n",output_file_hwmon);

	remove(output_file_hwmon);
	fclose(report_hwmon_fd);

	/*Indicate to both single objective and multi-objective indicator variables*/
	dat_process_done_hwmon = 1;

//	printf("Hit any key: \n");
//	getchar();

    return SUCCESS;

}

/*******************************************************************************
 * int launch_hwmon_fitness_function(void)                                     *
 *    - This function creates a thread to evaluate the fitness function        *
 *      using perf based pmu counters                                          *
 ******************************************************************************/

int launch_hwmon_fitness_function(void)
{

	int ret_val = -1;

	REPORT_FN_DETAILS;

    ret_val = pthread_create( &fitness_fn_hwmon_thread, NULL, record_hwmon_events, (void*) NULL);
//    ret_val = pthread_create( &fitness_fn_hwmon_thread, NULL, record_hwmon_events, (void*) NULL);
    if (ret_val)
	{
		printf("MAIN$$ ERROR: Unable to create hwmon fitness function thread\n");

	    return OS_ERROR;
	}

	return SUCCESS;

}

/*******************************************************************************
 * int report_hwmon_fitness_function (int node_index)                          *
 *    - This function evaluates the hwmon based fitness function using perf    *
 *      based pmu counters                                                     *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_hwmon_fitness_function(int node_index)
{

    static int counter_hwmon = 0;
    int ret_val = -1;

    REPORT_FN_DETAILS;

	ret_val = process_hwmon_dat_files(node_index);
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to report hwmon fitness function\n");

	    return OPERATION_ERROR;
	}

    counter_hwmon++;
	printf("MAIN$$ HWMON FF Counter value = %d\n",counter_hwmon);

    return SUCCESS;

}

/*******************************************************************************
 * static int process_pop_based_hwmon_dat_files (int node_index)                         *
 *    - This function appends the .dat counter values to a report              *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
static int process_pop_based_hwmon_dat_files(int node_index)
{

	FILE* fd = NULL;
	FILE* report_hwmon_fd = NULL;
	char str[HWMON_OUT_STR_LEN];
	char* tempstr;
	char* freemem;
	char file_info[64];
	char write_str[64];
	char delim[2] = ",";
	char delim1[2] = "_";
	char delim2[2] = ".";
	char* token;
	int tmp_score = -1;
	int x_pos = -1;
	int y_pos = -1;
	int found_best = 0;
	int loads = -1;
	int stores = -1;
    float average;
	float a53_score, a57_score, sys_score, gpu_score;
	float a53_average, a57_average, gpu_average, sys_average;
	int a53_score_count, a57_score_count, sys_score_count, gpu_score_count;
	int termination_countdown = 10; //Avoid an infinite loop by providing a countdown while waiting for the perf output to be generated
    int i = 0;
    int header_found, sys_found, a53_found, a57_found, gpu_found;
    int sys_index, a53_index, a57_index, gpu_index;
    time_t current_time;
	char timestamp[128];

    REPORT_FN_DETAILS;

    /*Progressively append the perf output to the final report.txt while computing the fitness scores
      in the internal data structures as we go along*/
	report_hwmon_fd = fopen(report_hwmon_name,"a+");
	if (NULL == report_hwmon_fd)
	{
		printf("MAIN$$ File open error for %s, Terminating...\n",report_hwmon_name);
		return IO_ERROR;
	}
	while ((fd == NULL) && (termination_countdown > 0))
	{
//	    printf("Trying to open %s attempt %d\n", output_file, 11-termination_countdown);
		fd = fopen(output_file_hwmon,"r");
		if (NULL == fd)
		{
			printf("MAIN$$ WARNING Is perf system exec prerequisite satisfied? mkdir -p /bin; ln -s /system/bin/sh /bin/sh\n");
			printf("MAIN$$ File open error for dat file, Retrying attempt %d...\n",11-termination_countdown);
			sleep(3);
			termination_countdown--;
		}
	}
//    printf("Termination count is %d\n",termination_countdown);
    if (0 == termination_countdown)
    {
        return OS_ERROR;
    }
	printf("MAIN$$ Processing %s ...\n",output_file_hwmon);


//	printf("and cat %s...\n",output_file);
//	sprintf(debugstr,"cat %s",output_file);
//	system(debugstr);

//    sprintf(file_info,"\n[%s]\n",output_file);
    sprintf(file_info,"\n[%s]\n",output_file_hwmon);
	fputs(file_info,report_hwmon_fd);

    current_time = time(NULL);
    sprintf(timestamp,"# started %s\n\n",asctime(localtime(&current_time)));
	fputs(timestamp,report_hwmon_fd);

	header_found = a53_found = a57_found = sys_found = gpu_found = 0;
	sys_index = a57_index = a53_index = gpu_index = -1;


    /*Process the perf output file line by line*/
	while (NULL != fgets(str,HWMON_OUT_STR_LEN,fd))
	{


        if (0 == header_found)
        {
            //Parse the header to determine the sys_pow measurement position in the csv line
            tempstr = strdup(str);
            freemem = tempstr;
            i = 0;


            while( (token = strsep(&tempstr,delim)) != NULL)
            {

                if (0 == strcmp(token,"sys_pow"))
                {
                    sys_index = i;
                    sys_found = 1;
                }
                if (0 == strcmp(token,"a53_pow"))
                {
                    a53_index = i;
                    a53_found = 1;
                }
                if (0 == strcmp(token,"a57_pow"))
                {
                    a57_index = i;
                    a57_found = 1;
                }
                if (0 == strcmp(token,"gpu_pow"))
                {
                    gpu_index = i;
                    gpu_found = 1;
               }

                if (sys_found && a53_found && a57_found && gpu_found)
                {
                    header_found = 1;
                    //Got all indices hence break
                    break;
                }

                i++;
            }
            free(freemem);


        }
        else
        {
            //Here compute the average score from the instantaneous measurements

            tempstr = strdup(str);
            freemem = tempstr;
            i = 0;


            while( (token = strsep(&tempstr,delim)) != NULL)
            {
                if (sys_index == i)
                {
                    //Extract the values and compute the average
                    sys_score = (float)atof(token);
                    sys_average += sys_score;
                    sys_score_count++;
                }
                if (a53_index == i)
                {
                    //Extract the values and compute the average
                    a53_score = (float)atof(token);
                    a53_average += a53_score;
                    a53_score_count++;
                }
                if (a57_index == i)
                {
                    //Extract the values and compute the average
                    a57_score = (float)atof(token);
                    a57_average += a57_score;
                    a57_score_count++;
                }
                if (gpu_index == i)
                {
                    //Extract the values and compute the average
                    gpu_score = (float)atof(token);
                    gpu_average += gpu_score;
                    gpu_score_count++;
                }
                i++;
            }
            free(freemem);


        }


	} //end of While loop processing the .dat file


    sys_average /= sys_score_count;
    a53_average /= a53_score_count;
    a57_average /= a57_score_count;
    gpu_average /= gpu_score_count;

    sprintf(write_str,"%ld,system_power\n",(long)(sys_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,a57_power\n",(long)(a57_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,a53_power\n",(long)(a53_average*1000));
	fputs(write_str,report_hwmon_fd);
    sprintf(write_str,"%ld,gpu_power\n",(long)(gpu_average*1000));
	fputs(write_str,report_hwmon_fd);

    printf("MAIN$$ system_power:: %ld \n", (long)(sys_average*1000));
    printf("MAIN$$ a53_power:: %ld \n", (long)(a53_average*1000));
    printf("MAIN$$ a57_power:: %ld \n", (long)(a57_average*1000));
    printf("MAIN$$ gpu_power:: %ld \n", (long)(gpu_average*1000));
//	printf("Hit any key: \n");
//	getchar();

	if ((node_index >= 0) && (node_index < SPEA2_POPULATION_SIZE))
    {
        sprintf(pop[node_index].hwmon_computed_fitness_score.counter_name[0],"%s","system_power");
        sprintf(pop[node_index].hwmon_computed_fitness_score.counter_name[1],"%s","a53_power");
        sprintf(pop[node_index].hwmon_computed_fitness_score.counter_name[2],"%s","a57_power");
        sprintf(pop[node_index].hwmon_computed_fitness_score.counter_name[3],"%s","gpu_power");

        if (stat_token_hwmon == AVERAGE_STAT_COUNT-1)
        {
		    pop[node_index].hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
            average = pop[node_index].hwmon_computed_fitness_score.counter_score[0] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[0], average, (long)average);
            pop[node_index].hwmon_computed_fitness_score.counter_score[0] = (long)average;

			pop[node_index].hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
            average = pop[node_index].hwmon_computed_fitness_score.counter_score[1] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[1], average, (long)average);
            pop[node_index].hwmon_computed_fitness_score.counter_score[1] = (long)average;

			pop[node_index].hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
            average = pop[node_index].hwmon_computed_fitness_score.counter_score[2] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[2], average, (long)average);
            pop[node_index].hwmon_computed_fitness_score.counter_score[2] = (long)average;

			pop[node_index].hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
            average = pop[node_index].hwmon_computed_fitness_score.counter_score[3] / AVERAGE_STAT_COUNT;
            printf("MAIN$$ Token %d Average score for %s is %f -> %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[3], average, (long)average);
            pop[node_index].hwmon_computed_fitness_score.counter_score[3] = (long)average;
        }
        else
        {
            pop[node_index].hwmon_computed_fitness_score.counter_score[0] += (long)(sys_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[0], pop[node_index].hwmon_computed_fitness_score.counter_score[0]);

            pop[node_index].hwmon_computed_fitness_score.counter_score[1] += (long)(a53_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[1], pop[node_index].hwmon_computed_fitness_score.counter_score[1]);

            pop[node_index].hwmon_computed_fitness_score.counter_score[2] += (long)(a57_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[2], pop[node_index].hwmon_computed_fitness_score.counter_score[2]);

            pop[node_index].hwmon_computed_fitness_score.counter_score[3] += (long)(gpu_average*1000);
            printf("MAIN$$ Token %d Running total for %s is %ld\n", stat_token_hwmon, pop[node_index].hwmon_computed_fitness_score.counter_name[3], pop[node_index].hwmon_computed_fitness_score.counter_score[3]);
        }
    }
	else
	    printf("MAIN$$ Should not reach here.....!!!!\n");



	printf("MAIN$$ Global Best score is ::\n");
//	printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",global_best.gen,global_best.x_pos, global_best.y_pos, global_best.explorer_index, global_best.LLC_load_misses, global_best.LLC_store_misses, global_best.load_miss_ratio, global_best.store_miss_ratio);
        printf("MAIN$$ Current generation Best score is ::\n");
//        printf("MAIN$$ In generation %d ---->  (%d,%d) e-index %d = load %d store %d load-ratio %f store-ratio %f\n",current_best.gen,current_best.x_pos, current_best.y_pos, current_best.explorer_index, current_best.LLC_load_misses, current_best.LLC_store_misses, current_best.load_miss_ratio, current_best.store_miss_ratio);

    //We want to write a summary after the end of all statistical samples are taken into account
	if (stat_token_hwmon == AVERAGE_STAT_COUNT-1)
	{
		if ((node_index >= 0) && (node_index < SPEA2_POPULATION_SIZE))
		{
            sprintf(str,"\n^^^^^Nextgen_node -> pl_size[%dx%d]",pl_rows_max,pl_cols_max);
            fputs(str,report_hwmon_fd);
			for (i=0; i<4; i++)
			{
                sprintf(str,"\n^^^^Nextgen_node[%d] -> %s = %ld", node_index,pop[node_index].hwmon_computed_fitness_score.counter_name[i], pop[node_index].hwmon_computed_fitness_score.counter_score[i]);
	    		fputs(str,report_hwmon_fd);
    		}
    	}
		else
			printf("MAIN$$ Should not reach here...!!!!\n");
	}


	fclose(fd);
	printf("MAIN$$ Trying to delete %s\n",output_file_hwmon);

	remove(output_file_hwmon);
	fclose(report_hwmon_fd);

	/*Indicate to both single objective and multi-objective indicator variables*/
	dat_process_done_hwmon = 1;

//	printf("Hit any key: \n");
//	getchar();

    return SUCCESS;

}


/*******************************************************************************
 * int report_pop_based_hwmon_fitness_function (int node_index)                          *
 *    - This function evaluates the hwmon based fitness function using perf    *
 *      based pmu counters                                                     *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_pop_based_hwmon_fitness_function(int node_index)
{

    static int counter_hwmon = 0;
    int ret_val = -1;

    REPORT_FN_DETAILS;

	ret_val = process_pop_based_hwmon_dat_files(node_index);
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to report hwmon fitness function\n");

	    return OPERATION_ERROR;
	}

    counter_hwmon++;
	printf("MAIN$$ HWMON FF Counter value = %d\n",counter_hwmon);

    return SUCCESS;

}
