/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/fitness_function_adaptor.h"

/*Global variables*/
pthread_t fitness_fn_pmu_thread;
pthread_t fitness_fn_hwmon_thread;
//pthread_t data_gen_thread[MAX_CPU_COUNT];
//pthread_t data_obsr_thread[MAX_CPU_COUNT];
pid_t ff_pmu_threadid;
pid_t ff_hwmon_threadid;
//pid_t dg_threadid[MAX_CPU_COUNT];
//pid_t do_threadid[MAX_CPU_COUNT];
mqd_t mqd_singleobj_ff;
mqd_t mqd_pmu_ff;
mqd_t mqd_hwmon_ff;
//FILE* report_pmu_fd;
//FILE* report_hwmon_fd;
/*Global variable to indicate data processing done to all single objective functions*/
//int dat_process_done = 0;
//char output_file_singleobj[PERF_OUT_STR_LEN];


/*******************************************************************************
 * int fitness_function_init_config(FIT_FN_T* fitness_fn)                      *
 *    - This function blah blah                                                *
 ******************************************************************************/
int fitness_function_init_config(FIT_FN_T* fitness_fn)
{
    char str[512];
    int fitnessfn_found = FALSE;
//    int cmdline_found = FALSE;
    int perf_cmdline_found = FALSE;
    int hwmon_cmdline_found = FALSE;
    int perf_instrument_found = FALSE;
    int energy_instrument_found = FALSE;
    int maxcounters_found = FALSE;
    char* tmpstr;
    char * token;
    int i;
    int j;
    int value;

    REPORT_FN_DETAILS;

    /*First determine if the needed configuration parameters PERF or ENERGY
      are provided via the config file. The first encountered will be used*/
    for (i = 0; i < MAX_CFG_SECTIONS; i++)
    {
        strcpy(str,target_config[i].section);
        if (0 == strcmp("PERF", str) || (0 == strcmp("ENERGY", str)) || (0 == strcmp("MULTI", str)) || (0 == strcmp("COMPOSITE", str)))
        {

            if (0 == strcmp("PERF", str))
            {
                *fitness_fn = PMU_MON;
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("perf_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(perf_cmdline, target_config[i].config[j].cfg_val);
                        perf_cmdline_found = TRUE;
                        printf("perf_cmdline found\n");
                    }
                    else if (0 == strcmp("max_counters",target_config[i].config[j].cfg))
                    {
                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        perf_max_counters = value;
                        maxcounters_found = TRUE;
                        printf("perf_maxcounters found\n");
                    }
                    else if (0 == strcmp("hwmon_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(hwmon_cmdline, target_config[i].config[j].cfg_val);
                        hwmon_cmdline_found = TRUE;
                        printf("hwmon_cmdline found\n");
                    }
                    else if (perf_cmdline_found && hwmon_cmdline_found && maxcounters_found)
                    {
                        fitnessfn_found = TRUE;
                        break;
                    }

                }

            }
            else if (0 == strcmp("ENERGY", str))
            {
                *fitness_fn = HW_MON;
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("hwmon_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(hwmon_cmdline, target_config[i].config[j].cfg_val);
                        hwmon_cmdline_found = TRUE;
                        printf("hwmon_cmdline found\n");
                    }
                    else if (0 == strcmp("perf_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(perf_cmdline, target_config[i].config[j].cfg_val);
                        perf_cmdline_found = TRUE;
                        printf("perf_cmdline found\n");
                    }
                    else if (0 == strcmp("max_counters",target_config[i].config[j].cfg))
                    {
                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        perf_max_counters = value;
                        maxcounters_found = TRUE;
                        printf("perf_maxcounters found\n");
                    }
                    else if (hwmon_cmdline_found && perf_cmdline_found && maxcounters_found)
                    {
                        fitnessfn_found = TRUE;
                        break;
                    }
                }

            }
            else if (0 == strcmp("COMPOSITE", str))
            {
                *fitness_fn = COMPOSITE_FF;
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("hwmon_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(hwmon_cmdline, target_config[i].config[j].cfg_val);
                        hwmon_cmdline_found = TRUE;
                        printf("hwmon_cmdline found\n");
                    }
                    else if (0 == strcmp("perf_cmdline", target_config[i].config[j].cfg))
                    {
                        strcpy(perf_cmdline, target_config[i].config[j].cfg_val);
                        perf_cmdline_found = TRUE;
                        printf("perf_cmdline found\n");
                    }
                    else if (0 == strcmp("max_counters",target_config[i].config[j].cfg))
                    {
                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        perf_max_counters = value;
                        maxcounters_found = TRUE;
                        printf("perf_maxcounters found\n");
                    }
                    else if (hwmon_cmdline_found && perf_cmdline_found && maxcounters_found)
                    {
                        fitnessfn_found = TRUE;
                        break;
                    }
                }

            }
            else /*Assume MULTI objective as unconditional else config*/
            {
                *fitness_fn = MULTI_OBJ;
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {

                    if (0 == strcmp("instrument", target_config[i].config[j].cfg))
                    {
                        if (0 == strcmp("PERF", target_config[i].config[j].cfg_val))
                        {
                            strcpy(multiobj_pmu_ff, target_config[i].config[j].cfg_val);
                            perf_instrument_found = TRUE;
                            printf("PERF instrument found\n");
                        }
                        else if (0 == strcmp("ENERGY", target_config[i].config[j].cfg_val))
                        {
                            strcpy(multiobj_hwmon_ff, target_config[i].config[j].cfg_val);
                            energy_instrument_found = TRUE;
                            printf("ENERGY instrument found\n");
                        }

                    }

                    if (0 == strcmp("perf_cmdline", target_config[i].config[j].cfg))
                    {
                        /*Record details of instrument only after instrument tag entry
                        is found first*/
                        if (perf_instrument_found)
                        {
                            strcpy(perf_cmdline, target_config[i].config[j].cfg_val);
                            perf_cmdline_found = TRUE;
                            printf("perf_cmdline found\n");
                        }
                        else
                        {
                            printf("perf_cmdline listed before PERF instrument entry!!! Incorrect order needs fixing!\n");

                        }
                    }
                    else if (0 == strcmp("max_counters",target_config[i].config[j].cfg))
                    {
                        /*Record details of instrument only after instrument tag entry
                        is found first*/
                        if (perf_instrument_found)
                        {
                            sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                            perf_max_counters = value;
                            maxcounters_found = TRUE;
                            printf("perf_maxcounters found\n");
                        }
                        else
                        {
                            printf("perf_maxcounters listed before PERF instrument entry!!! Incorrect order needs fixing!\n");

                        }
                    }
                    else if (0 == strcmp("hwmon_cmdline", target_config[i].config[j].cfg))
                    {
                        /*Record details of instrument only after instrument tag entry
                        is found first*/
                        if (energy_instrument_found)
                        {
                            strcpy(hwmon_cmdline, target_config[i].config[j].cfg_val);
                            hwmon_cmdline_found = TRUE;
                            printf("hwmon_cmdline found\n");
                        }
                        else
                        {
                            printf("hwmon_cmdline listed before ENERGY instrument entry!!! Incorrect order needs fixing!\n");

                        }
                    }
                    else if (perf_instrument_found && energy_instrument_found && perf_cmdline_found && hwmon_cmdline_found && maxcounters_found)
                    {
                        fitnessfn_found = TRUE;
                        break;
                    }

                }

            }

        }

        if (fitnessfn_found) /*Upon finding the first valid fitness function stop the search*/
            break;

    }

    if (!fitnessfn_found)
    {
        printf("MAIN$$ ERROR: No valid fitness function specified in target.config\n");

        return INCOMPLETE_CONFIG;
    }

    switch (*fitness_fn)
    {
        case PMU_MON:
                        printf("MAIN$$ Using %s as Fitness Function with command line %s with max counters %d\n", "PERF", perf_cmdline, perf_max_counters);
                        if ((perf_max_counters <= 0) || (perf_max_counters > MAX_COUNTERS))
                        {
                            return UNSUPPORTED_CONFIG;
                        }

                        printf("----PERF events selected-------\n");
                        /*Extract the PMU counters from the perf commandline and check it is more than max allowed*/
                        /*Do note that since the validity of the selected events happens at the host side
                          we take a permissive approach at the on target component. Don't bother to check again
                          the validity of the selected events*/
                        tmpstr = strdup(perf_cmdline);
                        while( (token = strsep(&tmpstr,"/=")) != NULL )
                        {
                            if (NULL != strstr(token,"cci_"))
                            {
//                                printf("%s\n",token);
                                sprintf(perf_event_list[perf_counters],"%s",token);
                                printf("%s\n",perf_event_list[perf_counters]);
                                perf_counters++;
                            }
                        }
                        printf("-------------------------------\n");
                        free(tmpstr); //strdup will malloc mem internally, so free it explicitly after use to avoid mem leak
                        if (perf_counters > perf_max_counters)
                            return UNSUPPORTED_CONFIG;

                        /*set the report file name*/
                        sprintf(report_singleobj_name,"report-pmu.txt");
                        sprintf(report_pmu_name,"report-pmu.txt");
                        sprintf(report_hwmon_name,"report-hwmon.txt");
                        sprintf(report_shadowobj_name,"report-hwmon.txt");


        break;

        case HW_MON:

                        if ((perf_max_counters <= 0) || (perf_max_counters > MAX_COUNTERS))
                        {
                            return UNSUPPORTED_CONFIG;
                        }

                        printf("----PERF events selected-------\n");
                        /*Extract the PMU counters from the perf commandline and check it is more than max allowed*/
                        /*Do note that since the validity of the selected events happens at the host side
                          we take a permissive approach at the on target component. Don't bother to check again
                          the validity of the selected events*/
                        tmpstr = strdup(perf_cmdline);
                        while( (token = strsep(&tmpstr,"/=")) != NULL )
                        {
                            if (NULL != strstr(token,"cci_"))
                            {
//                                printf("%s\n",token);
                                sprintf(perf_event_list[perf_counters],"%s",token);
                                printf("%s\n",perf_event_list[perf_counters]);
                                perf_counters++;
                            }
                        }
                        printf("-------------------------------\n");
                        free(tmpstr); //strdup will malloc mem internally, so free it explicitly after use to avoid mem leak
                        if (perf_counters > perf_max_counters)
                            return UNSUPPORTED_CONFIG;


                        /*set the report file name*/
                        sprintf(report_singleobj_name,"report-hwmon.txt");
                        sprintf(report_hwmon_name,"report-hwmon.txt");
                        sprintf(report_pmu_name,"report-pmu.txt");
                        sprintf(report_shadowobj_name,"report-pmu.txt");

        break;

        case COMPOSITE_FF:

                        if ((perf_max_counters <= 0) || (perf_max_counters > MAX_COUNTERS))
                        {
                            return UNSUPPORTED_CONFIG;
                        }

                        printf("----PERF events selected-------\n");
                        /*Extract the PMU counters from the perf commandline and check it is more than max allowed*/
                        /*Do note that since the validity of the selected events happens at the host side
                          we take a permissive approach at the on target component. Don't bother to check again
                          the validity of the selected events*/
                        tmpstr = strdup(perf_cmdline);
                        while( (token = strsep(&tmpstr,"/=")) != NULL )
                        {
                            if (NULL != strstr(token,"cci_"))
                            {
//                                printf("%s\n",token);
                                sprintf(perf_event_list[perf_counters],"%s",token);
                                printf("%s\n",perf_event_list[perf_counters]);
                                perf_counters++;
                            }
                        }
                        printf("-------------------------------\n");
                        free(tmpstr); //strdup will malloc mem internally, so free it explicitly after use to avoid mem leak
                        if (perf_counters > perf_max_counters)
                            return UNSUPPORTED_CONFIG;


                        /*set the report file name*/
                        sprintf(report_singleobj_name,"report-hwmon.txt");
                        sprintf(report_hwmon_name,"report-hwmon.txt");
                        sprintf(report_pmu_name,"report-pmu.txt");
                        sprintf(report_shadowobj_name,"report-pmu.txt");
                        sprintf(report_composite_name,"report-composite.txt");

        break;

        case MULTI_OBJ:
                        if (perf_instrument_found)
                        {
                            printf("MAIN$$ Using %s as Fitness Function with command line %s with max counters %d\n", "PERF", perf_cmdline, perf_max_counters);
                            if ((perf_max_counters <= 0) || (perf_max_counters > MAX_COUNTERS))
                            {
                                return UNSUPPORTED_CONFIG;
                            }

                            printf("----PERF events selected-------\n");
                            /*Extract the PMU counters from the perf commandline and check it is more than max allowed*/
                            /*Do note that since the validity of the selected events happens at the host side
                              we take a permissive approach at the on target component. Don't bother to check again
                              the validity of the selected events*/
                            tmpstr = strdup(perf_cmdline);
                            while( (token = strsep(&tmpstr,"/=")) != NULL )
                            {
                                if (NULL != strstr(token,"cci_"))
                                {
//                                  printf("%s\n",token);
                                    sprintf(perf_event_list[perf_counters],"%s",token);
                                    printf("%s\n",perf_event_list[perf_counters]);
                                    perf_counters++;
                                }
                            }
                            printf("-------------------------------\n");
                            free(tmpstr); //strdup will malloc mem internally, so free it explicitly after use to avoid mem leak
                            if (perf_counters > perf_max_counters)
                                return UNSUPPORTED_CONFIG;


                            /*set the report file name*/
                            sprintf(report_pmu_name,"report-pmu.txt");
                            sprintf(report_hwmon_name,"report-hwmon.txt");
                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;
    }


    return SUCCESS;
}


/*******************************************************************************
 * int fitness_function_startup(FIT_FN_T fitness_fn)                           *
 *    - This function blah blah                                                *
 ******************************************************************************/
int fitness_function_startup(FIT_FN_T fitness_fn)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    /*Each fitness function variant is implemented as launch_<variant>_fitness_function()
      method*/
    switch (fitness_fn)
    {
        case PMU_MON:
                        printf("MAIN$$ Using %s as Fitness Function with command line %s with max counters %d\n", "PERF", perf_cmdline, perf_max_counters);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_PMU_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_pmu_ff = mq_open(MQ_PMU_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);
                        /*Assign this to single objective fitness function*/
                        mqd_singleobj_ff = mqd_pmu_ff;


                        ret_val = launch_pmu_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch pmu fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        printf("MAIN$$ Running %s as Shadow Fitness Function with command line %s\n", "ENERGY", hwmon_cmdline);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_HWMON_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_hwmon_ff = mq_open(MQ_HWMON_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);

                        ret_val = launch_hwmon_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch hwmon fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }



        break;

        case HW_MON:
                        printf("MAIN$$ Using %s as Fitness Function with command line %s\n", "ENERGY", hwmon_cmdline);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_HWMON_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_hwmon_ff = mq_open(MQ_HWMON_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);
                        /*Assign this to single objective fitness function*/
                        mqd_singleobj_ff = mqd_hwmon_ff;


                        ret_val = launch_hwmon_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch hwmon fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        printf("MAIN$$ Running %s as Shadow Fitness Function with command line %s with max counters %d\n", "PERF", perf_cmdline, perf_max_counters);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_PMU_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_pmu_ff = mq_open(MQ_PMU_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);

                        ret_val = launch_pmu_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch pmu fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }


        break;

        case COMPOSITE_FF:
                        printf("MAIN$$ Using %s as Fitness Function with command line %s\n", "COMPOSITE_FF", hwmon_cmdline);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_HWMON_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_hwmon_ff = mq_open(MQ_HWMON_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);
                        /*Assign this to single objective fitness function*/
                        mqd_singleobj_ff = mqd_hwmon_ff;


                        ret_val = launch_hwmon_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch hwmon fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        printf("MAIN$$ Running %s as Shadow Fitness Function with command line %s with max counters %d\n", "PERF", perf_cmdline, perf_max_counters);

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_PMU_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_pmu_ff = mq_open(MQ_PMU_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);

                        ret_val = launch_pmu_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch pmu fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }


        break;

        case MULTI_OBJ:
                        printf("MAIN$$ Using %s as Fitness Function\n", "MULTI-OBJECTIVE");

                        /*Setup the FF MsgQ for IPC starting with a cleanup*/
                    	/*Cleanup anything that might be in stale state!*/
                        ret_val = mq_unlink(MQ_PMU_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        ret_val = mq_unlink(MQ_HWMON_FF_NAME);
                        if(ret_val)
                                printf("MAIN$$ Message queue unlink failed");
                        else
                                printf("MAIN$$ Message Queue unlinked\n");
                        /*Now create the Msg Q for PMU FF*/
                        mqd_pmu_ff = mq_open(MQ_PMU_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);

                        mqd_hwmon_ff = mq_open(MQ_HWMON_FF_NAME, O_WRONLY | O_CREAT, MQ_MODE, &attr);

                        ret_val = launch_multiobj_fitness_function();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch multiobj fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;

    }
    return SUCCESS;
}

/*******************************************************************************
 * int fitness_function_teardown(FIT_FN_T fitness_fn)                          *
 *    - This function blah blah                                                *
 ******************************************************************************/
int fitness_function_teardown(FIT_FN_T fitness_fn)
{
    int ret_val = -1;
    int* thread_return[1];

    REPORT_FN_DETAILS;

    switch (fitness_fn)
    {
        case PMU_MON:
                        printf("MAIN$$ Gracefully teardown pmu fitness function setup\n");
                        /*Fall through to the common cleanup logic outside the switch case*/

        break;

        case HW_MON:
                        printf("MAIN$$ Gracefully teardown hwmon fitness function setup\n");
                        /*Fall through to the common cleanup logic outside the switch case*/

        break;

        case COMPOSITE_FF:
                        printf("MAIN$$ Gracefully teardown composite fitness function setup\n");
                        /*Fall through to the common cleanup logic outside the switch case*/

        break;

        case MULTI_OBJ:
                        printf("MAIN$$ Gracefully teardown multiobj fitness function setup\n");

        break;

        default:
                        return NOT_IMPLEMENTED;
        break;

    }

    /*Common teardown logic*/
    printf("MAIN$$ Sending message.... ");
    ret_val = mq_send(mqd_pmu_ff, "TERMINATE", 10, MQ_MESSAGE_PRIORITY);
    if(ret_val)
        printf("MAIN$$ TERMINATE Failed\n");
    else
        printf("MAIN$$ TERMINATE Done\n");

    printf("MAIN$$ Sending message.... ");
    ret_val = mq_send(mqd_hwmon_ff, "TERMINATE", 10, MQ_MESSAGE_PRIORITY);
    if(ret_val)
        printf("MAIN$$ TERMINATE Failed\n");
    else
        printf("MAIN$$ TERMINATE Done\n");

    /*Wait for the fitness function thread before terminating*/
    pthread_join(fitness_fn_pmu_thread,(void**)&(thread_return[0]));
    printf("MAIN$$ PMU Fitness function thread returned %d\n",*(thread_return[0]));

    /*Wait for the fitness function thread before terminating*/
    pthread_join(fitness_fn_hwmon_thread,(void**)&(thread_return[0]));
    printf("MAIN$$ HWMON Fitness function thread returned %d\n",*(thread_return[0]));

    /* Now close the message queue */
    ret_val = mq_close(mqd_pmu_ff);
    if(ret_val)
        printf("MAIN$$ Message queue close failed\n");
    else
        printf("MAIN$$ Message Queue Closed\n");

    ret_val = mq_close(mqd_hwmon_ff);
    if(ret_val)
        printf("MAIN$$ Message queue close failed\n");
    else
        printf("MAIN$$ Message Queue Closed\n");

    ret_val = mq_unlink(MQ_PMU_FF_NAME);
    if(ret_val)
        printf("MAIN$$ Message queue unlink failed\n");
    else
        printf("MAIN$$ Message Queue unlinked\n");

    ret_val = mq_unlink(MQ_HWMON_FF_NAME);
    if(ret_val)
        printf("MAIN$$ Message queue unlink failed\n");
    else
        printf("MAIN$$ Message Queue unlinked\n");


    return SUCCESS;
}

/*******************************************************************************
 * int fitness_function_evaluate(FIT_FN_T fitness_fn)                          *
 *    - This function blah blah                                                *
 ******************************************************************************/
int fitness_function_evaluate(FIT_FN_T fitness_fn, int node_id)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

//	printf("Hit any key: ff=%d nodeid=%d\n", fitness_fn, node_id);
//	getchar();

    switch (fitness_fn)
    {
        case PMU_MON:
                        printf("MAIN$$ Evaluating PMU_MON the fitness function\n");
                        ret_val = report_pmu_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        printf("MAIN$$ Evaluating HW_MON [Shadow] the fitness function\n");
                        ret_val = report_hwmon_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

        break;

        case HW_MON:
                        printf("MAIN$$ Evaluating HW_MON the fitness function\n");
                        ret_val = report_hwmon_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        printf("MAIN$$ Evaluating PMU_MON [Shadow] the fitness function\n");
                        ret_val = report_pmu_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

        break;

        case COMPOSITE_FF:
                        printf("MAIN$$ Evaluating HW_MON the fitness function\n");
                        ret_val = report_hwmon_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }


//                        printf("\nAfter repoooooooooooooooorrrrrrrrrrrrrrrrt HWMON FF.    Hiiit any key");
//                        getchar();

                        printf("MAIN$$ Evaluating PMU_MON [Shadow] the fitness function\n");
                        ret_val = report_pmu_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

//                        printf("\nAfter repoooooooooooooooorrrrrrrrrrrrrrrrt PMUUU FF.    Hiiit any key");
//                        getchar();

                        printf("MAIN$$ Evaluating COMPOSIITE the fitness function\n");
                        ret_val = report_composite_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

//                        printf("\nAfter repoooooooooooooooorrrrrrrrrrrrrrrrt COMPOSIIIITE FF.    Hiiit any key");
//                        getchar();
        break;

        case MULTI_OBJ:
                        printf("MAIN$$ Evaluating MULTI OBJECTIVE the fitness function\n");
                        ret_val = report_multiobj_fitness_function(node_id);
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Evaluating fitness function - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;

    }
//	printf("Hit any key: \n");
//	getchar();

    return SUCCESS;
}
