#    Copyright 2016-2020 University of Leicester & Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from sbstauto.sbst_algo.fitnessfunction import pmu_counters
from sbstauto import config



class PMUCounterTests(unittest.TestCase):
    """
    Tests for device methods/operations.
    """

    def test_perfCounter_add_morethan_max_events(self):
        """
        Test if instance of a perCounter derived class check for maximum supported
        events during configuration stage

        :return:
        """
        pc = pmu_counters.TC2perfCounter()

        self.assertRaises(OverflowError, pc.configEventCounters, "cci_a7_read_data_stall", "cci_a15_read_data_stall",
                          "cci_a7_read_data_stall_trackerfull","cci_a15_read_data_stall_trackerfull",
                          "cci_a7_read_req_stall_id_hzrd")

    def test_perfCounter_unsupported_events(self):
        """
        Test if instance of a perCounter derived class rejects unsupported events
        encounted during configuration

        :return:
        """
        pc = pmu_counters.TC2perfCounter()

        self.assertRaises(ValueError, pc.configEventCounters, "cci_a7_read_data_stall", "Unsupported Event")

    def test_perfCounter_no_events(self):
        """
        Test if instance of a perCounter derived class mandates for atleast
        one event to be specified

        :return:
        """
        pc = pmu_counters.TC2perfCounter()

        self.assertRaises(TypeError, pc.configEventCounters)

    def test_perfCounter_configure_events(self):
        """
        Test if if instance of a perCounter derived class can succesfully add
        valid events as part of the configuration

        :return:
        """
        pc = pmu_counters.TC2perfCounter()
        ev1 = "cci_a7_read_data_stall"
        ev2 = "cci_a7_read_req_stall_id_hzrd"
        ev3 = "cci_a15_read_data_stall_trackerfull"

        pc.configEventCounters(ev1)
        pc.configEventCounters(ev2)
        pc.configEventCounters(ev3)
        pc.configEventCounters(ev2)

        self.assertEqual(ev1,pc._eventlist[0])
        self.assertEqual(ev2, pc._eventlist[1])
        self.assertEqual(ev3, pc._eventlist[2])
        self.assertEqual(ev2, pc._eventlist[3])
        self.assertEqual(4, len(pc._eventlist))

    def test_perfCounter_reset_list(self):
        """
        Test if instance of a perCounter derived class can succesfully empty
        the eventlist by the reset list method

        :return:
        """
        pc = pmu_counters.TC2perfCounter()
        ev1 = "cci_a7_read_data_stall"
        ev2 = "cci_a7_read_req_stall_id_hzrd"
        ev3 = "cci_a15_read_data_stall_trackerfull"

        pc.configEventCounters(ev1)
        pc.configEventCounters(ev2)
        pc.configEventCounters(ev3)
        pc.configEventCounters(ev2)

        self.assertEqual(4, len(pc._eventlist))

        pc.resetEventList()

        self.assertEqual(0, len(pc._eventlist))


    def test_perf_executable_query(self):
        """
        Test if instance of a perCounter derived class can return the correct
        perf executable for the instance created

        :return:
        """
        pc = pmu_counters.TC2perfCounter()
        tc2_perf = config.SBSTAUTO_HOME+"/tools/external/armeabi/perf"

        self.assertEqual(tc2_perf, pc.whichPerfExe())


if __name__ == "__main__":
    unittest.main()
