/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#ifndef _PAYLOAD_CCI_MEM_FITNESS_FUNCTION_COMMON_H_
#define _PAYLOAD_CCI_MEM_FITNESS_FUNCTION_COMMON_H_

/*There is a stronger coupling between the arm_cci_pmu ff and payload_cci_mem_operation payload*/
#include "../../fitnessfunction/include/arm_cci_pmu_fitness_function_common.h"

#define PL_SIZE_1K (16*16)
#define PL_SIZE_64K (256*64)
#define PL_SIZE_512K (256*512)
#define PL_SIZE_1M (256*1024)
#define PL_SIZE_32M (2048*4096)
#define PL_SIZE_64M (4096*4096)
#define PL_SIZE_80M (5120*4096)
#define PL_SIZE_96M (6144*4096)
#define PL_SIZE_128M (4096*8192)
#define PL_SIZE_256M (8192*8192)
#define PL_SIZE_512M (16384*8192)

#define PL_SIZE_MIN 1
#define PL_SIZE_MAX 524288
#define PL_SIZE_STEP 1024

#define DEFAULT_COLS 8192


typedef enum _constraint_type {UNCONSTRAINT=1,CONSTRAINT_LOW_MEDIUM=2,CONSTRAINT_HIGH_MEDIUM=10,CONSTRAINT_HIGHEST=100} CONSTRAINT_T;
typedef enum _actor_type {SINGLE_RDWR_AGENT=1,DUAL_RDWR_AGENT,MULTIPLE_RDWR_AGENT} ACTORPROFILE_T;

/*Used in defining array size for various payload related data structures*/
#define MAX_CPU_COUNT 10

typedef struct _cpu_info
{
    int cpu_id;
    int is_readcpu;
    int cluster_id;
} CPU_INFO_T;

typedef struct _cpu_topology
{
    int cluster_count;
    int cpu_count;
    CPU_INFO_T cpu_info[MAX_CPU_COUNT];
} CPU_TOP_T;

typedef struct _grid_pos
{
	long x_pos;
	long y_pos;
} GRID_POS_T;

typedef struct _composite_score
{
    int pmu_normalised;
    int hwmon_normalised;
} COMP_SCORE;

typedef struct _payload
{
	long long pl_size;
	int pl_size_row;
	int pl_size_col;
	CONSTRAINT_T sparcity_index;
	ACTORPROFILE_T actor_index;
	FIT_FN_T primary_fitness_fn;
	FIT_SCORE_T pmu_computed_fitness_score;
	FIT_SCORE_T hwmon_computed_fitness_score;
	COMP_SCORE composite_fitness_score;
	long fitness;
	int active;
	double distance[SPEA2_POPULATION_SIZE + SPEA2_ARCHIVE_SIZE];
	double density;
    /*The Chromosome encode is done as value type encoding
      in this case of length 3 where:
      Si Sp Ac
      is the representation with Si denoting Size,
      Sp denoting Sparcity Indec and
      Ac denoting actor index
      Size is a continuous number range which can be modified in discrete steps
      Sparcity index is a discrete range from a valid set
      Actor index is a discrete range from a valid set*/
} PAYLOAD_T;



#endif