/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/payload_adaptor.h"

int no_of_clusters = 0;

/*******************************************************************************
 * int payload_init_config (PL_T* payload)                                     *
 *    - This function blah blah                                                *
 ******************************************************************************/
int payload_init_config(PL_T* payload)
{
    char str[CFG_STR_LEN];
    int payload_found = FALSE;
    int payload_type_found = FALSE;
    int mem_region_found = FALSE;
    int clusters_found = FALSE;
    char* tmpstr;
    char * token;
    int i;
    int j;
    int value;

    REPORT_FN_DETAILS;

    /*First determine if the needed configuration parameters CCITEST is
      provided via the config file. The first encountered will be used*/
    for (i = 0; i < MAX_CFG_SECTIONS; i++)
    {
        strcpy(str,target_config[i].section);
        if (0 == strcmp("CCITEST", str))
        {
                strcpy(read_cpupart, "not-specified"); //default assumption is this is not specified
                for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
                {
                    if (0 == strcmp("type", target_config[i].config[j].cfg))
                    {
                        if (0 == strcmp("random_rd_wr", target_config[i].config[j].cfg_val))
                        {
                            *payload = CCI_MEM_RDM_RDWR;
                            payload_type_found = TRUE;
                            printf("payload type found\n");
                        }
                    }
                    else if (0 == strcmp("mem_region_mb", target_config[i].config[j].cfg))
                    {

                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        heap_init_size = value;
                        mem_region_found = TRUE;
                        printf("mem region size found\n");

                    }
                    else if (0 == strcmp("clusters", target_config[i].config[j].cfg))
                    {
                        sscanf(target_config[i].config[j].cfg_val,"%d",&value);
                        no_of_clusters = value;
                        clusters_found = TRUE;
                        printf("cluster count found\n");
                    }
                    else if (0 == strcmp("read_cpupart", target_config[i].config[j].cfg))
                    {
                        strcpy(read_cpupart, target_config[i].config[j].cfg_val);
                        printf("read cpu part found\n");
                    }

                    if (payload_type_found && mem_region_found && clusters_found)
                    {
                        payload_found = TRUE;
                    }


                }


            }
            else
            {
                //TODO: Add logic when the payload type is something else

            }

            if (payload_found) /*Upon finding the first valid payload stop the search*/
                break;


        }

    if (!payload_found)
    {
        printf("MAIN$$ ERROR: No valid payload specified in target.config\n");

        return INCOMPLETE_CONFIG;
    }

    switch (*payload)
    {
        case CCI_MEM_RDM_RDWR:
                        printf("MAIN$$ Using %s as Payload with clusters %d with heap memory initialised to %ld MB\n The read cpu part evaluates to %s\n", "CCITEST -> [memory randome read write]", no_of_clusters, heap_init_size, read_cpupart);
                        if (no_of_clusters < 2)
                        {
                            /*CCI testing assumes atleast 2 clusters*/
                            return UNSUPPORTED_CONFIG;
                        }
                        if (heap_init_size < 32)
                        {
                            /*CCI testing expects atleast 32MB of heap region allocated for the testing*/
                            printf("MAIN$$ Expects a minimum of 32MB memory region for this test, but only %dMB is specified!\n",heap_init_size);
                            return UNSUPPORTED_CONFIG;
                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;
    }

    return SUCCESS;
}



/*******************************************************************************
 * int payload_startup (PL_T payload)                                                      *
 *    - This function blah blah                                                *
 ******************************************************************************/
int payload_startup(PL_T payload)
{
    REPORT_FN_DETAILS;

    int ret_val = -1;

    REPORT_FN_DETAILS;

    /*Each payload variant is implemented as launch_<variant>_payload_operations()
      method*/
    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                        printf("MAIN$$ Using %s as Payload with clusters %d with heap memory initialised to %dMB\n The read cpu part evaluates to %s\n", "CCITEST -> [memory randome read write]", no_of_clusters, heap_init_size, read_cpupart);

                        ret_val = launch_ccitest_payload_operations();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to launch payload operations - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}

/*******************************************************************************
 * int payload_teardown(PL_T payload)                                          *
 *    - This function blah blah                                                *
 ******************************************************************************/
int payload_teardown(PL_T payload)
{
    int ret_val = -1;
    int i = 0;

    REPORT_FN_DETAILS;

    switch (payload)
    {
        case CCI_MEM_RDM_RDWR:
                        printf("MAIN$$ Gracefully teardown Payload setup\n");

                        ret_val = shutdown_ccitest_payload_operations();
                        if (SUCCESS != ret_val)
                        {
                            printf("MAIN$$ ERROR - Unable to shutdown payload operations - failed with error %d -\n",error_map[ret_val].err_code);
                            return ret_val;
                        }

                        /*Free the heap memory used for CCI testing*/
                        free(array_2d);
        break;

        default:
                        return NOT_IMPLEMENTED;
        break;

    }

    return SUCCESS;
}